// Type definitions for Bitwig Studio Control Surface Scripting API v6
// Project: https://bitwig.com
// Definitions by: Joseph Larson <https://github.com/joslarson>
// TypeScript Version: 4.1.2

declare namespace com.bitwig.extension {
  // source: com/bitwig/extension/Extension.java

  import Host = com.bitwig.extension.api.Host;

  class Extension<
    HostType extends Host = Host,
    DefinitionType extends ExtensionDefinition = ExtensionDefinition
  > {
    constructor(extensionDefinition: DefinitionType, host: HostType);

    getHost(): HostType;

    getExtensionDefinition(): DefinitionType;

    mExtensionDefinition: DefinitionType;

    mHost: HostType;
  }

  // source: com/bitwig/extension/ExtensionDefinition.java

  import UUID = java.util.UUID;

  /** Base class for defining any kind of extension for Bitwig Studio. */
  class ExtensionDefinition {
    /** The name of the extension. */
    getName(): string;

    /** The author of the extension. */
    getAuthor(): string;

    /** The version of the extension. */
    getVersion(): string;

    /** A unique id that identifies this extension. */
    getId(): UUID;

    /** The minimum API version number that this extensions requires. */
    getRequiredAPIVersion(): number;

    /**
     * Gets a path within the extension's jar file where documentation for this extension can be found or null
     * if there is none. At the moment this file needs to be a PDF file but other file formats maybe supported
     * in the future.
     */
    getHelpFilePath(): string;

    /**
     * If true then this extension should fail when it calls a deprecated method in the API. This is useful
     * during development.
     */
    shouldFailOnDeprecatedUse(): boolean;

    /**
     * An e-mail address that can be used to contact the author of this extension if a problem is detected with
     * it or null if none.
     */
    getErrorReportingEMail(): string;

    toString(): string;
  }
}

declare namespace com.bitwig.extension.api {
  // source: com/bitwig/extension/api/Color.java

  /**
   * This class represents an RGBA color with each component being stored as double.
   *
   * @since API version 5
   */
  class Color {
    constructor(red: number, green: number, blue: number, alpha: number);

    static fromRGB(red: number, green: number, blue: number): Color;

    static fromRGBA(
      red: number,
      green: number,
      blue: number,
      alpha: number
    ): Color;

    static fromRGB255(red: number, green: number, blue: number): Color;

    static fromRGBA255(
      red: number,
      green: number,
      blue: number,
      alpha: number
    ): Color;

    static fromHex(hex: string): Color;

    /**
     * Mixes two colors.
     * @since API version 4
     */
    static mix(c1: Color, c2: Color, blend: number): Color;

    static nullColor(): Color;

    static blackColor(): Color;

    static whiteColor(): Color;

    getRed(): number;

    getGreen(): number;

    getBlue(): number;

    getAlpha(): number;

    getRed255(): number;

    getGreen255(): number;

    getBlue255(): number;

    getAlpha255(): number;

    mRed: number;

    mGreen: number;

    mBlue: number;

    mAlpha: number;
  }

  // source: com/bitwig/extension/api/Host.java

  import OscModule = com.bitwig.extension.api.opensoundcontrol.OscModule;

  /**
   * Defines the interface through which an extension can talk to the host application.
   */
  interface Host {
    /**
     * Returns the latest supported API version of the host application.
     *
     * @return the latest supported API version of the host application
     * @since API version 1
     */
    getHostApiVersion(): number;

    /**
     * Returns the vendor of the host application.
     *
     * @return the vendor of the host application
     * @since API version 1
     */
    getHostVendor(): string;

    /**
     * Returns the product name of the host application.
     *
     * @return the product name of the host application
     * @since API version 1
     */
    getHostProduct(): string;

    /**
     * Returns the version number of the host application.
     *
     * @return the version number of the host application
     * @since API version 1
     */
    getHostVersion(): string;

    /**
     * The platform type that this host is running on.
     */
    getPlatformType(): PlatformType;

    /**
     * Sets an email address to use for reporting errors found in this script.
     *
     * @since API version 2
     */
    setErrorReportingEMail(address: string): void;

    /**
     * Gets the OpenSoundControl module.
     * @since API version 5
     */
    getOscModule(): OscModule;
  }

  // source: com/bitwig/extension/api/PlatformType.java

  enum PlatformType {
    WINDOWS = 0,
    LINUX = 1,
    MAC = 2,
  }
}

declare namespace com.bitwig.extension.api.opensoundcontrol {
  // source: com/bitwig/extension/api/opensoundcontrol/OscAddressSpace.java

  /**
   * An OSC address space.
   *
   * It contains the root OscContainer.
   *
   * @since API version 5
   */
  interface OscAddressSpace {
    /**
     * Register all the methods annotated with @OscMethod from object.
     * Also if a method is annotated with @OscNode, this method will be called and the returned object's method
     * will be registered.
     */
    registerObjectMethods(addressPrefix: string, object: object): void;

    /**
     * Low level way to register an Osc Method.
     * @param address The address to register the method at
     * @param typeTagPattern The globing pattern used to match the type tag. Pass "*" to match anything.
     * @param desc The method description.
     * @param callback Then call handler.
     */
    registerMethod(
      address: string,
      typeTagPattern: string,
      desc: string,
      callback: OscMethodCallback
    ): void;

    /**
     * This method will be called if no registered OscMethod could handle incoming OscPacket.
     */
    registerDefaultMethod(callback: OscMethodCallback): void;

    /**
     * Should the address spaces log the messages it dispatches?
     * Default is false.
     */
    setShouldLogMessages(shouldLogMessages: boolean): void;

    /**
     * This gives a display name for this address space.
     * It is useful if you have multiple address space to identify them when we generate the documentation.
     */
    setName(name: string): void;
  }

  // source: com/bitwig/extension/api/opensoundcontrol/OscBundle.java

  import List = java.util.List;

  /**
   * An OSC Bundle.
   *
   * @since API version 5
   */
  interface OscBundle extends OscPacket {
    getNanoseconds(): number;

    getPackets(): List<OscPacket>;
  }

  // source: com/bitwig/extension/api/opensoundcontrol/OscConnection.java

  /**
   * This interface lets you send OscMessage through an connection which can be via Tcp, Udp, or whatever.
   *
   * OscPackets are sent when all the startBundle() have a matching endBundle().
   * If you call sendMessage() with startBundle() before, then the message will be sent directly.
   *
   * Our maximum packet size is 64K.
   *
   * @since API version 5
   */
  interface OscConnection {
    /**
     * Starts an OscBundle.
     */
    startBundle(): void;

    /**
     * Supported object types:
     * - Integer for int32
     * - Long for int64
     * - Float for float
     * - Double for double
     * - null for nil
     * - Boolean for true and false
     * - String for string
     * - byte[] for blob
     */
    sendMessage(address: string, ...args: object[]): void;

    /**
     * Finishes the previous bundle, and if it was not inside an other bundle, it will send the message
     * directly.
     */
    endBundle(): void;
  }

  // source: com/bitwig/extension/api/opensoundcontrol/OscMessage.java

  /**
   * An OSC message.
   *
   * @since API version 5
   */
  interface OscMessage extends OscPacket {
    getAddressPattern(): string;

    getTypeTag(): string;

    getArguments(): List<object>;

    getString(index: number): string;

    getBlob(index: number): number;

    getInt(index: number): number;

    getLong(index: number): number;

    getFloat(index: number): number;

    getDouble(index: number): number;

    getBoolean(index: number): boolean;
  }

  // source: com/bitwig/extension/api/opensoundcontrol/OscMethod.java

  // source: com/bitwig/extension/api/opensoundcontrol/OscMethodCallback.java

  interface OscMethodCallback {
    (source: OscConnection, message: OscMessage): void;
  }

  // source: com/bitwig/extension/api/opensoundcontrol/OscModule.java

  /**
   * Interface to create Osc related object.
   *
   * @since API version 5
   */
  interface OscModule {
    /**
     * Creates a new OscAddressSpace.
     *
     * In short the OscAddressSpace dispatches the incoming messages to services.
     * An OscAddressSpace is an OscService.
     *
     * @since API version 5
     */
    createAddressSpace(): OscAddressSpace;

    /**
     * Creates a new OSC Server.
     *
     * @param service Use createAddressSpace
     *
     * @return a new OscServer
     * @since API version 5
     */
    createUdpServer(port: number, addressSpace: OscAddressSpace): void;

    /**
     * Tries to connect to an OscServer.
     *
     * @param addressSpace can be null
     *
     * @return a new OscConnection
     * @since API version 5
     */
    connectToUdpServer(
      host: string,
      port: number,
      addressSpace: OscAddressSpace
    ): OscConnection;
  }

  // source: com/bitwig/extension/api/opensoundcontrol/OscNode.java

  // source: com/bitwig/extension/api/opensoundcontrol/OscPacket.java

  /**
   * Base class for OscPackets.
   *
   * @since API version 5
   */
  interface OscPacket {
    /**
     * If the message was part of a bundle, get a pointer back to it.
     * If not, this methods returns null.
     */
    getParentBundle(): OscBundle;
  }
}

declare namespace com.bitwig.extension.api.util.midi {
  // source: com/bitwig/extension/api/util/midi/ShortMidiMessage.java

  class ShortMidiMessage {
    static NOTE_OFF: number;

    static NOTE_ON: number;

    static POLY_PRESSURE: number;

    static CONTROL_CHANGE: number;

    static PROGRAM_CHANGE: number;

    static CHANNEL_PRESSURE: number;

    static PITCH_BEND: number;

    static MIDI_TIME_CODE: number;

    static SONG_POSITION_POINTER: number;

    static SONG_SELECT: number;

    static TUNE_REQUEST: number;

    static TIMING_CLOCK: number;

    static START: number;

    static CONTINUE: number;

    static STOP: number;

    static ACTIVE_SENSING: number;

    static SYSTEM_RESET: number;

    constructor(status: number, data1: number, data2: number);

    getStatusByte(): number;

    getData1(): number;

    getData2(): number;

    getChannel(): number;

    getStatusMessage(): number;

    isNoteOff(): boolean;

    isNoteOn(): boolean;

    isPolyPressure(): boolean;

    isControlChange(): boolean;

    isProgramChange(): boolean;

    isChannelPressure(): boolean;

    mData: number;
  }

  // source: com/bitwig/extension/api/util/midi/SysexBuilder.java

  class SysexBuilder {
    static MAX_LENGTH: number;

    static fromHex(hexString: string): SysexBuilder;

    addByte(value: number): SysexBuilder;

    addString(string: string, length: number): SysexBuilder;

    add(bytes: number): SysexBuilder;

    addHex(hex: string): SysexBuilder;

    terminate(): number;

    array(): number;

    mData: number;

    mLength: number;
  }
}

declare namespace com.bitwig.extension.callback {
  // source: com/bitwig/extension/callback/BooleanValueChangedCallback.java

  interface BooleanValueChangedCallback extends ValueChangedCallback {
    (newValue: boolean): void;
  }

  // source: com/bitwig/extension/callback/Callback.java

  interface Callback {}

  // source: com/bitwig/extension/callback/ClipLauncherSlotBankPlaybackStateChangedCallback.java

  interface ClipLauncherSlotBankPlaybackStateChangedCallback extends Callback {
    /**
     * Registers an observer that reports the playback state of clips / slots. The reported states include
     * `stopped`, `playing`, `recording`, but also `queued for stop`, `queued for playback`, `queued for
     * recording`.
     *
     * @param callback
     *           a callback function that receives three parameters: 1. the slot index (integer), 2. the queued
     *           or playback state: `0` when stopped, `1` when playing, or `2` when recording, and 3. a boolean
     *           parameter indicating if the second argument is referring to the queued state (`true`) or the
     *           actual playback state (`false`)
     * @since API version 1
     */
    (slotIndex: number, playbackState: number, isQueued: boolean): void;
  }

  // source: com/bitwig/extension/callback/ColorValueChangedCallback.java

  interface ColorValueChangedCallback extends ValueChangedCallback {
    /**
     * As alpha component was introduced after this interface was released,
     * the alpha component is not part of the parameter and would have to be
     * checked manually.
     */
    (red: number, green: number, blue: number): void;
  }

  // source: com/bitwig/extension/callback/ConnectionEstablishedCallback.java

  import RemoteConnection = com.bitwig.extension.controller.api.RemoteConnection;

  interface ConnectionEstablishedCallback extends Callback {
    (connection: RemoteConnection): void;
  }

  // source: com/bitwig/extension/callback/DataReceivedCallback.java

  interface DataReceivedCallback extends Callback {
    (data: number): void;
  }

  // source: com/bitwig/extension/callback/DirectParameterDisplayedValueChangedCallback.java

  interface DirectParameterDisplayedValueChangedCallback extends Callback {
    (id: string, value: string): void;
  }

  // source: com/bitwig/extension/callback/DirectParameterNameChangedCallback.java

  interface DirectParameterNameChangedCallback extends Callback {
    (id: string, name: string): void;
  }

  // source: com/bitwig/extension/callback/DirectParameterNormalizedValueChangedCallback.java

  interface DirectParameterNormalizedValueChangedCallback extends Callback {
    (id: string, normalizedValue: number): void;
  }

  // source: com/bitwig/extension/callback/DoubleValueChangedCallback.java

  interface DoubleValueChangedCallback extends ValueChangedCallback {
    (newValue: number): void;
  }

  // source: com/bitwig/extension/callback/EnumValueChangedCallback.java

  interface EnumValueChangedCallback
    extends ObjectValueChangedCallback<string> {}

  // source: com/bitwig/extension/callback/FloatValueChangedCallback.java

  interface FloatValueChangedCallback extends Callback {
    (newValue: number): void;
  }

  // source: com/bitwig/extension/callback/IndexedBooleanValueChangedCallback.java

  interface IndexedBooleanValueChangedCallback
    extends IndexedValueChangedCallback {
    /**
     * Registers an observer that reports the names of the scenes and slots. The slot names reflect the names
     * of containing clips.
     *
     * @param callback
     *           a callback function receiving two parameters: 1. the slot index (integer) within the
     *           configured window, and 2. the name of the scene/slot (string)
     * @since API version 1
     */
    (index: number, newValue: boolean): void;
  }

  // source: com/bitwig/extension/callback/IndexedColorValueChangedCallback.java

  interface IndexedColorValueChangedCallback
    extends IndexedValueChangedCallback {
    /**
     * Registers an observer that reports the names of the scenes and slots. The slot names reflect the names
     * of containing clips.
     *
     * @param callback
     *           a callback function receiving two parameters: 1. the slot index (integer) within the
     *           configured window, and 2. the name of the scene/slot (string)
     * @since API version 1
     */
    (index: number, red: number, green: number, blue: number): void;
  }

  // source: com/bitwig/extension/callback/IndexedStringValueChangedCallback.java

  interface IndexedStringValueChangedCallback
    extends IndexedValueChangedCallback {
    /**
     * Registers an observer that reports the names of the scenes and slots. The slot names reflect the names
     * of containing clips.
     *
     * @param callback
     *           a callback function receiving two parameters: 1. the slot index (integer) within the
     *           configured window, and 2. the name of the scene/slot (string)
     * @since API version 1
     */
    (index: number, newValue: string): void;
  }

  // source: com/bitwig/extension/callback/IndexedValueChangedCallback.java

  interface IndexedValueChangedCallback extends Callback {}

  // source: com/bitwig/extension/callback/IntegerValueChangedCallback.java

  interface IntegerValueChangedCallback extends ValueChangedCallback {
    (newValue: number): void;
  }

  // source: com/bitwig/extension/callback/NoArgsCallback.java

  interface NoArgsCallback extends Callback {
    (): void;
  }

  // source: com/bitwig/extension/callback/NotePlaybackCallback.java

  interface NotePlaybackCallback extends Callback {
    (isNoteOn: boolean, key: number, velocity: number): void;
  }

  // source: com/bitwig/extension/callback/ObjectValueChangedCallback.java

  interface ObjectValueChangedCallback<ValueType> extends ValueChangedCallback {
    (newValue: ValueType): void;
  }

  // source: com/bitwig/extension/callback/ShortMidiDataReceivedCallback.java

  interface ShortMidiDataReceivedCallback extends Callback {
    /**
     * Registers a callback for receiving short (normal) MIDI messages on this MIDI input port.
     *
     * @param callback
     *           a callback function that receives three integer parameters: 1. the status byte 2. the data1
     *           value 2. the data2 value
     * @ @since API version 1
     */
    (statusByte: number, data1: number, data2: number): void;
  }

  // source: com/bitwig/extension/callback/ShortMidiMessageReceivedCallback.java

  import ShortMidiMessage = com.bitwig.extension.api.util.midi.ShortMidiMessage;

  interface ShortMidiMessageReceivedCallback
    extends ShortMidiDataReceivedCallback {
    /**
     * Registers a callback for receiving short (normal) MIDI messages on this MIDI input port.
     *
     * @param callback
     *           a callback function that receives a ShortMidiMessage instance.
     * @ @since API version 2
     */
    (msg: ShortMidiMessage): void;

    (statusByte: number, data1: number, data2: number): void;
  }

  // source: com/bitwig/extension/callback/StepDataChangedCallback.java

  interface StepDataChangedCallback extends Callback {
    /**
     * A callback function that receives three parameters: 1. the x (step) coordinate within the note grid
     * (integer), 2. the y (key) coordinate within the note grid (integer), and 3. an integer value that
     * indicates if the step is empty (`0`) or if a note continues playing (`1`) or starts playing (`2`).
     */
    (x: number, y: number, state: number): void;
  }

  // source: com/bitwig/extension/callback/StringArrayValueChangedCallback.java

  interface StringArrayValueChangedCallback
    extends ObjectValueChangedCallback<string[]> {}

  // source: com/bitwig/extension/callback/StringValueChangedCallback.java

  interface StringValueChangedCallback
    extends ObjectValueChangedCallback<string> {}

  // source: com/bitwig/extension/callback/SysexMidiDataReceivedCallback.java

  interface SysexMidiDataReceivedCallback extends Callback {
    /**
     * @param data
     *           The data encoded as a hex string
     */
    (data: string): void;
  }

  // source: com/bitwig/extension/callback/ValueChangedCallback.java

  interface ValueChangedCallback extends Callback {}
}

declare namespace com.bitwig.extension.controller {
  // source: com/bitwig/extension/controller/AutoDetectionMidiPortNames.java

  class AutoDetectionMidiPortNames {
    constructor(inputNames: string[], outputNames: string[]);

    getInputNames(): string[];

    getOutputNames(): string[];

    mInputNames: string[];
  }

  // source: com/bitwig/extension/controller/AutoDetectionMidiPortNamesList.java

  import List = java.util.List;

  class AutoDetectionMidiPortNamesList {
    add(inputNames: string[], outputNames: string[]): void;

    getPortNames(): List<AutoDetectionMidiPortNames>;

    getCount(): number;

    getPortNamesAt(index: number): AutoDetectionMidiPortNames;

    mList: List<AutoDetectionMidiPortNames>;
  }

  // source: com/bitwig/extension/controller/ControllerExtension.java

  import Extension = com.bitwig.extension.Extension;
  import ControllerHost = com.bitwig.extension.controller.api.ControllerHost;
  import MidiIn = com.bitwig.extension.controller.api.MidiIn;
  import MidiOut = com.bitwig.extension.controller.api.MidiOut;

  /**
   * Defines an extension that enabled a controller to work with Bitwig Studio.
   */
  class ControllerExtension extends Extension<
    ControllerHost,
    ControllerExtensionDefinition
  > {
    constructor(
      definition: ControllerExtensionDefinition,
      host: ControllerHost
    );

    getMidiInPort(index: number): MidiIn;

    getMidiOutPort(index: number): MidiOut;

    /**
     * Initializes this controller extension. This will be called once when the extension is started. During initialization the
     * extension should call the various create methods available via the {@link ControllerHost} interface in order to
     * create objects used to communicate with various parts of the Bitwig Studio application (e.g
     * {@link ControllerHost#createCursorTrack(int, int)}.
     */
    init(): void;

    /**
     * Called once when this controller extension is stopped.
     */
    exit(): void;

    /**
     * Called when this controller extension should flush any pending updates to the controller.
     */
    flush(): void;
  }

  // source: com/bitwig/extension/controller/ControllerExtensionDefinition.java

  import ExtensionDefinition = com.bitwig.extension.ExtensionDefinition;
  import PlatformType = com.bitwig.extension.api.PlatformType;

  /**
   * Defines an extension that enabled a controller to work with Bitwig Studio.
   * */
  class ControllerExtensionDefinition extends ExtensionDefinition {
    toString(): string;

    /** The vendor of the controller that this extension is for. */
    getHardwareVendor(): string;

    /** The model name of the controller that this extension is for. */
    getHardwareModel(): string;

    /** The number of MIDI in ports that this controller extension has. */
    getNumMidiInPorts(): number;

    /** The number of MIDI out ports that this controller extension has. */
    getNumMidiOutPorts(): number;

    /**
     * Obtains a {@link AutoDetectionMidiPortNamesList} that defines the names of the MIDI in and out ports
     * that can be used for auto detection of the controller for the supplied platform type.
     */
    getAutoDetectionMidiPortNamesList(
      platformType: PlatformType
    ): AutoDetectionMidiPortNamesList;

    /**
     * Lists the {@link AutoDetectionMidiPortNames} that defines the names of the MIDI in and out ports
     * that can be used for auto detection of the controller for the supplied platform type.
     */
    listAutoDetectionMidiPortNames(
      list: AutoDetectionMidiPortNamesList,
      platformType: PlatformType
    ): void;

    /** Creates an instance of this extension. */
    createInstance(host: ControllerHost): ControllerExtension;
  }
}

declare namespace com.bitwig.extension.controller.api {
  // source: com/bitwig/extension/controller/api/Action.java

  /**
   * Instances of this interface represent actions in Bitwig Studio, such as commands that can be launched from
   * the main menu or via keyboard shortcuts.
   *
   * To receive the list of all actions provided by Bitwig Studio call {@link Application#getActions()}. The
   * list of actions that belong to a certain category can be queried by calling
   * {@link ActionCategory#getActions()}. Access to specific actions is provided in
   * {@link Application#getAction(String)}.
   *
   * @since API version 1
   */
  interface Action {
    /**
     * Returns a string the identifies this action uniquely.
     *
     * @return the identifier string
     * @since API version 1
     */
    getId(): string;

    /**
     * Returns the name of this action.
     *
     * @return the name string
     * @since API version 1
     */
    getName(): string;

    /**
     * Returns the category of this action.
     *
     * @return the category string
     * @since API version 1
     */
    getCategory(): ActionCategory;

    /**
     * Returns the text that is displayed in menu items associated with this action.
     *
     * @return the menu item text
     * @since API version 1
     */
    getMenuItemText(): string;

    /**
     * Invokes the action.
     *
     * @since API version 1
     */
    invoke(): void;
  }

  // source: com/bitwig/extension/controller/api/ActionCategory.java

  /**
   * Instances of this interface are used to categorize actions in Bitwig Studio. The list of action categories
   * provided by Bitwig Studio can be queried by calling {@link Application#getActionCategories()}. To receive a
   * specific action category call {@link Application#getActionCategory(String)}.
   *
   * @see Application#getActionCategories()
   * @see Application#getActionCategory(String)
   * @since API version 1
   */
  interface ActionCategory {
    /**
     * Returns a string the identifies this action category uniquely.
     *
     * @return the identifier string
     * @since API version 1
     */
    getId(): string;

    /**
     * Returns the name of this action category.
     *
     * @return the name string
     * @since API version 1
     */
    getName(): string;

    /**
     * Lists all actions in this category.
     *
     * @return the array of actions in this category
     * @since API version 1
     */
    getActions(): Action[];
  }

  // source: com/bitwig/extension/controller/api/Application.java

  import BooleanValueChangedCallback = com.bitwig.extension.callback.BooleanValueChangedCallback;
  import StringValueChangedCallback = com.bitwig.extension.callback.StringValueChangedCallback;

  /**
   * An interface that provides methods for accessing the most common global application commands.<br/>
   *
   * In addition, functions are provided for accessing any application action in a generic and categorized way,
   * pretty much as displayed in the Bitwig Studio commander dialog (see {@link #getActions()},
   * {@link #getAction(String)}, {@link #getActionCategories()}), {@link #getActionCategory(String)}).<br/>
   *
   * To receive an instance of the application interface call {@link ControllerHost#createApplication()}.
   *
   * @since API version 1
   */
  interface Application {
    /**
     * This identifier can be used as parameter for {@link #setPanelLayout(String)} in order to switch to the
     * `ARRANGE` panel layout that is available in various display profiles.
     */
    PANEL_ARRANGE_ARRANGE: 'ARRANGE';

    /**
     * This identifier can be used as parameter for {@link #setPanelLayout(String)} in order to switch to the
     * `EDIT` panel layout that is available in various display profiles.
     */
    PANEL_MIX_ARRANGE: 'MIX';

    /**
     * This identifier can be used as parameter for {@link #setPanelLayout(String)} in order to switch to the
     * `MIX` panel layout that is available in various display profiles.
     */
    PANEL_EDIT_ARRANGE: 'EDIT';

    /**
     * Creates a new audio track at the given position.
     *
     * @param position
     *           the index within the list of main tracks where the new track should be inserted, or `-1` in
     *           case the track should be inserted at the end of the list. Values outside the valid range will
     *           get pinned to the valid range, so the actual position might be different from the provided
     *           parameter value.
     * @since API version 1
     */
    createAudioTrack(position: number): void;

    /**
     * Creates a new instrument track at the given position.
     *
     * @param position
     *           the index within the list of main tracks where the new track should be inserted, or `-1` in
     *           case the track should be inserted at the end of the list. Values outside the valid range will
     *           get pinned to the valid range, so the actual position might be different from the provided
     *           parameter value.
     * @since API version 1
     */
    createInstrumentTrack(position: number): void;

    /**
     * Creates a new effect track at the given position.
     *
     * @param position
     *           the index within the list of effect tracks where the new track should be inserted, or `-1` in
     *           case the track should be inserted at the end of the list. Values outside the valid range will
     *           get pinned to the valid range, so the actual position might be different from the provided
     *           parameter value.
     * @since API version 1
     */
    createEffectTrack(position: number): void;

    /**
     * Returns a list of actions that the application supports. Actions are commands in Bitwig Studio that are
     * typically accessible through menus or keyboard shortcuts.
     *
     * Please note that many of the commands encapsulated by the reported actions are also accessible through
     * other (probably more convenient) interfaces methods of the API. In contrast to that, this method
     * provides a more generic way to find available application functionality.
     *
     * @return the list of actions
     * @ @since API version 1
     */
    getActions(): Action[];

    /**
     * Returns the action for the given action identifier. For a list of available actions, see
     * {@link #getActions()}.
     *
     * @param id
     *           the action identifier string, must not be `null`
     * @return the action associated with the given id, or null in case there is no action with the given
     *         identifier.
     * @ @since API version 1
     */
    getAction(id: string): Action;

    /**
     * Returns a list of action categories that is used by Bitwig Studio to group actions into categories.
     *
     * @return the list of action categories
     * @ @since API version 1
     */
    getActionCategories(): ActionCategory[];

    /**
     * Returns the action category associated with the given identifier. For a list of available action
     * categories, see {@link #getActionCategories()}.
     *
     * @param id
     *           the category identifier string, must not be `null`
     * @return the action associated with the given id, or null in case there is no category with the given
     *         identifier
     * @ @since API version 1
     */
    getActionCategory(id: string): ActionCategory;

    /**
     * Activates the audio engine in Bitwig Studio.
     *
     * @since API version 1
     */
    activateEngine(): void;

    /**
     * Deactivates the audio engine in Bitwig Studio.
     *
     * @since API version 1
     */
    deactivateEngine(): void;

    /**
     * Value that reports whether an audio engine is active or not.
     *
     * @since API version 2
     */
    hasActiveEngine(): BooleanValue;

    /**
     * Registers an observer that gets called when the audio engine becomes active or inactive.
     *
     * @param callable
     *           a callback function that accepts a single boolean parameter. The callback parameter indicates
     *           whether the audio engine became active (true) or inactive (false).
     * @since API version 1
     * @deprecated Use {@link #hasActiveEngine()} instead.
     */
    addHasActiveEngineObserver(callable: BooleanValueChangedCallback): void;

    /**
     * Value that reports the name of the current project.
     *
     * @since API version 2
     */
    projectName(): StringValue;

    /**
     * Registers an observer that reports the name of the current project.
     *
     * @param callback
     *           a callback function that accepts a single string parameter.
     * @param maxChars
     *           the maximum length of the reported name. Longer names will get truncated.
     * @since API version 1
     * @deprecated Use {@link #projectName()} instead.
     */
    addProjectNameObserver(
      callback: StringValueChangedCallback,
      maxChars: number
    ): void;

    /**
     * Switches to the next project tab in Bitwig Studio.
     *
     * @since API version 1
     */
    nextProject(): void;

    /**
     * Switches to the previous project tab in Bitwig Studio.
     *
     * @since API version 1
     */
    previousProject(): void;

    /**
     * Set BitwigStudio to navigate into the group.
     *
     * @since API version 2
     */
    navigateIntoTrackGroup(track: Track): void;

    /**
     * Set BitwigStudio to navigate into the parent group.
     *
     * @since API version 2
     */
    navigateToParentTrackGroup(): void;

    /**
     * Sends an undo command to Bitwig Studio.
     *
     * @since API version 1
     */
    undo(): void;

    /**
     * Sends a redo command to Bitwig Studio.
     *
     * @since API version 1
     */
    redo(): void;

    /**
     * Switches the Bitwig Studio user interface to the panel layout with the given name. The list of available
     * panel layouts depends on the active display profile.
     *
     * @param panelLayout
     *           the name of the new panel layout
     * @since API version 1
     */
    setPanelLayout(panelLayout: string): void;

    /**
     * Switches to the next panel layout of the active display profile in Bitwig Studio.
     *
     * @since API version 1
     */
    nextPanelLayout(): void;

    /**
     * Switches to the previous panel layout of the active display profile in Bitwig Studio.
     *
     * @since API version 1
     */
    previousPanelLayout(): void;

    /**
     * Value that reports the name of the active panel layout.
     *
     * @since API version 2
     */
    panelLayout(): StringValue;

    /**
     * Registers an observer that reports the name of the active panel layout.
     *
     * @param callable
     *           a callback function object that accepts a single string parameter
     * @param maxChars
     *           the maximum length of the panel layout name
     * @since API version 1
     * @deprecated Use {@link #panelLayout()} instead.
     */
    addPanelLayoutObserver(
      callable: StringValueChangedCallback,
      maxChars: number
    ): void;

    /**
     * Value that reports the name of the active display profile.
     *
     * @since API version 2
     */
    displayProfile(): StringValue;

    /**
     * Registers an observer that reports the name of the active display profile.
     *
     * @param callable
     *           a callback function object that accepts a single string parameter
     * @param maxChars
     *           the maximum length of the display profile name
     * @since API version 1
     * @deprecated Use {@link #displayProfile()} instead.
     */
    addDisplayProfileObserver(
      callable: StringValueChangedCallback,
      maxChars: number
    ): void;

    /**
     * Toggles the visibility of the inspector panel.
     *
     * @since API version 1
     */
    toggleInspector(): void;

    /**
     * Toggles the visibility of the device chain panel.
     *
     * @since API version 1
     */
    toggleDevices(): void;

    /**
     * Toggles the visibility of the mixer panel.
     *
     * @since API version 1
     */
    toggleMixer(): void;

    /**
     * Toggles the visibility of the note editor panel.
     *
     * @since API version 1
     */
    toggleNoteEditor(): void;

    /**
     * Toggles the visibility of the automation editor panel.
     *
     * @since API version 1
     */
    toggleAutomationEditor(): void;

    /**
     * Toggles the visibility of the browser panel.
     *
     * @since API version 1
     */
    toggleBrowserVisibility(): void;

    /**
     * Shows the previous detail panel (note editor, device, automation).
     *
     * @since API version 1
     */
    previousSubPanel(): void;

    /**
     * Shows the next detail panel (note editor, device, automation).
     *
     * @since API version 1
     */
    nextSubPanel(): void;

    /**
     * Equivalent to an Arrow-Left key stroke on the computer keyboard. The concrete functionality depends on
     * the current keyboard focus in Bitwig Studio.
     *
     * @since API version 1
     */
    arrowKeyLeft(): void;

    /**
     * Equivalent to an Arrow-Right key stroke on the computer keyboard. The concrete functionality depends on
     * the current keyboard focus in Bitwig Studio.
     *
     * @since API version 1
     */
    arrowKeyRight(): void;

    /**
     * Equivalent to an Arrow-Up key stroke on the computer keyboard. The concrete functionality depends on the
     * current keyboard focus in Bitwig Studio.
     *
     * @since API version 1
     */
    arrowKeyUp(): void;

    /**
     * Equivalent to an Arrow-Down key stroke on the computer keyboard. The concrete functionality depends on
     * the current keyboard focus in Bitwig Studio.
     *
     * @since API version 1
     */
    arrowKeyDown(): void;

    /**
     * Equivalent to an Enter key stroke on the computer keyboard. The concrete functionality depends on the
     * current keyboard focus in Bitwig Studio.
     *
     * @since API version 1
     */
    enter(): void;

    /**
     * Equivalent to an Escape key stroke on the computer keyboard. The concrete functionality depends on the
     * current keyboard focus in Bitwig Studio.
     *
     * @since API version 1
     */
    escape(): void;

    /**
     * Selects all items according the current selection focus in Bitwig Studio.
     *
     * @since API version 1
     */
    selectAll(): void;

    /**
     * Deselects any items according the current selection focus in Bitwig Studio.
     *
     * @since API version 1
     */
    selectNone(): void;

    /**
     * Cuts the selected items in Bitwig Studio if applicable.
     *
     * @since API version 1
     */
    cut(): void;

    /**
     * Copies the selected items in Bitwig Studio to the clipboard if applicable.
     *
     * @since API version 1
     */
    copy(): void;

    /**
     * Pastes the clipboard contents into the current selection focus in Bitwig Studio if applicable.
     *
     * @since API version 1
     */
    paste(): void;

    /**
     * Duplicates the active selection in Bitwig Studio if applicable.
     *
     * @since API version 1
     */
    duplicate(): void;

    /**
     * Deletes the selected items in Bitwig Studio if applicable. Originally this function was called `delete`
     * (Bitwig Studio 1.0). But as `delete` is reserved in JavaScript this function got renamed to `remove` in
     * Bitwig Studio 1.0.9.
     *
     * @since API version 1
     */
    remove(): void;

    /**
     * Opens a text input field in Bitwig Studio for renaming the selected item.
     *
     * @since API version 1
     */
    rename(): void;

    /**
     * Zooms in one step into the currently focused editor of the Bitwig Studio user interface.
     *
     * @since API version 1
     */
    zoomIn(): void;

    /**
     * Zooms out one step in the currently focused editor of the Bitwig Studio user interface.
     *
     * @since API version 1
     */
    zoomOut(): void;

    /**
     * Adjusts the zoom level of the currently focused editor so that it matches the active selection.
     *
     * @since API version 1
     */
    zoomToSelection(): void;

    /**
     * Adjusts the zoom level of the currently focused editor so that all content becomes visible.
     *
     * @since API version 1
     */
    zoomToFit(): void;

    /**
     * Moves the panel focus to the panel on the left of the currently focused panel.
     *
     * @since API version 1
     */
    focusPanelToLeft(): void;

    /**
     * Moves the panel focus to the panel right to the currently focused panel.
     *
     * @since API version 1
     */
    focusPanelToRight(): void;

    /**
     * Moves the panel focus to the panel above the currently focused panel.
     *
     * @since API version 1
     */
    focusPanelAbove(): void;

    /**
     * Moves the panel focus to the panel below the currently focused panel.
     *
     * @since API version 1
     */
    focusPanelBelow(): void;

    /**
     * Toggles between full screen and windowed user interface.
     *
     * @since API version 1
     */
    toggleFullScreen(): void;

    /**
     * @deprecated Use {@link #setPanelLayout(java.lang.String)} instead.
     * @since API version 1
     */
    setPerspective(perspective: string): void;

    /**
     * @deprecated Use {@link #nextPanelLayout()} instead.
     * @since API version 1
     */
    nextPerspective(): void;

    /**
     * @deprecated Use {@link #previousPanelLayout()} instead.
     * @since API version 1
     */
    previousPerspective(): void;

    /**
     * @deprecated Use {@link #addPanelLayoutObserver(org.mozilla.javascript.Callable, int)} instead.
     * @since API version 1
     */
    addSelectedModeObserver(
      callable: StringValueChangedCallback,
      maxChars: number,
      fallbackText: string
    ): void;
  }

  // source: com/bitwig/extension/controller/api/Arranger.java

  /**
   * An interface representing various commands which can be performed on the Bitwig Studio arranger.<br/>
   *
   * To receive an instance of the application interface call {@link ControllerHost#createArranger}.
   *
   * @since API version 1
   */
  interface Arranger {
    /**
     * Gets an object that allows to enable/disable arranger playback follow. Observers can be registered on
     * the returned object for receiving notifications when the setting switches between on and off.
     *
     * @return a boolean value object that represents the enabled state of arranger playback follow
     * @since API version 1
     */
    isPlaybackFollowEnabled(): SettableBooleanValue;

    /**
     * Gets an object that allows to control the arranger track height. Observers can be registered on the
     * returned object for receiving notifications when the track height changes.
     *
     * @return a boolean value object that has the state `true` when the tracks have double row height and
     *         `false` when the tracks have single row height.
     * @since API version 1
     */
    hasDoubleRowTrackHeight(): SettableBooleanValue;

    /**
     * Gets an object that allows to show/hide the cue markers in the arranger panel. Observers can be
     * registered on the returned object for receiving notifications when the cue marker lane switches between
     * shown and hidden.
     *
     * @return a boolean value object that represents the cue marker section visibility
     * @since API version 1
     */
    areCueMarkersVisible(): SettableBooleanValue;

    /**
     * Gets an object that allows to show/hide the clip launcher in the arranger panel. Observers can be
     * registered on the returned object for receiving notifications when the clip launcher switches between
     * shown and hidden.
     *
     * @return a boolean value object that represents the clip launcher visibility
     * @since API version 1
     */
    isClipLauncherVisible(): SettableBooleanValue;

    /**
     * Gets an object that allows to show/hide the timeline in the arranger panel. Observers can be registered
     * on the returned object for receiving notifications when the timeline switches between shown and hidden.
     *
     * @return a boolean value object that represents the timeline visibility
     * @since API version 1
     */
    isTimelineVisible(): SettableBooleanValue;

    /**
     * Gets an object that allows to show/hide the track input/output choosers in the arranger panel. Observers
     * can be registered on the returned object for receiving notifications when the I/O section switches
     * between shown and hidden.
     *
     * @return a boolean value object that represents the visibility of the track I/O section
     * @since API version 1
     */
    isIoSectionVisible(): SettableBooleanValue;

    /**
     * Gets an object that allows to show/hide the effect tracks in the arranger panel. Observers can be
     * registered on the returned object for receiving notifications when the effect track section switches
     * between shown and hidden.
     *
     * @return a boolean value object that represents the visibility of the effect track section
     * @since API version 1
     */
    areEffectTracksVisible(): SettableBooleanValue;

    /**
     * Returns an object that provides access to a bank of successive cue markers using a window configured with
     * the given size, that can be scrolled over the list of markers.
     *
     * @param size
     *           the number of simultaneously accessible items
     * @return the requested item bank object
     */
    createCueMarkerBank(size: number): CueMarkerBank;

    /**
     * Registers an observer that reports if playback-follow is enabled.
     *
     * @param callback
     *           a callback function object that accepts a single bool parameter
     * @see #isPlaybackFollowEnabled()
     * @deprecated call `isPlaybackFollowEnabled().addValueObserver` instead
     * @since API version 1
     */
    addPlaybackFollowObserver(callback: BooleanValueChangedCallback): void;

    /**
     * Registers an observer that reports the current configuration of the arranger track row height.
     *
     * @param callback
     *           a callback function object that accepts a single bool parameter. The parameter indicates if
     *           the row height is double (`true`) or single (`false`).
     * @see #hasDoubleRowTrackHeight()
     * @deprecated call `hasDoubleRowTrackHeight().addValueObserver` instead
     * @since API version 1
     */
    addTrackRowHeightObserver(callback: BooleanValueChangedCallback): void;

    /**
     * Registers an observer that reports if the cue marker lane is visible.
     *
     * @param callback
     *           a callback function object that accepts a single bool parameter.
     * @see #areCueMarkersVisible()
     * @deprecated call `areCueMarkersVisible().addValueObserver` instead
     * @since API version 1
     */
    addCueMarkerVisibilityObserver(callback: BooleanValueChangedCallback): void;

    /**
     * Toggles the playback follow state.
     *
     * @see #isPlaybackFollowEnabled()
     * @deprecated call `isPlaybackFollowEnabled().toggle` instead
     * @since API version 1
     */
    togglePlaybackFollow(): void;

    /**
     * Toggles the arranger track row height between `double` and `single`.
     *
     * @see #hasDoubleRowTrackHeight()
     * @deprecated call `hasDoubleRowTrackHeight().toggle` instead
     * @since API version 1
     */
    toggleTrackRowHeight(): void;

    /**
     * Toggles the visibility of the arranger cue marker lane.
     *
     * @see #areCueMarkersVisible()
     * @deprecated call `areCueMarkersVisible().toggle` instead
     * @since API version 1
     */
    toggleCueMarkerVisibility(): void;
  }

  // source: com/bitwig/extension/controller/api/Bank.java

  /**
   * A bank provides access to a range of items in Bitwig Studio. Instances of a bank are configured with a
   * fixed number of items and represent an excerpt of a larger list of items. Various methods are provided for
   * scrolling to different sections of the item list. It basically acts like a window moving over the list of
   * underlying items.
   *
   * @since API version 2
   */
  interface Bank<ItemType extends ObjectProxy = ObjectProxy>
    extends ObjectProxy,
      Scrollable {
    /**
     * The fixed size of this bank.
     *
     * @since API version 2
     */
    getSizeOfBank(): number;

    scrollPageForwards(): void;

    scrollPageBackwards(): void;

    /**
     * Gets the item in the bank at the supplied index. The index must be >= 0 and < {@link #getSizeOfBank()}.
     *
     * @since API version 2
     */
    getItemAt(index: number): ItemType;

    /**
     * Value that reports the underlying total item count (not the number of items available in the bank
     * window).
     *
     * @since API version 2
     */
    itemCount(): IntegerValue;

    /**
     * An integer value that defines the location of the cursor that this bank is following. If there is no
     * cursor or the cursor is not within the bank then the value is -1.
     *
     * @since API version 2
     */
    cursorIndex(): SettableIntegerValue;
  }

  // source: com/bitwig/extension/controller/api/BeatTimeFormatter.java

  /**
   * Defines a formatter for a beat time that can convert a beat time to a string for display to the user.
   *
   * @since API version 2
   */
  interface BeatTimeFormatter {
    /**
     * Formats the supplied beat time as a string in the supplied time signature.
     *
     * @param beatTime
     *           The beat time to be formatted
     * @param isAbsolute
     *           If true the beat time represents an absolute time (such as a time on the arranger) otherwise
     *           it represents a beat time duration (such as the length of a clip).
     *
     * @since API version 2
     */
    formatBeatTime(
      beatTime: number,
      isAbsolute: boolean,
      timeSignatureNumerator: number,
      timeSignatureDenominator: number,
      timeSignatureTicks: number
    ): string;
  }

  // source: com/bitwig/extension/controller/api/BeatTimeValue.java

  import DoubleValueChangedCallback = com.bitwig.extension.callback.DoubleValueChangedCallback;

  /**
   * Instances of this interface represent beat time values.
   *
   * @since API version 1
   */
  interface BeatTimeValue extends DoubleValue {
    /**
     * Add an observer which receives the internal raw of the parameter as floating point.
     *
     * @param callback
     *           a callback function that receives a single numeric parameter with double precision.
     * @since API version 1
     * @deprecated This exists for backwards compatibility. Use
     *             {@link #addValueObserver(DoubleValueChangedCallback)} instead.
     */
    addRawValueObserver(callback: DoubleValueChangedCallback): void;

    /**
     * Gets the current beat time formatted according to the supplied formatter.
     *
     * @since API version 2
     */
    getFormatted(formatter: BeatTimeFormatter): string;

    /**
     * Gets the current beat time formatted according to the default beat time formatter.
     *
     * @see ControllerHost#setDefaultBeatTimeFormatter(BeatTimeFormatter)
     * @since API version 2
     */
    getFormatted(): string;

    /**
     * Registers an observer that reports the internal beat time value as formatted text, for example
     * "012:03:00:01".
     *
     * @param separator
     *           the character used to separate the segments of the formatted beat time, typically ":", "." or
     *           "-"
     * @param barsLen
     *           the number of digits reserved for bars
     * @param beatsLen
     *           the number of digits reserved for beats
     * @param subdivisionLen
     *           the number of digits reserved for beat subdivisions
     * @param ticksLen
     *           the number of digits reserved for ticks
     * @param callback
     *           a callback function that receives a single string parameter
     * @since API version 1
     * @deprecated Use {@link #getFormatted(BeatTimeFormatter)} instead.
     */
    addTimeObserver(
      separator: string,
      barsLen: number,
      beatsLen: number,
      subdivisionLen: number,
      ticksLen: number,
      callback: StringValueChangedCallback
    ): void;
  }

  // source: com/bitwig/extension/controller/api/BitwigBrowsingSession.java

  /**
   * Instances of this interface are used for browsing Bitwig Studio document such as devices, presets,
   * multi-samples, or clips. Full access to all filter columns and the result column as shown in Bitwig
   * Studio's contextual browser window is provided.
   *
   * @see BrowsingSession
   * @since API version 1
   * @deprecated Use {@link PopupBrowser} instead.
   */
  interface BitwigBrowsingSession extends BrowsingSession {
    /**
     * Returns the creator filter as shown in the category column of Bitwig Studio's contextual browser.
     *
     * @return the requested creator filter object.
     * @since API version 1
     */
    getCreatorFilter(): BrowserFilterColumn;

    /**
     * Returns the tags filter as shown in the category column of Bitwig Studio's contextual browser.
     *
     * @return the requested tags filter object.
     * @since API version 1
     */
    getTagsFilter(): BrowserFilterColumn;
  }

  // source: com/bitwig/extension/controller/api/BooleanValue.java

  interface BooleanValue extends Value<BooleanValueChangedCallback> {
    /**
     * Gets the current value.
     *
     * @since API version 2
     */
    get(): boolean;
  }

  // source: com/bitwig/extension/controller/api/Browser.java

  /**
   * Instances of this interface represent a contextual browser in Bitwig Studio.
   *
   * @since API version 1
   * @deprecated Use {@link PopupBrowser} instead.
   */
  interface Browser extends ObjectProxy {
    /**
     * Registers an observer that reports if a browsing session was started.
     *
     * @param callback
     *           a callback function that receivers a single boolean parameter.
     * @since API version 1
     * @deprecated Use {@link #exists()} instead.
     */
    addIsBrowsingObserver(callback: BooleanValueChangedCallback): void;

    /**
     * Starts a new browser session.
     *
     * @since API version 1
     */
    startBrowsing(): void;

    /**
     * Cancels the current browser session.
     *
     * @since API version 1
     */
    cancelBrowsing(): void;

    /**
     * Finished the browser session by loading the selected item.
     *
     * @since API version 1
     */
    commitSelectedResult(): void;

    /**
     * Activates the given search session. Please note that only one search session can be active at a time.
     *
     * @param session
     *           the session that should be activated.
     * @since API version 1
     */
    activateSession(session: BrowsingSession): void;

    /**
     * Return an object allows to observe and control if the browser window should be small or full-sized.
     *
     * @return a boolean value object
     * @since API version 1
     */
    isWindowMinimized(): SettableBooleanValue;

    /**
     * Return an object allows to observe and control if the selected result should be auditioned.
     *
     * @return a boolean value object
     * @since API version 1
     */
    shouldAudition(): SettableBooleanValue;

    /**
     * Returns an object that provided bank-wise navigation of the available search sessions. Each search
     * session is dedicated to a certain material type, as shown in the tabs of Bitwig Studio's contextual
     * browser.
     *
     * @param size
     *           the size of the windows used to navigate the available browsing sessions.
     * @return the requested file column bank object
     * @since API version 1
     */
    createSessionBank(size: number): BrowsingSessionBank;

    /**
     * Returns an object that represents the selected tab as shown in Bitwig Studio's contextual browser
     * window.
     *
     * @return the requested browsing session cursor
     * @since API version 1
     */
    createCursorSession(): CursorBrowsingSession;

    /**
     * Returns an object that provides access to the contents of the device tab as shown in Bitwig Studio's
     * contextual browser window.
     *
     * @return the requested device browsing session instance
     * @since API version 1
     */
    getDeviceSession(): DeviceBrowsingSession;

    /**
     * Returns an object that provides access to the contents of the preset tab as shown in Bitwig Studio's
     * contextual browser window.
     *
     * @return the requested preset browsing session instance
     * @since API version 1
     */
    getPresetSession(): PresetBrowsingSession;

    /**
     * Returns an object that provides access to the contents of the samples tab as shown in Bitwig Studio's
     * contextual browser window.
     *
     * @return the requested sample browsing session instance
     * @since API version 1
     */
    getSampleSession(): SampleBrowsingSession;

    /**
     * Returns an object that provides access to the contents of the multi-samples tab as shown in Bitwig
     * Studio's contextual browser window.
     *
     * @return the requested multi-sample browsing session instance
     * @since API version 1
     */
    getMultiSampleSession(): MultiSampleBrowsingSession;

    /**
     * Returns an object that provides access to the contents of the clips tab as shown in Bitwig Studio's
     * contextual browser window.
     *
     * @return the requested clip browsing session instance
     * @since API version 1
     */
    getClipSession(): ClipBrowsingSession;

    /**
     * Returns an object that provides access to the contents of the music tab as shown in Bitwig Studio's
     * contextual browser window.
     *
     * @return the requested music browsing session instance
     * @since API version 1
     */
    getMusicSession(): MusicBrowsingSession;
  }

  // source: com/bitwig/extension/controller/api/BrowserColumn.java

  import IntegerValueChangedCallback = com.bitwig.extension.callback.IntegerValueChangedCallback;

  /**
   * Instances of this interface are used to navigate a column in the Bitwig Studio browser.
   *
   * @since API version 1
   */
  interface BrowserColumn extends ObjectProxy {
    /**
     * Registers an observer that reports if the column exists.
     *
     * @param callback
     *           a callback function that receives a single boolean parameter
     * @since API version 1
     * @deprecated Use {@link #exists()} instead.
     */
    addExistsObserver(callback: BooleanValueChangedCallback): void;

    /**
     * Value that reports the underlying total count of column entries (not the size of the column window).
     *
     * @since API version 2
     */
    entryCount(): IntegerValue;

    /**
     * Registers an observer that reports the underlying total count of column entries (not the size of the
     * column window).
     *
     * @param callback
     *           a callback function that receives a single integer parameter
     * @since API version 1
     * @deprecated Use {@link #entryCount()}.addValueObserver(callback)
     */
    addEntryCountObserver(callback: IntegerValueChangedCallback): void;

    /**
     * Returns the cursor item, which can be used to navigate over the list of entries.
     *
     * @return the requested filter item object
     * @since API version 1
     */
    createCursorItem(): BrowserItem;

    /**
     * Returns an object that provides access to a bank of successive entries using a window configured with
     * the given size, that can be scrolled over the list of entries.
     *
     * @param size
     *           the number of simultaneously accessible items
     * @return the requested item bank object
     */
    createItemBank(size: number): BrowserItemBank;
  }

  // source: com/bitwig/extension/controller/api/BrowserFilterColumn.java

  /**
   * Instances of this interface are used to navigate a filter column in the Bitwig Studio browser.
   *
   * @since API version 1
   */
  interface BrowserFilterColumn extends BrowserColumn {
    /**
     * Returns the filter item that represents the top-level all/any/everything wildcard item.
     *
     * @return the requested filter item object
     * @since API version 1
     */
    getWildcardItem(): BrowserFilterItem;

    /**
     * Returns the cursor filter item, which can be used to navigate over the list of entries.
     *
     * @return the requested filter item object
     * @since API version 1
     */
    createCursorItem(): BrowserFilterItem;

    /**
     * Returns an object that provides access to a bank of successive entries using a window configured with
     * the given size, that can be scrolled over the list of entries.
     *
     * @param size
     *           the number of simultaneously accessible items
     * @return the requested item bank object
     */
    createItemBank(size: number): BrowserFilterItemBank;

    /**
     * Value that reports the name of the filter column.
     *
     * @since API version2
     */
    name(): StringValue;

    /**
     * Registers an observer that reports the name of the filter column.
     *
     * @param callback
     *           a callback function that receives a single string argument.
     * @since API version 1
     * @deprecated Use {@link #name()} instead.
     */
    addNameObserver(
      maxCharacters: number,
      textWhenUnassigned: string,
      callback: StringValueChangedCallback
    ): void;
  }

  // source: com/bitwig/extension/controller/api/BrowserFilterColumnBank.java

  /**
   * Instances of this interface are used to navigate the columns of a Bitwig Studio browser session.
   *
   * @since API version 1
   * @deprecated Use {@link PopupBrowser} instead.
   */
  interface BrowserFilterColumnBank extends Bank<BrowserFilterColumn> {
    /**
     * Returns the window size that was used to configure the filter column during creation.
     *
     * @return the size of the filter column.
     */
    getSize(): number;

    /**
     * Returns the filter column for the given index.
     *
     * @param index
     *           the item index, must be in the range `[0..getSize-1]`
     * @return the requested filter column object
     */
    getItem(index: number): BrowserFilterColumn;

    /**
     * Scrolls the filter columns one item up.
     *
     * @since API version 1
     */
    scrollUp(): void;

    /**
     * Scrolls the filter columns one item down.
     *
     * @since API version 1
     */
    scrollDown(): void;

    /**
     * Scrolls the filter columns one page up. For example if the bank is configured with a window size of 8
     * entries and is currently showing items [1..8], calling this method would scroll the window to show
     * columns [9..16].
     *
     * @since API version 1
     */
    scrollPageUp(): void;

    /**
     * Scrolls the filter columns one page up. For example if the bank is configured with a window size of 8
     * entries and is currently showing items [9..16], calling this method would scroll the window to show
     * columns [1..8].
     *
     * @since API version 1
     */
    scrollPageDown(): void;

    /**
     * Registers an observer that reports the current scroll position, more specifically the position of the
     * first item within the underlying list of columns, that is shown as the first column within the window.
     *
     * @param callback
     *           a callback function that receives a single integer number parameter. The parameter reflects
     *           the scroll position, or `-1` in case the column has no content.
     * @since API version 1
     */
    addScrollPositionObserver(callback: IntegerValueChangedCallback): void;

    /**
     * Registers an observer that reports if the columns can be scrolled further up.
     *
     * @param callback
     *           a callback function that receives a single boolean parameter
     * @since API version 1
     */
    addCanScrollUpObserver(callback: BooleanValueChangedCallback): void;

    /**
     * Registers an observer that reports if the columns can be scrolled further down.
     *
     * @param callback
     *           a callback function that receives a single boolean parameter
     * @since API version 1
     */
    addCanScrollDownObserver(callback: BooleanValueChangedCallback): void;

    /**
     * Registers an observer that reports the underlying total count of columns (not the size of the window).
     *
     * @param callback
     *           a callback function that receives a single integer parameter
     * @since API version 1
     */
    addEntryCountObserver(callback: IntegerValueChangedCallback): void;
  }

  // source: com/bitwig/extension/controller/api/BrowserFilterItem.java

  /**
   * Instances of this interface represent entries in a browser filter column.
   *
   * @since API version 1
   */
  interface BrowserFilterItem extends BrowserItem {
    /**
     * Value that reports the hit count of the filter item.
     *
     * @since API version 2
     */
    hitCount(): IntegerValue;

    /**
     * Registers an observer that reports the hit count of the filter item.
     *
     * @param callback
     *           a callback function that receives a single integer parameter
     * @since API version 1
     * @deprecated Use {@link #hitCount()} instead.
     */
    addHitCountObserver(callback: IntegerValueChangedCallback): void;
  }

  // source: com/bitwig/extension/controller/api/BrowserFilterItemBank.java

  /**
   * Instances of this interface are used to navigate a filter column in the Bitwig Studio browser.
   *
   * @since API version 1
   */
  interface BrowserFilterItemBank extends BrowserItemBank<BrowserFilterItem> {}

  // source: com/bitwig/extension/controller/api/BrowserItem.java

  /**
   * Instances of this interface represent entries in a browser filter column.
   *
   * @since API version 1
   */
  interface BrowserItem extends ObjectProxy {
    /**
     * Registers an observer that reports if the item exists.
     *
     * @param callback
     *           a callback function that receives a single boolean parameter
     * @since API version 1
     * @deprecated Use {@link #exists()} instead.
     */
    addExistsObserver(callback: BooleanValueChangedCallback): void;

    /**
     * Value that reports the name of the browser item.
     *
     * @since API version 2
     */
    name(): StringValue;

    /**
     * Registers an observer that reports the string value of the browser item.
     *
     * @param maxCharacters
     * @param textWhenUnassigned
     * @param callback
     *           a callback function that receives a single string argument
     * @since API version 1
     * @deprecated Use {@link #name()} instead.
     */
    addValueObserver(
      maxCharacters: number,
      textWhenUnassigned: string,
      callback: StringValueChangedCallback
    ): void;

    /**
     * Returns an object that provides access to the selected state of the browser item.
     *
     * @return an boolean value object
     * @since API version 1
     */
    isSelected(): SettableBooleanValue;
  }

  // source: com/bitwig/extension/controller/api/BrowserItemBank.java

  /**
   * Instances of this interface are used to navigate a column in the Bitwig Studio browser.
   *
   * @since API version 1
   */
  interface BrowserItemBank<ItemType extends BrowserItem = BrowserItem>
    extends Bank<ItemType> {
    /**
     * Returns the window size that was used to configure the filter column during creation.
     *
     * @return the size of the filter column.
     * @since API version 1
     * @deprecated Use {@link #getSizeOfBank()} instead.
     */
    getSize(): number;

    /**
     * Returns the item for the given index.
     *
     * @param index
     *           the item index, must be in the range `[0..getSize-1]`
     * @return the requested item object
     * @since API version 1
     * @deprecated Use {@link #getItemAt(int)} instead.
     */
    getItem(index: number): BrowserItem;

    /**
     * Scrolls the filter column entries one item up.
     *
     * @since API version 1
     * @deprecated Use {@link #scrollBackwards()} instead.
     */
    scrollUp(): void;

    /**
     * Scrolls the filter column entries one item down.
     *
     * @since API version 1
     * @deprecated Use {@link #scrollForwards()} instead.
     */
    scrollDown(): void;

    /**
     * Scrolls the filter column entries one page up. For example if the column is configured with a window
     * size of 8 entries and is currently showing items [1..8], calling this method would scroll the column to
     * show items [9..16].
     *
     * @since API version 1
     * @deprecated Use {@link #scrollPageBackwards()} instead.
     */
    scrollPageUp(): void;

    /**
     * Scrolls the filter column entries one page up. For example if the column is configured with a window
     * size of 8 entries and is currently showing items [9..16], calling this method would scroll the column to
     * show items [1..8].
     *
     * @since API version 1
     * @deprecated Use {@link #scrollPageForwards()} instead.
     */
    scrollPageDown(): void;

    /**
     * Registers an observer that reports the current scroll position, more specifically the position of the
     * first item within the underlying list of entries, that is shown as the first entry within the window.
     *
     * @param callback
     *           a callback function that receives a single integer number parameter. The parameter reflects
     *           the scroll position, or `-1` in case the column has no content.
     * @since API version 1
     * @deprecated Use {@link #scrollPosition()} instead.
     */
    addScrollPositionObserver(callback: IntegerValueChangedCallback): void;

    /**
     * Registers an observer that reports if the column entries can be scrolled further up.
     *
     * @param callback
     *           a callback function that receives a single boolean parameter
     * @since API version 1
     * @deprecated Use {@link #canScrollBackwards()} instead.
     */
    addCanScrollUpObserver(callback: BooleanValueChangedCallback): void;

    /**
     * Registers an observer that reports if the column entries can be scrolled further down.
     *
     * @param callback
     *           a callback function that receives a single boolean parameter
     * @since API version 1
     * @deprecated Use {@link #canScrollForwards()} instead.
     */
    addCanScrollDownObserver(callback: BooleanValueChangedCallback): void;
  }

  // source: com/bitwig/extension/controller/api/BrowserResultsColumn.java

  /**
   * Instances of this interface are used to navigate a results column in the Bitwig Studio browser.
   *
   * @since API version 1
   */
  interface BrowserResultsColumn extends BrowserColumn {
    /**
     * Returns the cursor result item, which can be used to navigate over the list of entries.
     *
     * @return the requested filter item object
     * @since API version 1
     */
    createCursorItem(): BrowserResultsItem;

    /**
     * Returns an object that provides access to a bank of successive entries using a window configured with
     * the given size, that can be scrolled over the list of entries.
     *
     * @param size
     *           the number of simultaneously accessible items
     * @return the requested item bank object
     */
    createItemBank(size: number): BrowserResultsItemBank;
  }

  // source: com/bitwig/extension/controller/api/BrowserResultsItem.java

  /**
   * Instances of this interface represent entries in a browser results column.
   *
   * @since API version 1
   */
  interface BrowserResultsItem extends BrowserItem {}

  // source: com/bitwig/extension/controller/api/BrowserResultsItemBank.java

  /**
   * Instances of this interface are used to navigate the results column in the Bitwig Studio browser.
   *
   * @since API version 1
   */
  interface BrowserResultsItemBank
    extends BrowserItemBank<BrowserResultsItem> {}

  // source: com/bitwig/extension/controller/api/BrowsingSession.java

  /**
   * Instances of this interface are used for browsing material according to a certain type. Possible material
   * types are devices, presets, samples, multi-samples, clips, or files from your music collection.
   *
   * In Bitwig Studio's contextual browser window the search sessions for the various material kinds are shown
   * in tabs. Just like the tabs in the browser window, instances of this interface provide access to multiple
   * filter columns and one result column. The filter columns are used to control the content of the results
   * column.
   *
   * @since API version 1
   * @deprecated Use {@link PopupBrowser} instead.
   */
  interface BrowsingSession extends ObjectProxy {
    /**
     * Registers an observer that reports if the browser session is available for the current context.
     *
     * @param callback
     *           a callback function that receives a single boolean argument.
     * @since API version 1
     * @deprecated Use {@link #exists()} instead.
     */
    addIsAvailableObserver(callback: BooleanValueChangedCallback): void;

    /**
     * Registers an observer that reports if the browser session is currently active.
     *
     * @param callback
     *           a callback function that receives a single boolean argument.
     * @since API version 1
     */
    addIsActiveObserver(callback: BooleanValueChangedCallback): void;

    /**
     * Activates the given search session, same as calling {@link Browser#activateSession
     * Browser#activateSession(this)}. Please note that only one search session can be active at a time.
     *
     * @since API version 1
     * @see Browser#activateSession
     */
    activate(): void;

    /**
     * Returns an object that represents the column which shows the results according to the current filter
     * settings in Bitwig Studio's contextual browser.
     *
     * @return the requested results browser column.
     * @since API version 1
     */
    getResults(): BrowserResultsColumn;

    /**
     * Returns an object used for navigating the entries in the results column of Bitwig Studio's contextual
     * browser.
     *
     * @return the requested cursor object.
     * @since API version 1
     */
    getCursorResult(): CursorBrowserResultItem;

    /**
     * Returns an object that represents the currently loaded material item.
     *
     * @return the requested settled result object
     * @since API version 1
     */
    getSettledResult(): BrowserResultsItem;

    /**
     * Returns an object that can be used to navigate over the various filter sections of the browsing session.
     *
     * @return the requested filter cursor object
     */
    getCursorFilter(): CursorBrowserFilterColumn;

    /**
     * Returns an object that provided bank-wise navigation of filter columns.
     *
     * @return the requested file column bank object
     * @since API version 1
     * @param numColumns
     *           the number of columns that are simultaneously accessible.
     */
    createFilterBank(numColumns: number): BrowserFilterColumnBank;

    /**
     * Value that reports the number of results available for the current filter settings.
     *
     * @since API version 2
     */
    hitCount(): IntegerValue;

    /**
     * Registers an observer that reports the number of results available for the current filter settings.
     *
     * @param callback
     *           a callback function that receives a single integer argument.
     * @since API version 1
     * @deprecated Use {@link #hitCount()} instead.
     */
    addHitCountObserver(callback: IntegerValueChangedCallback): void;
  }

  // source: com/bitwig/extension/controller/api/BrowsingSessionBank.java

  /**
   * Instances of this interface are used to navigate the available sessions in Bitwig Studio's contextual
   * browser. The sessions are shown as tabs in the graphical user interface of the browser.
   *
   * @since API version 1
   */
  interface BrowsingSessionBank extends Bank<GenericBrowsingSession> {
    /**
     * Returns the window size that was used to configure the session bank during creation.
     *
     * @return the size of the session bank.
     * @since API version 1
     */
    getSize(): number;

    /**
     * Returns the browser session for the given index.
     *
     * @param index
     *           the session index, must be in the range `[0..getSize-1]`
     * @return the requested browser session object
     * @since API version 1
     */
    getSession(index: number): GenericBrowsingSession;

    /**
     * Scrolls the browser sessions one item up.
     *
     * @since API version 1
     * @deprecated Use {@link #scrollBackwards()} instead.
     */
    scrollUp(): void;

    /**
     * Scrolls the browser sessions one item down.
     *
     * @since API version 1
     * @deprecated Use {@link #canScrollForwards()} instead.
     */
    scrollDown(): void;

    /**
     * Scrolls the browser sessions one page up. For example if the bank is configured with a window size of 8
     * entries and is currently showing items [1..8], calling this method would scroll the window to show items
     * [9..16].
     *
     * @since API version 1
     * @deprecated Use {@link #scrollPageBackwards()} instead.
     */
    scrollPageUp(): void;

    /**
     * Scrolls the filter columns one page up. For example if the bank is configured with a window size of 8
     * entries and is currently showing items [9..16], calling this method would scroll the window to show
     * items [1..8].
     *
     * @since API version 1
     * @deprecated Use {@link #scrollPageForwards()} instead.
     */
    scrollPageDown(): void;

    /**
     * Registers an observer that reports the current scroll position, more specifically the position of the
     * first item within the underlying list of browser sessions, that is shown as the first session within the
     * window.
     *
     * @param callback
     *           a callback function that receives a single integer number parameter. The parameter reflects
     *           the scroll position, or `-1` in case the column has no content.
     * @since API version 1
     * @deprecated Use {@link #scrollPosition()} instead.
     */
    addScrollPositionObserver(callback: IntegerValueChangedCallback): void;

    /**
     * Registers an observer that reports if the browser sessions can be scrolled further up.
     *
     * @param callback
     *           a callback function that receives a single boolean parameter
     * @since API version 1
     * @deprecated Use {@link #canScrollBackwards()} instead.
     */
    addCanScrollUpObserver(callback: BooleanValueChangedCallback): void;

    /**
     * Registers an observer that reports if the browser sessions can be scrolled further down.
     *
     * @param callback
     *           a callback function that receives a single boolean parameter
     * @since API version 1
     * @deprecated Use {@link #canScrollForwards()} instead.
     */
    addCanScrollDownObserver(callback: BooleanValueChangedCallback): void;

    /**
     * Registers an observer that reports the underlying total count of browser sessions (not the size of the
     * window).
     *
     * @param callback
     *           a callback function that receives a single integer parameter
     * @since API version 1
     * @deprecated Use {@link #itemCount()} instead.
     */
    addEntryCountObserver(callback: IntegerValueChangedCallback): void;
  }

  // source: com/bitwig/extension/controller/api/ChainSelector.java

  /**
   * This interface represents a chain selector device which can be:
   * - instrument selector
   * - effect selector
   *
   * @since API version 6
   */
  interface ChainSelector extends ObjectProxy, Cursor {
    /**
     * The index of the active chain in the chain selector.
     * In case the chain selector has no chains or the value is not connected to the chain selector,
     * then the value will be 0.
     *
     * @since API version 6
     */
    activeChainIndex(): SettableIntegerValue;

    /**
     * The number of chains in the chain selector.
     *
     * @since API version 6
     */
    chainCount(): IntegerValue;

    /**
     * The active device layer.
     *
     * @since API version 6
     */
    activeChain(): DeviceLayer;

    /**
     * Cycle to the next chain.
     * If the current active chain is the last one, then moves to the first one.
     *
     * @since API version 6
     */
    cycleNext(): void;

    /**
     * Cycle to the previous chain.
     * If the current active chain the first one, then moves to the last one.
     *
     * @since API version 6
     */
    cyclePrevious(): void;
  }

  // source: com/bitwig/extension/controller/api/Channel.java

  import ColorValueChangedCallback = com.bitwig.extension.callback.ColorValueChangedCallback;
  import NotePlaybackCallback = com.bitwig.extension.callback.NotePlaybackCallback;

  /**
   * This interface defines access to the common attributes and operations of channels, such as tracks or nested
   * device channels.
   *
   * @since API version 1
   */
  interface Channel extends DeviceChain {
    /**
     * Returns an object that represents the activated state of the channel.
     *
     * @return an object that provides access to the channels activated state.
     * @since API version 1
     */
    isActivated(): SettableBooleanValue;

    /**
     * Gets a representation of the channels volume control.
     *
     * @return an object that provides access to the channels volume control.
     * @since API version 1
     * @deprecated Use {@link #volume()} instead.
     */
    getVolume(): Parameter;

    /**
     * Gets a representation of the channels volume control.
     *
     * @return an object that provides access to the channels volume control.
     * @since API version 5
     */
    volume(): Parameter;

    /**
     * Gets a representation of the channels pan control.
     *
     * @return an object that provides access to the channels pan control.
     * @since API version 1
     * @deprecated Use {@link #pan()} instead.
     */
    getPan(): Parameter;

    /**
     * Gets a representation of the channels pan control.
     *
     * @return an object that provides access to the channels pan control.
     * @since API version 5
     */
    pan(): Parameter;

    /**
     * Gets a representation of the channels mute control.
     *
     * @return an object that provides access to the channels mute control.
     * @since API version 1
     * @deprecated Use {@link #mute()} instead.
     */
    getMute(): SettableBooleanValue;

    /**
     * Gets a representation of the channels mute control.
     *
     * @return an object that provides access to the channels mute control.
     * @since API version 5
     */
    mute(): SettableBooleanValue;

    /**
     * Gets a representation of the channels solo control.
     *
     * @return an object that provides access to the channels solo control.
     * @since API version 1
     * @deprecated Use {@link #solo()} instead.
     */
    getSolo(): SoloValue;

    /**
     * Gets a representation of the channels solo control.
     *
     * @return an object that provides access to the channels solo control.
     * @since API version 1
     */
    solo(): SoloValue;

    /**
     * Registers an observer for the VU-meter of this track.
     *
     * @param range
     *           the number of steps to which the reported values should be scaled. For example a range of 128
     *           would cause the callback to be called with values between 0 and 127.
     * @param channel
     *           0 for left channel, 1 for right channel, -1 for the sum of both
     * @param peak
     *           when `true` the peak value is reported, otherwise the RMS value
     * @param callback
     *           a callback function that takes a single numeric argument. The value is in the range
     *           [0..range-1].
     * @throws com.bitwig.base.control_surface.ControlSurfaceException
     * @since API version 1
     */
    addVuMeterObserver(
      range: number,
      channel: number,
      peak: boolean,
      callback: IntegerValueChangedCallback
    ): void;

    /**
     * Registers an observer that reports notes when they are played on the channel.
     *
     * @param callback
     *           a callback function that receives three parameters: 1. on/off state (boolean), 2. key (int),
     *           and 3. velocity (float).
     * @since API version 1
     *
     * @deprecated use {@link #playingNotes()} instead.
     */
    addNoteObserver(callback: NotePlaybackCallback): void;

    /**
     * Returns an array of the playing notes.
     *
     * @since API version 2
     */
    playingNotes(): PlayingNoteArrayValue;

    /**
     * Registers an observer that receives notifications about the color of the channel. The callback gets
     * called at least once immediately after this function call to report the current color. Additional calls
     * are fired each time the color changes.
     *
     * @param callback
     *           a callback function that receives three float parameters in the range [0..1]: 1. red, 2.
     *           green, and 3. blue.
     * @since API version 1
     * @deprecated use {@link #color()} instead.
     */
    addColorObserver(callback: ColorValueChangedCallback): void;

    /**
     * Get the color of the channel.
     *
     * @since API version 2
     */
    color(): SettableColorValue;

    /**
     * Gets a {@link SendBank} that can be used to navigate the sends of this channel.
     *
     * @since API version 2
     */
    sendBank(): SendBank;

    /**
     * Gets a representation of the channels send control at the given index.
     *
     * @param index
     *           the index of the send, must be valid
     * @return an object that provides access to the requested send control.
     * @since API version 1
     * @deprecated Use {@link #sendBank()} instead.
     */
    getSend(index: number): Send;

    /**
     * Duplicates the track.
     *
     * @since API version 1
     */
    duplicate(): void;

    /**
     * Selects the device chain in the Bitwig Studio mixer, in case it is a selectable object.
     *
     * @since API version 1
     */
    selectInMixer(): void;

    /**
     * Registers an observer that reports if the device chain is selected in Bitwig Studio mixer.
     *
     * @param callback
     *           a callback function that takes a single boolean parameter.
     * @since API version 1
     */
    addIsSelectedInMixerObserver(callback: BooleanValueChangedCallback): void;

    /**
     * Tries to scroll the contents of the arrangement editor so that the channel becomes visible.
     *
     * @since API version 1
     */
    makeVisibleInArranger(): void;

    /**
     * Tries to scroll the contents of the mixer panel so that the channel becomes visible.
     *
     * @since API version 1
     */
    makeVisibleInMixer(): void;
  }

  // source: com/bitwig/extension/controller/api/ChannelBank.java

  /**
   * A channel bank provides access to a range of channels in Bitwig Studio, such as tracks or device layers.
   * Instances of channel bank are typically configured with support for a fixed number of channels and
   * represent an excerpt of a larger list of channels. Various methods are provided for scrolling to different
   * sections of the channel list. It basically acts like a window moving over the list of channels.
   *
   * @since API version 1
   */
  interface ChannelBank<ChannelType extends Channel = Channel>
    extends ObjectProxy,
      Bank<ChannelType> {
    /**
     * Returns the channel for the given index.
     *
     * @param indexInBank
     *           the channel index within this bank, not the index within the list of all Bitwig Studio
     *           channels. Must be in the range [0..sizeOfBank-1].
     * @return the channel object
     * @since API version 1
     * @deprecated Use {@link #getItemAt(int)} instead.
     */
    getChannel(indexInBank: number): Channel;

    /**
     * Sets the step size used for scrolling the channel bank.
     *
     * @param stepSize
     *           the step size used for scrolling. Default is `1`.
     * @since API version 1
     */
    setChannelScrollStepSize(stepSize: number): void;

    /**
     * Scrolls the channels one page up. For example if the channel bank is configured with a window size of 8
     * channels and is currently showing channel [1..8], calling this method would scroll the channel bank to
     * show channel [9..16].
     *
     * @since API version 1
     * @deprecated {@link #scrollPageBackwards()}
     */
    scrollChannelsPageUp(): void;

    /**
     * Scrolls the channels one page up. For example if the channel bank is configured with a window size of 8
     * channels and is currently showing channel [9..16], calling this method would scroll the channel bank to
     * show channel [1..8].
     *
     * @since API version 1
     * @deprecated {@link #scrollPageForwards()}
     */
    scrollChannelsPageDown(): void;

    /**
     * Scrolls the channel window up by the amount specified via {@link #setChannelScrollStepSize(int)} (by
     * default one channel).
     *
     * @since API version 1
     * @deprecated {@link #scrollBackwards()}
     */
    scrollChannelsUp(): void;

    /**
     * Scrolls the channel window down by the amount specified via {@link #setChannelScrollStepSize(int)} (by
     * default one channel).
     *
     * @since API version 1
     * @deprecated {@link #scrollForwards()}
     */
    scrollChannelsDown(): void;

    /**
     * Scrolls the channel bank window so that the channel at the given position becomes visible.
     *
     * @param position
     *           the index of the channel within the underlying full list of channels (not the index within the
     *           bank). The position is typically directly related to the layout of the channel list in Bitwig
     *           Studio, starting with zero in case of the first channel.
     * @since API version 1
     * @deprecated {@link #scrollPosition()}
     */
    scrollToChannel(position: number): void;

    /**
     * Value that reports the current scroll position, more specifically the position of the
     * first channel within the underlying list of channels, that is shown as channel zero within the bank.
     *
     * @since API version 2
     * @deprecated {@link #scrollPosition()}
     */
    channelScrollPosition(): IntegerValue;

    /**
     * Registers an observer that reports the current scroll position, more specifically the position of the
     * first channel within the underlying list of channels, that is shown as channel zero within the bank.
     *
     * @param callback
     *           a callback function that receives a single integer number parameter
     * @param valueWhenUnassigned
     *           a default value for the channel position that gets reported in case the channel bank is not
     *           connected to a list of channels in Bitwig Studio.
     * @since API version 1
     * @deprecated Use {@link #channelScrollPosition()} instead
     */
    addChannelScrollPositionObserver(
      callback: IntegerValueChangedCallback,
      valueWhenUnassigned: number
    ): void;

    /**
     * Value that reports if the channel bank can be scrolled further down.
     *
     * @since API version 2
     */
    canScrollChannelsUp(): BooleanValue;

    /**
     * Registers an observer that reports if the channel bank can be scrolled further up.
     *
     * @param callback
     *           a callback function that receives a single boolean parameter
     * @since API version 1
     * @deprecated Use canScrollChannelsUp().addValueObserver(callback)
     */
    addCanScrollChannelsUpObserver(callback: BooleanValueChangedCallback): void;

    /**
     * Value that reports if the channel bank can be scrolled further down.
     *
     * @since API version 2
     */
    canScrollChannelsDown(): BooleanValue;

    /**
     * Registers an observer that reports if the channel bank can be scrolled further down.
     *
     * @param callback
     *           a callback function that receives a single boolean parameter
     * @since API version 1
     * @deprecated Use {@link #canScrollChannelsDown()}.addValueObserver(callback)
     */
    addCanScrollChannelsDownObserver(
      callback: BooleanValueChangedCallback
    ): void;

    /**
     * Value that reports the underlying total channel count (not the number of channels available in the bank
     * window).
     *
     * @since API version 2
     */
    channelCount(): IntegerValue;

    /**
     * Registers an observer that reports the underlying total channel count (not the number of channels
     * available in the bank window).
     *
     * @param callback
     *           a callback function that receives a single integer parameter
     * @since API version 1
     * @deprecated Use {@link #channelCount()}.addValueObserver(callback)
     */
    addChannelCountObserver(callback: IntegerValueChangedCallback): void;

    /**
     * Scrolls the sends one page up.
     *
     * @since API version 1
     * @deprecated Does nothing.
     */
    scrollSendsPageUp(): void;

    /**
     * Scrolls the sends one page down.
     *
     * @since API version 1
     * @deprecated Does nothing.
     */
    scrollSendsPageDown(): void;

    /**
     * Scrolls the sends one step up.
     *
     * @since API version 1
     * @deprecated Does nothing.
     */
    scrollSendsUp(): void;

    /**
     * Scrolls the sends one step down.
     *
     * @since API version 1
     * @deprecated Does nothing.
     */
    scrollSendsDown(): void;

    /**
     * Scrolls to the send.
     *
     * @param position
     *           the index of the send.
     * @since API version 1
     * @deprecated Does nothing.
     */
    scrollToSend(position: number): void;

    /**
     * Registers an observer that reports if the sends window can be scrolled further up.
     *
     * @param callback
     *           a callback function that takes a single boolean parameter
     * @since API version 1
     * @deprecated Does nothing.
     */
    addCanScrollSendsUpObserver(callback: BooleanValueChangedCallback): void;

    /**
     * Registers an observer that reports if the sends window can be scrolled further down.
     *
     * @param callback
     *           a callback function that takes a single boolean parameter
     * @since API version 1
     * @deprecated Does nothing.
     */
    addCanScrollSendsDownObserver(callback: BooleanValueChangedCallback): void;

    /**
     * Registers an observer that reports the underlying total send count (not the number of sends available in
     * the bank window).
     *
     * @param callback
     *           a callback function that receives a single integer parameter
     * @since API version 1
     * @deprecated Does nothing.
     */
    addSendCountObserver(callback: IntegerValueChangedCallback): void;
  }

  // source: com/bitwig/extension/controller/api/Clip.java

  import StepDataChangedCallback = com.bitwig.extension.callback.StepDataChangedCallback;

  /**
   * An interface that provides access to the contents of a clip in Bitwig Studio.
   *
   * The note content of the clip is exposed in terms of steps and keys, mainly targeted to x-y-grid
   * applications such as step sequencers.
   *
   * @since API version 1
   */
  interface Clip extends ObjectProxy {
    /**
     * Scroll the note grid so that the given key becomes visible.
     *
     * @param key
     *           the key that should become visible
     * @since API version 1
     */
    scrollToKey(key: number): void;

    /**
     * Scrolls the note grid keys one page up. For example if the note grid is configured to show 12 keys and
     * is currently showing keys [36..47], calling this method would scroll the note grid to key range
     * [48..59].
     *
     * @since API version 1
     */
    scrollKeysPageUp(): void;

    /**
     * Scrolls the note grid keys one page down. For example if the note grid is configured to show 12 keys and
     * is currently showing keys [36..47], calling this method would scroll the note grid to key range
     * [48..59].
     *
     * @since API version 1
     */
    scrollKeysPageDown(): void;

    /**
     * Scrolls the note grid keys one key up. For example if the note grid is configured to show 12 keys and is
     * currently showing keys [36..47], calling this method would scroll the note grid to key range [37..48].
     *
     * @since API version 1
     */
    scrollKeysStepUp(): void;

    /**
     * Scrolls the note grid keys one key down. For example if the note grid is configured to show 12 keys and
     * is currently showing keys [36..47], calling this method would scroll the note grid to key range
     * [35..46].
     *
     * @since API version 1
     */
    scrollKeysStepDown(): void;

    /**
     * Scroll the note grid so that the given step becomes visible.
     *
     * @param step
     *           the step that should become visible
     * @since API version 1
     */
    scrollToStep(step: number): void;

    /**
     * Scrolls the note grid steps one page forward. For example if the note grid is configured to show 16
     * steps and is currently showing keys [0..15], calling this method would scroll the note grid to key range
     * [16..31].
     *
     * @since API version 1
     */
    scrollStepsPageForward(): void;

    /**
     * Scrolls the note grid steps one page backwards. For example if the note grid is configured to show 16
     * steps and is currently showing keys [16..31], calling this method would scroll the note grid to key
     * range [0..16].
     *
     * @since API version 1
     */
    scrollStepsPageBackwards(): void;

    /**
     * Scrolls the note grid steps one step forward. For example if the note grid is configured to show 16
     * steps and is currently showing keys [0..15], calling this method would scroll the note grid to key range
     * [1..16].
     *
     * @since API version 1
     */
    scrollStepsStepForward(): void;

    /**
     * Scrolls the note grid steps one step backwards. For example if the note grid is configured to show 16
     * steps and is currently showing keys [1..16], calling this method would scroll the note grid to key range
     * [0..15].
     *
     * @since API version 1
     */
    scrollStepsStepBackwards(): void;

    /**
     * Value that reports if the note grid keys can be scrolled further up.
     *
     * @since API version 2
     */
    canScrollKeysUp(): BooleanValue;

    /**
     * Registers an observer that reports if the note grid keys can be scrolled further up.
     *
     * @param callback
     *           a callback function that receives a single boolean parameter.
     * @since API version 1
     * @deprecated Use {@link #canScrollKeysUp()} instead.
     */
    addCanScrollKeysUpObserver(callback: BooleanValueChangedCallback): void;

    /**
     * Value that reports if the note grid keys can be scrolled further down.
     *
     * @since API version 2
     */
    canScrollKeysDown(): BooleanValue;

    /**
     * Registers an observer that reports if the note grid keys can be scrolled further down.
     *
     * @param callback
     *           a callback function that receives a single boolean parameter.
     * @since API version 1
     * @deprecated Use {@link #canScrollKeysDown()} instead.
     */
    addCanScrollKeysDownObserver(callback: BooleanValueChangedCallback): void;

    /**
     * Value that reports if the note grid if the note grid steps can be scrolled backwards.
     *
     * @since API version 2
     */
    canScrollStepsBackwards(): BooleanValue;

    /**
     * Registers an observer that reports if the note grid steps can be scrolled backwards.
     *
     * @param callback
     *           a callback function that receives a single boolean parameter.
     * @since API version 1
     * @deprecated Use {@link #canScrollStepsBackwards()} instead.
     */
    addCanScrollStepsBackwardsObserver(
      callback: BooleanValueChangedCallback
    ): void;

    /**
     * Value that reports if the note grid if the note grid steps can be scrolled forwards.
     *
     * @since API version 2
     */
    canScrollStepsForwards(): BooleanValue;

    /**
     * Registers an observer that reports if the note grid keys can be scrolled forward.
     *
     * @param callback
     *           a callback function that receives a single boolean parameter.
     * @since API version 1
     * @deprecated Use {@link #canScrollStepsForwards()} instead.
     */
    addCanScrollStepsForwardObserver(
      callback: BooleanValueChangedCallback
    ): void;

    /**
     * Toggles the existence of a note in the note grid cell specified by the given x and y arguments.
     *
     * @param x
     *           the x position within the note grid, defining the step/time of the target note
     * @param y
     *           the y position within the note grid, defining the key of the target note
     * @param insertVelocity
     *           the velocity of the target note in case a new note gets inserted
     * @since API version 1
     */
    toggleStep(x: number, y: number, insertVelocity: number): void;

    /**
     * Creates a note in the grid cell specified by the given x and y arguments. Existing notes are
     * overwritten.
     *
     * @param x
     *           the x position within the note grid, defining the step/time of the new note
     * @param y
     *           the y position within the note grid, defining the key of the new note
     * @param insertVelocity
     *           the velocity of the new note
     * @param insertDuration
     *           the duration of the new note
     * @since API version 1
     */
    setStep(
      x: number,
      y: number,
      insertVelocity: number,
      insertDuration: number
    ): void;

    /**
     * Removes the note in the grid cell specified by the given x and y arguments. Calling this method does
     * nothing in case no note exists at the given x-y-coordinates.
     *
     * @param x
     *           the x position within the note grid, defining the step/time of the target note
     * @param y
     *           the y position within the note grid, defining the key of the target note
     * @since API version 1
     */
    clearStep(x: number, y: number): void;

    /**
     * Removes all notes in the grid row specified by the given y argument.
     *
     * @param y
     *           the y position within the note grid, defining the key of the target note
     * @since API version 1
     */
    clearSteps(y: number): void;

    /**
     * Removes all notes in the grid.
     *
     * @since API version 1
     */
    clearSteps(): void;

    /**
     * Selects the note in the grid cell specified by the given x and y arguments, in case there actually is a
     * note at the given x-y-coordinates.
     *
     * @param x
     *           the x position within the note grid, defining the step/time of the target note
     * @param y
     *           the y position within the note grid, defining the key of the target note
     * @param clearCurrentSelection
     *           `true` if the existing selection should be cleared, {@false} if the note should be added to
     *           the current selection.
     * @since API version 1
     */
    selectStepContents(
      x: number,
      y: number,
      clearCurrentSelection: boolean
    ): void;

    /**
     * Sets the beat time duration that is represented by one note grid step.
     *
     * @param lengthInBeatTime
     *           the length of one note grid step in beat time.
     * @since API version 1
     */
    setStepSize(lengthInBeatTime: number): void;

    /**
     * Registers an observer that reports which note grid steps/keys contain notes.
     *
     * @param callback
     *           A callback function that receives three parameters: 1. the x (step) coordinate within the note
     *           grid (integer), 2. the y (key) coordinate within the note grid (integer), and 3. an integer
     *           value that indicates if the step is empty (`0`) or if a note continues playing (`1`) or starts
     *           playing (`2`).
     * @since API version 1
     */
    addStepDataObserver(callback: StepDataChangedCallback): void;

    /**
     * Value that reports note grid cells as they get played by the sequencer.
     *
     * @since API version 2
     */
    playingStep(): IntegerValue;

    /**
     * Registers an observer that reports note grid cells as they get played by the sequencer.
     *
     * @param callback
     *           A callback function that receives a single integer parameter, which reflects the step
     *           coordinate that is played, or -1 if no step is associated with the current playback position.
     * @since API version 1
     * @deprecated Use {@link #playingStep()} instead.
     */
    addPlayingStepObserver(callback: IntegerValueChangedCallback): void;

    /**
     * Updates the name of the clip.
     *
     * @param name
     *           the new clip name
     * @since API version 1
     */
    setName(name: string): void;

    /**
     * Returns shuffle settings of the clip.
     *
     * @return the value object that represents the clips shuffle setting.
     * @since API version 1
     */
    getShuffle(): SettableBooleanValue;

    /**
     * Returns accent setting of the clip.
     *
     * @return the ranged value object that represents the clips accent setting.
     * @since API version 1
     */
    getAccent(): SettableRangedValue;

    /**
     * Returns the start of the clip in beat time.
     *
     * @return the beat time object that represents the clips start time.
     * @since API version 1
     */
    getPlayStart(): SettableBeatTimeValue;

    /**
     * Returns the length of the clip in beat time.
     *
     * @return the beat time object that represents the duration of the clip.
     * @since API version 1
     */
    getPlayStop(): SettableBeatTimeValue;

    /**
     * Returns an object that provides access to the loop enabled state of the clip.
     *
     * @return a boolean value object.
     * @since API version 1
     */
    isLoopEnabled(): SettableBooleanValue;

    /**
     * Returns the loop start time of the clip in beat time.
     *
     * @return the beat time object that represents the clips loop start time.
     * @since API version 1
     */
    getLoopStart(): SettableBeatTimeValue;

    /**
     * Returns the loop length of the clip in beat time.
     *
     * @return the beat time object that represents the clips loop length.
     * @since API version 1
     */
    getLoopLength(): SettableBeatTimeValue;

    /**
     * Registers an observer that reports the clip color.
     *
     * @param callback
     *           a callback function that receives three parameters: 1. the red coordinate of the RBG color
     *           value, 2. the green coordinate of the RBG color value, and 3. the blue coordinate of the RBG
     *           color value
     * @since API version 1
     * @deprecated use {@link #color()} instead.
     */
    addColorObserver(callback: ColorValueChangedCallback): void;

    /**
     * Get the color of the clip.
     *
     * @since API version 2
     */
    color(): SettableColorValue;

    /**
     * Duplicates the clip.
     *
     * @since API version 1
     */
    duplicate(): void;

    /**
     * Duplicates the content of the clip.
     *
     * @since API version 1
     */
    duplicateContent(): void;

    /**
     * Transposes all notes in the clip by the given number of semitones.
     *
     * @param semitones
     *           the amount of semitones to transpose, can be a positive or negative integer value.
     * @since API version 1
     */
    transpose(semitones: number): void;

    /**
     * Quantizes the start time of all notes in the clip according to the given amount. The note lengths remain
     * the same as before.
     *
     * @param amount
     *           a factor between `0` and `1` that allows to morph between the original note start and the
     *           quantized note start.
     * @since API version 1
     */
    quantize(amount: number): void;

    /**
     * Gets the track that contains the clip.
     *
     * @return a track object that represents the track which contains the clip.
     * @since API version 1
     */
    getTrack(): Track;
  }

  // source: com/bitwig/extension/controller/api/ClipBrowsingSession.java

  /**
   * Instances of this interface are used for browsing clips, including access to all filter columns and the
   * result column as shown in the 'Clips' tab of Bitwig Studio's contextual browser window.
   *
   * @see BrowsingSession
   * @since API version 1
   * @deprecated Use {@link PopupBrowser} instead.
   */
  interface ClipBrowsingSession extends BrowsingSession {
    /**
     * Returns the file type filter as shown in the category column of the browser.
     *
     * @return the requested file type filter object.
     * @since API version 1
     */
    getFileTypeFilter(): BrowserFilterColumn;
  }

  // source: com/bitwig/extension/controller/api/ClipLauncherSlot.java

  interface ClipLauncherSlot extends ClipLauncherSlotOrScene {
    /**
     * Value that reports whether this slot is selected or not.
     *
     * @since API version 2
     */
    isSelected(): BooleanValue;

    /**
     * Value that reports whether this slot has content or not.
     *
     * @since API version 2
     */
    hasContent(): BooleanValue;

    /**
     * Value that reports whether this slot is playing or not.
     *
     * @since API version 2
     */
    isPlaying(): BooleanValue;

    /**
     * Value that reports whether this slot is queued for playback or not.
     *
     * @since API version 2
     */
    isPlaybackQueued(): BooleanValue;

    /**
     * Value that reports whether this slot is recording or not.
     *
     * @since API version 2
     */
    isRecording(): BooleanValue;

    /**
     * Value that reports whether this slot is queued for recording or not.
     *
     * @since API version 2
     */
    isRecordingQueued(): BooleanValue;

    /**
     * Value that reports whether this slot is queued for recording or not.
     *
     * @since API version 2
     */
    isStopQueued(): BooleanValue;

    /**
     * Value that reports the color of this slot.
     *
     * @since API version 2
     */
    color(): ColorValue;

    /**
     * Starts browsing for content that can be inserted in this slot in Bitwig Studio's popup browser.
     *
     * @since API version 2
     */
    browseToInsertClip(): void;
  }

  // source: com/bitwig/extension/controller/api/ClipLauncherSlotBank.java

  import ClipLauncherSlotBankPlaybackStateChangedCallback = com.bitwig.extension.callback.ClipLauncherSlotBankPlaybackStateChangedCallback;
  import IndexedBooleanValueChangedCallback = com.bitwig.extension.callback.IndexedBooleanValueChangedCallback;
  import IndexedColorValueChangedCallback = com.bitwig.extension.callback.IndexedColorValueChangedCallback;

  /**
   * Instances of this interface represent a scrollable fixed-size window that is connected to a section of the
   * clip launcher slots for a specific track.
   *
   * @since API version 1
   */
  interface ClipLauncherSlotBank
    extends ClipLauncherSlotOrSceneBank<ClipLauncherSlot> {
    /**
     * Selects the slot with the given index.
     *
     * @param slot
     *           the index of the slot within the slot window.
     * @since API version 1
     */
    select(slot: number): void;

    /**
     * Starts recording into the slot with the given index.
     *
     * @param slot
     *           the index of the slot within the slot window.
     * @since API version 1
     */
    record(slot: number): void;

    /**
     * Makes the clip content of the slot with the given index visible in the note or audio editor.
     *
     * @param slot
     *           the index of the slot within the slot window.
     * @since API version 1
     */
    showInEditor(slot: number): void;

    /**
     * Creates an new clip in the slot with the given index.
     *
     * @param slot
     *           the index of the slot within the slot window.
     * @since API version 1
     */
    createEmptyClip(slot: number, lengthInBeats: number): void;

    /**
     * Deletes the clip in the slot with the given index.
     *
     * @param slot
     *           the index of the slot within the slot window.
     * @since API version 1
     */
    deleteClip(slot: number): void;

    /**
     * Duplicates the clip in the slot with the given index.
     *
     * @param slot
     *           the index of the slot within the slot window.
     * @since API version 1
     */
    duplicateClip(slot: number): void;

    /**
     * Registers an observer that reports selection changes for the slots inside the window.
     *
     * @param callback
     *           a callback function that receives two parameters: 1. the slot index (integer), and 2. a
     *           boolean parameter indicating if the slot at that index is selected (`true`) or not (`false`)
     * @since API version 1
     */
    addIsSelectedObserver(callback: IndexedBooleanValueChangedCallback): void;

    /**
     * Registers an observer that reports which slots contain clips.
     *
     * @param callback
     *           a callback function that receives two parameters: 1. the slot index (integer), and 2. a
     *           boolean parameter indicating if the slot at that index contains a clip (`true`) or not
     *           (`false`)
     * @since API version 1
     */
    addHasContentObserver(callback: IndexedBooleanValueChangedCallback): void;

    /**
     * Registers an observer that reports the playback state of clips / slots. The reported states include
     * `stopped`, `playing`, `recording`, but also `queued for stop`, `queued for playback`, `queued for
     * recording`.
     *
     * @param callback
     *           a callback function that receives three parameters: 1. the slot index (integer), 2. the queued
     *           or playback state: `0` when stopped, `1` when playing, or `2` when recording, and 3. a boolean
     *           parameter indicating if the second argument is referring to the queued state (`true`) or the
     *           actual playback state (`false`)
     * @since API version 1
     */
    addPlaybackStateObserver(
      callback: ClipLauncherSlotBankPlaybackStateChangedCallback
    ): void;

    /**
     * Registers an observer that reports which slots have clips that are currently playing.
     *
     * @param callback
     *           a callback function that receives two parameters: 1. the slot index (integer), and 2. a
     *           boolean parameter indicating if the slot at that index has a clip that is currently playing
     *           (`true`) or not (`false`)
     * @since API version 1
     */
    addIsPlayingObserver(callback: IndexedBooleanValueChangedCallback): void;

    /**
     * Registers an observer that reports which slots have clips that are currently recording.
     *
     * @param callback
     *           a callback function that receives two parameters: 1. the slot index (integer), and 2. a
     *           boolean parameter indicating if the slot at that index has a clip that is currently recording
     *           (`true`) or not (`false`)
     * @since API version 1
     */
    addIsRecordingObserver(callback: IndexedBooleanValueChangedCallback): void;

    /**
     * Add an observer if clip playback is queued on the slot.
     *
     * @param callback
     *           a callback function that receives two parameters: 1. the slot index (integer), and 2. a
     *           boolean parameter indicating if the slot at that index has a clip that is currently queued for
     *           playback (`true`) or not (`false`)
     * @since API version 1
     */
    addIsPlaybackQueuedObserver(
      callback: IndexedBooleanValueChangedCallback
    ): void;

    /**
     * Add an observer if clip recording is queued on the slot.
     *
     * @param callback
     *           a callback function that receives two parameters: 1. the slot index (integer), and 2. a
     *           boolean parameter indicating if the slot at that index has a clip that is currently queued for
     *           recording (`true`) or not (`false`)
     * @since API version 1
     */
    addIsRecordingQueuedObserver(
      callback: IndexedBooleanValueChangedCallback
    ): void;

    /**
     * Add an observer if clip playback is queued to stop on the slot.
     *
     * @param callback
     *           a callback function that receives two parameters: 1. the slot index (integer), and 2. a
     *           boolean parameter indicating if the slot at that index has a clip that is currently queued for
     *           stop (`true`) or not (`false`)
     * @since API version 1
     */
    addIsStopQueuedObserver(callback: IndexedBooleanValueChangedCallback): void;

    /**
     * @deprecated Use {@link #addIsPlaybackQueuedObserver} instead.
     * @since API version 1
     */
    addIsQueuedObserver(callback: IndexedBooleanValueChangedCallback): void;

    /**
     * Registers an observer that reports the colors of clip in the current slot window.
     *
     * @param callback
     *           a callback function that receives four parameters: 1. the slot index (integer), 2. the red
     *           coordinate of the RBG color value, 3. the green coordinate of the RBG color value, and 4. the
     *           blue coordinate of the RBG color value
     * @since API version 1
     */
    addColorObserver(callback: IndexedColorValueChangedCallback): void;

    /**
     * Specifies if the Bitwig Studio clip launcher should indicate which slots are part of the window. By
     * default indications are disabled.
     *
     * @param shouldIndicate
     *           `true` if visual indications should be enabled, `false` otherwise
     * @since API version 1
     */
    setIndication(shouldIndicate: boolean): void;

    /**
     * Returns an object that can be used to observe and toggle if the slots on a connected track group show
     * either scenes launch buttons (for launching the content of the track group) or the clips of the group
     * master track.
     *
     * @return a boolean value object.
     */
    isMasterTrackContentShownOnTrackGroups(): SettableBooleanValue;
  }

  // source: com/bitwig/extension/controller/api/ClipLauncherSlotOrScene.java

  interface ClipLauncherSlotOrScene extends ObjectProxy {
    /**
     * Returns an object that provides access to the name of the scene.
     *
     * @return a string value object that represents the scene name.
     * @since API version 2
     */
    name(): StringValue;

    /**
     * Launches the scene.
     *
     * @since API version 1
     */
    launch(): void;

    /**
     * Value that reports the position of the scene within the list of Bitwig Studio scenes.
     *
     * @since API version 2
     */
    sceneIndex(): IntegerValue;

    /**
     * Copies the current slot or scene into the dest slot or scene.
     *
     * @since API version 4
     */
    copyFrom(source: ClipLauncherSlotOrScene): void;

    /**
     * Moves the current slot or scene into the destination slot or scene.
     *
     * @since API version 4
     */
    moveTo(dest: ClipLauncherSlotOrScene): void;
  }

  // source: com/bitwig/extension/controller/api/ClipLauncherSlotOrSceneBank.java

  import IndexedStringValueChangedCallback = com.bitwig.extension.callback.IndexedStringValueChangedCallback;

  /**
   * An abstract interface that represents the clip launcher scenes or slots of a single track.
   *
   * @since API version 1
   */
  interface ClipLauncherSlotOrSceneBank<
    ItemType extends ClipLauncherSlotOrScene = ClipLauncherSlotOrScene
  > extends Bank<ItemType> {
    /**
     * Launches the scene/slot with the given index.
     *
     * @param slot
     *           the index of the slot that should be launched
     * @since API version 1
     */
    launch(slot: number): void;

    /**
     * Stops clip launcher playback for the associated track.
     *
     * @since API version 1
     */
    stop(): void;

    /**
     * Performs a return-to-arrangement operation on the related track, which caused playback to be taken over
     * by the arrangement sequencer.
     *
     * @since API version 1
     */
    returnToArrangement(): void;

    /**
     * Registers an observer that reports the names of the scenes and slots. The slot names reflect the names
     * of containing clips.
     *
     * @param callback
     *           a callback function receiving two parameters: 1. the slot index (integer) within the
     *           configured window, and 2. the name of the scene/slot (string)
     * @since API version 1
     */
    addNameObserver(callback: IndexedStringValueChangedCallback): void;
  }

  // source: com/bitwig/extension/controller/api/ColorValue.java

  interface ColorValue extends Value<ColorValueChangedCallback> {
    /**
     * Gets the red component of the current value.
     *
     * @since API version 2
     */
    red(): number;

    /**
     * Gets the green component of the current value.
     *
     * @since API version 2
     */
    green(): number;

    /**
     * Gets the blue component of the current value.
     *
     * @since API version 2
     */
    blue(): number;

    /**
     * Gets the alpha component of the current value.
     *
     * @since API version 5
     */
    alpha(): number;
  }

  // source: com/bitwig/extension/controller/api/ControllerHost.java

  import Host = com.bitwig.extension.api.Host;
  import ConnectionEstablishedCallback = com.bitwig.extension.callback.ConnectionEstablishedCallback;
  import DataReceivedCallback = com.bitwig.extension.callback.DataReceivedCallback;

  /**
   * @mainpage Introduction
   *
   * Welcome to the Bitwig Control Surface API.<br/>
   *
   * The pages shown here include the reference documentation for the various interfaces and functions provided
   * by the API.<br/>
   *
   * The best starting point for becoming familiar with the API within these pages is the documentation of the
   * {@link Host} interface. A singleton instance of that interface is available in the scope of each script.
   * In addition it is highly recommended to also walk through the <b>Control Surface Scripting Guide</b> that is
   * available from the @em Help menu in Bitwig Studio.
   */
  interface ControllerHost extends Host {
    /**
     * Loads the supplied API version into the calling script. This is only intended to be called from a controller
     * script. It cannot be called from a Java controller extension.
     */
    loadAPI(version: number): void;

    /**
     * Determines whether the calling script should fail if it calls a deprecated method based on the API version
     * that it requested.
     */
    shouldFailOnDeprecatedUse(): boolean;

    /**
     * Sets whether the calling script should fail if it calls a deprecated method based on the API version
     * that it requested. This is only intended to be called from a controller
     * script. It cannot be called from a Java controller extension.
     */
    setShouldFailOnDeprecatedUse(value: boolean): void;

    /**
     * Loads the script defined by the supplied path. This is only intended to be called from a controller
     * script. It cannot be called from a Java controller extension.
     */
    load(path: string): void;

    /**
     * Indicates if the host platform is Windows.
     *
     * @return `true` if the host platform is Windows, `false` otherwise.
     * @since API version 1
     */
    platformIsWindows(): boolean;

    /**
     * Indicates if the host platform is Apple Mac OS X.
     *
     * @return `true` if the host platform is Mac, `false` otherwise.
     * @since API version 1
     */
    platformIsMac(): boolean;

    /**
     * Indicates if the host platform is Linux.
     *
     * @return `true` if the host platform is Linux, `false` otherwise.
     * @since API version 1
     */
    platformIsLinux(): boolean;

    /**
     * Registers a controller script with the given parameters. This function must be called once at the global
     * scope of the script.
     *
     * @param vendor
     *           the name of the hardware vendor. Must not be <code>null</code>.
     * @param name
     *           the name of the controller script as listed in the user interface of Bitwig Studio. Must not
     *           be <code>null</code>.
     * @param version
     *           the version of the controller script. Must not be <code>null</code>.
     * @param uuid
     *           a universal unique identifier (UUID) string that is used to distinguish one script from
     *           another, for example `550e8400-e29b-11d4-a716-446655440000`. Must not be <code>null</code>.
     *           For generating random UUID strings several free web tools are available.
     * @param author
     *           the name of the script author
     * @since API version 1
     */
    defineController(
      vendor: string,
      name: string,
      version: string,
      uuid: string,
      author: string
    ): void;

    /**
     * Defines the number of MIDI ports for input and output that the device uses. This method should be called
     * once in the global scope if the script is supposed to exchange MIDI messages with the device, or if the
     * script adds entries to the MIDI input/output choosers in Bitwig Studio. After calling this method the
     * individual port objects can be accessed using {@link #getMidiInPort(int index)} and
     * {@link #getMidiInPort(int index)}.
     *
     * @param numInports
     *           the number of input ports
     * @param numOutports
     *           the number of output ports
     * @since API version 1
     */
    defineMidiPorts(numInports: number, numOutports: number): void;

    /**
     * Returns the MIDI input port with the given index.
     *
     * @param index
     *           the index of the MIDI input port, must be valid.
     * @return the requested MIDI input port
     * @since API version 1
     */
    getMidiInPort(index: number): MidiIn;

    /**
     * Returns the MIDI output port with the given index.
     *
     * @param index
     *           the index of the MIDI output port, must be valid.
     * @return the requested MIDI output port
     * @since API version 1
     */
    getMidiOutPort(index: number): MidiOut;

    /**
     * Registers patterns which are used to automatically detect hardware devices that can be used with the
     * script.<br/>
     *
     * When the user clicks on the `detect` button in the Bitwig Studio controller preferences dialog, Bitwig
     * Studio searches for connected controller hardware by comparing the parameters passed into this function
     * are compared with the port names of the available MIDI drivers. Found controller scripts are
     * automatically added with their input/output ports configured.<br/>
     *
     * Calling this function is optional, but can also be called multiple times in the global script scope in
     * order to support alternative driver names.
     *
     * @param inputs
     *           the array of strings used to detect MIDI input ports, must not be `null`.
     * @param outputs
     *           the array of strings used to detect MIDI output ports, must not be `null`.
     * @since API version 1
     */
    addDeviceNameBasedDiscoveryPair(inputs: string[], outputs: string[]): void;

    /**
     * Registers the `Identity Reply Universal SysEx` message (if any) that the MIDI device sends after
     * receiving the `Identity Request Universal SysEx` message (`F0 7E 7F 06 01 F7`), as defined in the MIDI
     * standard.<br/>
     *
     * This function may be called at the global scope of the script, but is optional. Please note that this
     * function is only applicable to scripts with one MIDI input and one MIDI output. Also note that not all
     * MIDI hardware supports SysEx identity messages.
     *
     * @param reply
     *           the `Identity Reply Universal SysEx` message. Must not be <code>null</code>
     * @deprecated
     * @since API version 1
     */
    defineSysexIdentityReply(reply: string): void;

    /**
     * Creates a preferences object that can be used to insert settings into the Controller Preferences panel
     * in Bitwig Studio.
     *
     * @return an object that provides access to custom controller preferences
     * @since API version 1
     */
    getPreferences(): Preferences;

    /**
     * Creates a document state object that can be used to insert settings into the Studio I/O Panel in Bitwig
     * Studio.
     *
     * @return an object that provides access to custom document settings
     * @since API version 1
     */
    getDocumentState(): DocumentState;

    /**
     * Returns an object that is used to configure automatic notifications. Bitwig Studio supports automatic
     * visual feedback from controllers that shows up as popup notifications. For example when the selected
     * track or the current device preset was changed on the controller these notifications are shown,
     * depending on your configuration.
     *
     * @return a configuration object used to enable/disable the various automatic notifications supported by
     *         Bitwig Studio
     * @since API version 1
     */
    getNotificationSettings(): NotificationSettings;

    /**
     * Returns an object for controlling various aspects of the currently selected project.
     *
     * @since API version 1
     */
    getProject(): Project;

    /**
     * Returns an object for controlling and monitoring the elements of the `Transport` section in Bitwig
     * Studio. This function should be called once during initialization of the script if transport access is
     * desired.
     *
     * @return an object that represents the `Transport` section in Bitwig Studio.
     * @since API version 1
     */
    createTransport(): Transport;

    /**
     * Returns an object for controlling and monitoring the `Groove` section in Bitwig Studio. This function
     * should be called once during initialization of the script if groove control is desired.
     *
     * @return an object that represents the `Groove` section in Bitwig Studio.
     * @since API version 1
     */
    createGroove(): Groove;

    /**
     * Returns an object that provides access to general application functionality, including global view
     * settings, the list of open projects, and other global settings that are not related to a certain
     * document.
     *
     * @return an application object.
     * @since API version 1
     */
    createApplication(): Application;

    /**
     * Returns an object which provides access to the `Arranger` panel of Bitwig Studio. Calling this function
     * is equal to `createArranger(-1)`.
     *
     * @return an arranger object
     * @see #createArranger(int)
     * @since API version 1
     */
    createArranger(): Arranger;

    /**
     * Returns an object which provides access to the `Arranger` panel inside the specified window.
     *
     * @param window
     *           the index of the window where the arranger panel is shown, or -1 in case the first arranger
     *           panel found on any window should be taken
     * @return an arranger object
     * @since API version 1
     */
    createArranger(window: number): Arranger;

    /**
     * Returns an object which provides access to the `Mixer` panel of Bitwig Studio. Calling this function is
     * equal to `createMixer(-1, null)`.
     *
     * @return a `Mixer` object
     * @since API version 1
     */
    createMixer(): Mixer;

    /**
     * Returns an object which provides access to the `Mixer` panel that belongs to the specified panel layout.
     * Calling this function is equal to `createMixer(-1, panelLayout)`.
     *
     * @param panelLayout
     *           the name of the panel layout that contains the mixer panel, or `null` in case the selected
     *           panel layout in Bitwig Studio should be followed. Empty strings or invalid names are treated
     *           the same way as `null`. To receive the list of available panel layouts see
     *           {@link Application#addPanelLayoutObserver}.
     * @return a `Mixer` object
     * @since API version 1
     */
    createMixer(panelLayout: string): Mixer;

    /**
     * Returns an object which provides access to the `Mixer` panel inside the specified window. Calling this
     * function is equal to `createMixer(window, null)`.
     *
     * @param window
     *           the index of the window where the mixer panel is shown, or -1 in case the first mixer panel
     *           found on any window should be taken
     * @return a `Mixer` object
     * @since API version 1
     */
    createMixer(window: number): Mixer;

    /**
     * Returns an object which provides access to the `Mixer` panel that matches the specified parameters.
     *
     * @param panelLayout
     *           the name of the panel layout that contains the mixer panel, or `null` in case the selected
     *           panel layout in Bitwig Studio should be followed. Empty strings or invalid names are treated
     *           the same way as `null`. To receive the list of available panel layouts see
     *           {@link Application#addPanelLayoutObserver}.
     * @param window
     *           the index of the window where the mixer panel is shown, or -1 in case the first mixer panel
     *           found on any window should be taken
     * @return a `Mixer` object
     * @since API version 1
     */
    createMixer(panelLayout: string, window: number): Mixer;

    /**
     * Returns a track bank with the given number of tracks, sends and scenes.<br/>
     *
     * A track bank can be seen as a fixed-size window onto the list of tracks in the current document
     * including their sends and scenes, that can be scrolled in order to access different parts of the track
     * list. For example a track bank configured for 8 tracks can show track 1-8, 2-9, 3-10 and so on.<br/>
     *
     * The idea behind the `bank pattern` is that hardware typically is equipped with a fixed amount of channel
     * strips or controls, for example consider a mixing console with 8 channels, but Bitwig Studio documents
     * contain a dynamic list of tracks, most likely more tracks than the hardware can control simultaneously.
     * The track bank returned by this function provides a convenient interface for controlling which tracks
     * are currently shown on the hardware.<br/>
     *
     * Creating a track bank using this method will consider all tracks in the document, including effect
     * tracks and the master track. Use {@link #createMainTrackBank} or {@link #createEffectTrackBank} in case
     * you are only interested in tracks of a certain kind.
     *
     * @param numTracks
     *           the number of tracks spanned by the track bank
     * @param numSends
     *           the number of sends spanned by the track bank
     * @param numScenes
     *           the number of scenes spanned by the track bank
     * @return an object for bank-wise navigation of tracks, sends and scenes
     * @see Track#createTrackBank
     * @see #createMainTrackBank
     * @see #createEffectTrackBank
     * @since API version 1
     */
    createTrackBank(
      numTracks: number,
      numSends: number,
      numScenes: number
    ): TrackBank;

    /**
     * Returns a track bank with the given number of child tracks, sends and scenes.<br/>
     *
     * A track bank can be seen as a fixed-size window onto the list of tracks in the connected track group
     * including their sends and scenes, that can be scrolled in order to access different parts of the track
     * list. For example a track bank configured for 8 tracks can show track 1-8, 2-9, 3-10 and so on.<br/>
     *
     * The idea behind the `bank pattern` is that hardware typically is equipped with a fixed amount of channel
     * strips or controls, for example consider a mixing console with 8 channels, but Bitwig Studio documents
     * contain a dynamic list of tracks, most likely more tracks than the hardware can control simultaneously.
     * The track bank returned by this function provides a convenient interface for controlling which tracks
     * are currently shown on the hardware.<br/>
     *
     * Creating a track bank using this method will consider all tracks in the document, including effect
     * tracks and the master track. Use {@link #createMainTrackBank} or {@link #createEffectTrackBank} in case
     * you are only interested in tracks of a certain kind.
     *
     * @param numTracks
     *           the number of child tracks spanned by the track bank
     * @param numSends
     *           the number of sends spanned by the track bank
     * @param numScenes
     *           the number of scenes spanned by the track bank
     * @param hasFlatTrackList
     *           specifies whether the track bank should operate on a flat list of all nested child tracks or
     *           only on the direct child tracks of the connected group track.
     * @return an object for bank-wise navigation of tracks, sends and scenes
     * @see #createMainTrackBank
     * @see #createEffectTrackBank
     * @since API version 1
     */
    createTrackBank(
      numTracks: number,
      numSends: number,
      numScenes: number,
      hasFlatTrackList: boolean
    ): TrackBank;

    /**
     * Returns a track bank with the given number of tracks, sends and scenes. Only audio tracks, instrument
     * tracks and hybrid tracks are considered. For more information about track banks and the `bank pattern`
     * in general, see the documentation for {@link #createTrackBank}.
     *
     * @param numTracks
     *           the number of tracks spanned by the track bank
     * @param numSends
     *           the number of sends spanned by the track bank
     * @param numScenes
     *           the number of scenes spanned by the track bank
     * @return an object for bank-wise navigation of tracks, sends and scenes
     * @see Track#createMainTrackBank
     * @see #createTrackBank
     * @see #createEffectTrackBank
     * @since API version 1
     */
    createMainTrackBank(
      numTracks: number,
      numSends: number,
      numScenes: number
    ): TrackBank;

    /**
     * Returns a track bank with the given number of effect tracks and scenes. Only effect tracks are
     * considered. For more information about track banks and the `bank pattern` in general, see the
     * documentation for {@link #createTrackBank}.
     *
     * @param numTracks
     *           the number of tracks spanned by the track bank
     * @param numScenes
     *           the number of scenes spanned by the track bank
     * @return an object for bank-wise navigation of tracks, sends and scenes
     * @see Track#createEffectTrackBank
     * @see #createTrackBank
     * @see #createMainTrackBank
     * @since API version 1
     */
    createEffectTrackBank(numTracks: number, numScenes: number): TrackBank;

    /**
     * Returns an object that represents the master track of the document.
     *
     * @param numScenes
     *           the number of scenes for bank-wise navigation of the master tracks clip launcher slots.
     * @return an object representing the master track.
     * @see Track#createMasterTrack
     * @since API version 1
     */
    createMasterTrack(numScenes: number): MasterTrack;

    /**
     * Returns an object that represents the cursor item of the arranger track selection.
     *
     * @param numSends
     *           the number of sends for bank-wise navigation of the sends that are associated with the track
     *           selection
     * @param numScenes
     *           the number of scenes for bank-wise navigation of the clip launcher slots that are associated
     *           with the track selection
     * @return an object representing the currently selected arranger track (in the future also multiple
     *         tracks)
     * @since API version 1
     */
    createArrangerCursorTrack(numSends: number, numScenes: number): CursorTrack;

    /**
     * Returns an object that represents a named cursor track, that is independent from the arranger or mixer
     * track selection in the user interface of Bitwig Studio.
     *
     * @param name
     *           the name of the track cursor
     * @param numSends
     *           the number of sends for bank-wise navigation of the sends that are associated with the track
     *           selection
     * @param numScenes
     *           the number of scenes for bank-wise navigation of the clip launcher slots that are associated
     *           with the track selection
     * @return an object representing the currently selected arranger track (in the future also multiple
     *         tracks).
     * @since API version 1
     * @deprecated use {@link #createCursorTrack(String, String, int, int, boolean)} instead.
     */
    createCursorTrack(
      name: string,
      numSends: number,
      numScenes: number
    ): CursorTrack;

    /**
     * Returns an object that represents a named cursor track, that is independent from the arranger or mixer
     * track selection in the user interface of Bitwig Studio.
     *
     * @param name
     *           the name of the track cursor
     * @param numSends
     *           the number of sends for bank-wise navigation of the sends that are associated with the track
     *           selection
     * @param numScenes
     *           the number of scenes for bank-wise navigation of the clip launcher slots that are associated
     *           with the track selection
     * @return an object representing the currently selected arranger track (in the future also multiple
     *         tracks).
     * @since API version 1
     */
    createCursorTrack(
      id: string,
      name: string,
      numSends: number,
      numScenes: number,
      shouldFollowSelection: boolean
    ): CursorTrack;

    /**
     * Returns a scene bank with the given number of scenes.<br/>
     *
     * A scene bank can be seen as a fixed-size window onto the list of scenes in the current document, that
     * can be scrolled in order to access different parts of the scene list. For example a scene bank
     * configured for 8 scenes can show scene 1-8, 2-9, 3-10 and so on.<br/>
     *
     * The idea behind the `bank pattern` is that hardware typically is equipped with a fixed amount of channel
     * strips or controls, for example consider a mixing console with 8 channels, but Bitwig Studio documents
     * contain a dynamic list of scenes, most likely more scenes than the hardware can control simultaneously.
     * The scene bank returned by this function provides a convenient interface for controlling which scenes
     * are currently shown on the hardware.<br/>
     *
     * @param numScenes
     *           the number of scenes spanned by the track bank
     * @return an object for bank-wise navigation of scenes
     * @since API version 1
     */
    createSceneBank(numScenes: number): SceneBank;

    /**
     * Returns an object that represents the cursor device in devices selections made by the user in Bitwig
     * Studio. Calling this method is equal to the following code: {@code
     * var cursorTrack = createArrangerCursorTrack(numSends, numScenes);
     * var cursorDevice = cursorTrack.createCursorDevice();
     * } To create a custom device selection that is not connected to the main device selection in the user
     * interface, call {@link Track#createCursorDevice(String) cursorTrack.createCursorDevice(String name)}.
     *
     * @return an object representing the currently selected device.
     * @see Track#createCursorDevice()
     * @see Track#createCursorDevice(String)
     * @see #createEditorCursorDevice(int numSends)
     * @since API version 1
     * @deprecated
     */
    createEditorCursorDevice(): CursorDevice;

    /**
     * Returns an object that represents the cursor device in devices selections made by the user in Bitwig
     * Studio. Calling this method is equal to the following code: {@code
     * var cursorTrack = createArrangerCursorTrack(numSends, numScenes);
     * var cursorDevice = cursorTrack.createCursorDevice();
     * } To create a custom device selection that is not connected to the main device selection in the user
     * interface, call {@link Track#createCursorDevice(String) cursorTrack.createCursorDevice(String name)}.
     *
     * @param numSends
     *           the number of sends that are simultaneously accessible in nested channels.
     * @return an object representing the currently selected device.
     * @deprecated Use cursorTrack.createCursorDevice().
     * @see Track#createCursorDevice()
     * @see Track#createCursorDevice(String)
     * @since API version 1
     */
    createEditorCursorDevice(numSends: number): CursorDevice;

    /**
     * @deprecated Use {@link #createLauncherCursorClip(int, int)} or
     *             {@link #createArrangerCursorClip(int, int)} instead.
     * @since API version 1
     */
    createCursorClip(gridWidth: number, gridHeight: number): Clip;

    /**
     * Returns a clip object that represents the cursor of the launcher clip selection. The gridWidth and
     * gridHeight parameters specify the grid dimensions used to access the note content of the clip.
     *
     * @param gridWidth
     *           the number of steps spanned by one page of the note content grid.
     * @param gridHeight
     *           the number of keys spanned by one page of the note content grid.
     * @return an object representing the currently selected cursor clip
     * @since API version 1
     */
    createLauncherCursorClip(gridWidth: number, gridHeight: number): Clip;

    /**
     * Returns a clip object that represents the cursor of the arranger clip selection. The gridWidth and
     * gridHeight parameters specify the grid dimensions used to access the note content of the clip.
     *
     * @param gridWidth
     *           the number of steps spanned by one page of the note content grid.
     * @param gridHeight
     *           the number of keys spanned by one page of the note content grid.
     * @return an object representing the currently selected cursor clip
     * @since API version 1
     */
    createArrangerCursorClip(gridWidth: number, gridHeight: number): Clip;

    /**
     * Returns an object that is used to define a bank of custom user controls. These controls are available to
     * the user for free controller assignments and are typically used when bank-wise navigation is
     * inconvenient.
     *
     * @param numControllers
     *           the number of controls that are available for free assignments
     * @return An object that represents a set of custom user controls.
     * @since API version 1
     */
    createUserControls(numControllers: number): UserControlBank;

    /**
     * Schedules the given callback function for execution after the given delay. For timer applications call
     * this method once initially and then from within the callback function.
     *
     * @param callback
     *           the callback function that will be called
     * @param args
     *           that array of arguments that gets passed into the callback function, may be `null`
     * @param delay
     *           the duration after which the callback function will be called in milliseconds
     * @deprecated
     * @since API version 1
     */
    scheduleTask(callback: () => void, args: object[], delay: number): void;

    /**
     * Schedules the given callback function for execution after the given delay. For timer applications call
     * this method once initially and then from within the callback function.
     *
     * @param callback
     *           the callback function that will be called
     * @param args
     *           that array of arguments that gets passed into the callback function, may be `null`
     * @param delay
     *           the duration after which the callback function will be called in milliseconds
     * @since API version 2
     */
    scheduleTask(callback: () => void, delay: number): void;

    /**
     * Requests that the driver's flush method gets called.
     *
     * @since API version 2
     */
    requestFlush(): void;

    /**
     * Prints the given string in the control surface console window. The console window can be opened in the
     * view menu of Bitwig Studio.
     *
     * @param s
     *           the string to be printed
     * @since API version 1
     */
    println(s: string): void;

    /**
     * Prints the given string in the control surface console window using a text style that highlights the
     * string as error. The console window can be opened in the view menu of Bitwig Studio.
     *
     * @param s
     *           the error string to be printed
     * @since API version 1
     */
    errorln(s: string): void;

    /**
     * Shows a temporary text overlay on top of the application GUI, that will fade-out after a short interval.
     * If the overlay is already shown, it will get updated with the given text.
     *
     * @param text
     *           the text to be shown
     * @since API version 1
     */
    showPopupNotification(text: string): void;

    /**
     * Opens a TCP (Transmission Control Protocol) host socket for allowing network connections from other
     * hardware and software.
     *
     * @param name
     *           a meaningful name that describes the purpose of this connection.
     * @param defaultPort
     *           the port that should be used for the connection. If the port is already in use, then another
     *           port will be used. Check {@link RemoteSocket#getPort()} on the returned object to be sure.
     * @return the object that represents the socket
     * @since API version 1
     */
    createRemoteConnection(name: string, defaultPort: number): RemoteSocket;

    /**
     * Connects to a remote TCP (Transmission Control Protocol) socket.
     *
     * @param host
     *           the host name or IP address to connect to.
     * @param port
     *           the port to connect to
     * @param callback
     *           the callback function that gets called when the connection gets established. A single
     *           {@link RemoteConnection} parameter is passed into the callback function.
     * @since API version 1
     */
    connectToRemoteHost(
      host: string,
      port: number,
      callback: ConnectionEstablishedCallback
    ): void;

    /**
     * Sends a UDP (User Datagram Protocol) packet with the given data to the specified host.
     *
     * @param host
     *           the destination host name or IP address
     * @param port
     *           the destination port
     * @param data
     *           the data to be send. When creating a numeric byte array in JavaScript, the byte values must be
     *           signed (in the range -128..127).
     * @since API version 1
     */
    sendDatagramPacket(host: string, port: number, data: number): void;

    /**
     * Adds an observer for incoming UDP (User Datagram Protocol) packets on the selected port.
     *
     * @param name
     *           a meaningful name that describes the purpose of this observer.
     * @param port
     *           the port that should be used
     * @param callback
     *           the callback function that gets called when data arrives. The function receives a single
     *           parameter that contains the data byte array.
     *
     * @return {@true} if was possible to bind the port, false otherwise
     * @since API version 1
     */
    addDatagramPacketObserver(
      name: string,
      port: number,
      callback: DataReceivedCallback
    ): boolean;

    /**
     * @deprecated Use {@link #defineController(String, String, String, String, String)} instead.
     * @since API version 1
     */
    defineController(
      vendor: string,
      name: string,
      version: string,
      uuid: string
    ): void;

    /**
     * @deprecated Use {@link #createTransport} instead.
     * @since API version 1
     */
    createTransportSection(): Transport;

    /**
     * @since API version 1
     */
    createCursorTrack(numSends: number, numScenes: number): CursorTrack;

    /**
     * @deprecated Use {@link #createGroove()} instead.
     * @since API version 1
     */
    createGrooveSection(): Groove;

    /**
     * @deprecated Use {@link #createApplication()} instead.
     * @since API version 1
     */
    createApplicationSection(): Application;

    /**
     * @deprecated Use {@link #createArranger(int)} instead.
     * @since API version 1
     */
    createArrangerSection(screenIndex: number): Arranger;

    /**
     * @deprecated Use {@link #createMixer(String, int)} instead.
     * @since API version 1
     */
    createMixerSection(perspective: string, screenIndex: number): Mixer;

    /**
     * @deprecated Use {@link #createTrackBank(int, int, int)} instead.
     * @since API version 1
     */
    createTrackBankSection(
      numTracks: number,
      numSends: number,
      numScenes: number
    ): TrackBank;

    /**
     * @deprecated Use {@link #createMainTrackBank(int, int, int)} instead.
     * @since API version 1
     */
    createMainTrackBankSection(
      numTracks: number,
      numSends: number,
      numScenes: number
    ): TrackBank;

    /**
     * @deprecated Use {@link #createEffectTrackBank(int, int)} instead.
     * @since API version 1
     */
    createEffectTrackBankSection(
      numTracks: number,
      numScenes: number
    ): TrackBank;

    /**
     * @deprecated Use {@link #createArrangerCursorTrack} instead.
     * @since API version 1
     */
    createCursorTrackSection(numSends: number, numScenes: number): CursorTrack;

    /**
     * @deprecated Use {@link #createMasterTrack(int)} instead.
     * @since API version 1
     */
    createMasterTrackSection(numScenes: number): Track;

    /**
     * @deprecated Use {@link #createCursorClip(int, int)} instead.
     * @since API version 1
     */
    createCursorClipSection(gridWidth: number, gridHeight: number): Clip;

    /**
     * @deprecated Use {@link #createEditorCursorDevice createEditorCursorDevice()} instead.
     * @since API version 1
     */
    createCursorDeviceSection(numControllers: number): CursorDevice;

    /**
     * @deprecated Use {@link #createEditorCursorDevice createEditorCursorDevice()} instead.
     * @since API version 1
     */
    createCursorDevice(): CursorDevice;

    /**
     * @deprecated Use {@link #createUserControls(int)} instead.
     * @since API version 1
     */
    createUserControlsSection(numControllers: number): UserControlBank;

    /**
     * @deprecated Use {@link #defineSysexIdentityReply(String)} instead.
     * @since API version 1
     */
    defineSysexDiscovery(request: string, reply: string): void;

    /**
     * Creates a {@link PopupBrowser} that represents the pop-up browser in Bitwig Studio.
     *
     * @since API version 2
     */
    createPopupBrowser(): PopupBrowser;

    /**
     * {@link BeatTimeFormatter} used to format beat times by default. This will be used to format beat times
     * when asking for a beat time in string format without providing any formatting options. For example by
     * calling {@link BeatTimeStringValue#get()}.
     *
     * @since API version 2
     */
    defaultBeatTimeFormatter(): BeatTimeFormatter;

    /**
     * Sets the {@link BeatTimeFormatter} to use by default for formatting beat times.
     *
     * @see #defaultBeatTimeFormatter()
     * @since API version 2
     */
    setDefaultBeatTimeFormatter(formatter: BeatTimeFormatter): void;

    /**
     * Creates a {@link BeatTimeFormatter} that can be used to format beat times.
     *
     * @param separator
     *           the character used to separate the segments of the formatted beat time, typically ":", "." or
     *           "-"
     * @param barsLen
     *           the number of digits reserved for bars
     * @param beatsLen
     *           the number of digits reserved for beats
     * @param subdivisionLen
     *           the number of digits reserved for beat subdivisions
     * @param ticksLen
     *           the number of digits reserved for ticks
     *
     * @since API version 2
     */
    createBeatTimeFormatter(
      separator: string,
      barsLen: number,
      beatsLen: number,
      subdivisionLen: number,
      ticksLen: number
    ): BeatTimeFormatter;
  }

  // source: com/bitwig/extension/controller/api/CueMarker.java

  /**
   * This interface defines access to the common attributes and operations of cue markers.
   *
   * @since API version 2
   */
  interface CueMarker extends ObjectProxy {
    /**
     * Launches playback at the marker position.
     *
     * @param quantized Specified if the cue marker should be launched quantized or immediately
     * @since API version 2
     */
    launch(quantized: boolean): void;

    /**
     * Gets a representation of the marker name.
     *
     * @since API version 2
     */
    getName(): StringValue;

    /**
     * Gets a representation of the marker color.
     *
     * @since API version 2
     */
    getColor(): ColorValue;
  }

  // source: com/bitwig/extension/controller/api/CueMarkerBank.java

  /**
   * A cue marker bank provides access to a range of cue markers in Bitwig Studio.
   * Instances are typically configured with a fixed number of markers and represent an excerpt
   * of a larger list of markers. It basically acts like a window moving over the list of markers.
   *
   * @since API version 2
   */
  interface CueMarkerBank extends Bank<CueMarker> {
    /**
     * Scrolls the cue marker bank window so that the marker at the given position becomes visible.
     *
     * @param position
     *           the index of the marker within the underlying full list of markers (not the index within the
     *           bank). The position is typically directly related to the layout of the marker list in Bitwig
     *           Studio, starting with zero in case of the first marker.
     * @since API version 2
     */
    scrollToMarker(position: number): void;
  }

  // source: com/bitwig/extension/controller/api/Cursor.java

  /**
   * A generic interface that provides the foundation for working with selections.
   *
   * Implementations of this interface can either represent custom selection cursors that are created by
   * controller scripts, or represent the cursor of user selections as shown in Bitwig Studio editors, such as
   * the Arranger track selection cursor, the note editor event selection cursor and so on.
   *
   * @since API version 1
   */
  interface Cursor {
    /**
     * Select the previous item.
     *
     * @since API version 1
     */
    selectPrevious(): void;

    /**
     * Select the next item.
     *
     * @since API version 1
     */
    selectNext(): void;

    /**
     * Select the first item.
     *
     * @since API version 1
     */
    selectFirst(): void;

    /**
     * Select the last item.
     *
     * @since API version 1
     */
    selectLast(): void;

    /**
     * Boolean value that reports whether there is an item after the current cursor position.
     *
     * @since API version 2
     */
    hasNext(): BooleanValue;

    /**
     * Boolean value that reports whether there is an item before the current cursor position.
     *
     * @since API version 2
     */
    hasPrevious(): BooleanValue;

    /**
     * Registers a function with bool argument that gets called when the previous item gains or remains
     * selectable.
     *
     * @since API version 1
     * @deprecated Use hasNext() instead.
     */
    addCanSelectPreviousObserver(callback: BooleanValueChangedCallback): void;

    /**
     * Registers a function with bool argument that gets called when the next item gains or remains
     * selectable.
     *
     * @since API version 1
     */
    addCanSelectNextObserver(callback: BooleanValueChangedCallback): void;
  }

  // source: com/bitwig/extension/controller/api/CursorBrowserFilterColumn.java

  /**
   * Instances of this interface are used to navigate the filter columns of a Bitwig Studio browsing session.
   *
   * @since API version 1
   */
  interface CursorBrowserFilterColumn extends BrowserFilterColumn, Cursor {}

  // source: com/bitwig/extension/controller/api/CursorBrowserFilterItem.java

  /**
   * Instances of this interface represent entries in a browser filter column.
   *
   * @since API version 1
   */
  interface CursorBrowserFilterItem
    extends BrowserFilterItem,
      CursorBrowserItem {
    /**
     * Select the parent item.
     *
     * @since API version 1
     */
    selectParent(): void;

    /**
     * Select the first child item.
     *
     * @since API version 1
     */
    selectFirstChild(): void;

    /**
     * Select the last child item.
     *
     * @since API version 1
     */
    selectLastChild(): void;

    /**
     * Select the previous item.
     *
     * @since API version 1
     */
    moveToPrevious(): void;

    /**
     * Select the next item.
     *
     * @since API version 1
     */
    moveToNext(): void;

    /**
     * Select the first item.
     *
     * @since API version 1
     */
    moveToFirst(): void;

    /**
     * Select the last item.
     *
     * @since API version 1
     */
    moveToLast(): void;

    /**
     * Select the parent item.
     *
     * @since API version 1
     */
    moveToParent(): void;

    /**
     * Move the cursor to the first child item.
     *
     * @since API version 1
     */
    moveToFirstChild(): void;

    /**
     * Move the cursor to the last child item.
     *
     * @since API version 1
     */
    moveToLastChild(): void;
  }

  // source: com/bitwig/extension/controller/api/CursorBrowserItem.java

  /**
   * Instances of this interface represent entries in a browser filter column.
   *
   * @since API version 1
   */
  interface CursorBrowserItem extends BrowserItem, Cursor {
    /**
     * Returns a bank object that provides access to the siblings of the cursor item. The bank will
     * automatically scroll so that the cursor item is always visible.
     *
     * @param numSiblings
     *           the number of simultaneously accessible siblings
     * @return the requested item bank object
     */
    createSiblingsBank(numSiblings: number): BrowserItemBank;
  }

  // source: com/bitwig/extension/controller/api/CursorBrowserResultItem.java

  /**
   * Instances of this interface represent entries in a browser column.
   *
   * @since API version 1
   */
  interface CursorBrowserResultItem
    extends BrowserResultsItem,
      CursorBrowserItem {}

  // source: com/bitwig/extension/controller/api/CursorBrowsingSession.java

  /**
   * Instances of this interface are used for navigating the various browsing sessions of Bitwig Studio's
   * contextual browser.
   *
   * @since API version 1
   */
  interface CursorBrowsingSession extends GenericBrowsingSession, Cursor {}

  // source: com/bitwig/extension/controller/api/CursorChannel.java

  /**
   * A special kind of channel that follows a channel selection cursor in Bitwig Studio. The selection can
   * either be a custom selection cursor that gets created by the controller script, or represent the user
   * selection cursor as shown in the Bitwig Studio editors, such as the Arranger track selection cursor.
   *
   * @since API version 1
   */
  interface CursorChannel extends Cursor, Channel {
    /**
     * Points the cursor to the given channel.
     *
     * @param channel
     *           the channel that this channel cursor should point to
     * @since API version 1
     */
    selectChannel(channel: Channel): void;
  }

  // source: com/bitwig/extension/controller/api/CursorDevice.java

  /**
   * A special kind of selection cursor used for devices.
   *
   * @since API version 1
   */
  interface CursorDevice extends Cursor, Device {
    /**
     * Returns the channel that this cursor device was created on. Currently this will always be a track or
     * cursor track instance.
     *
     * @return the track or cursor track object that was used for creation of this cursor device.
     * @since API version 1
     * @deprecated Use {@link #channel()} instead.
     */
    getChannel(): Channel;

    /**
     * Returns the channel that this cursor device was created on. Currently this will always be a track or
     * cursor track instance.
     *
     * @return the track or cursor track object that was used for creation of this cursor device.
     * @since API version 5
     */
    channel(): Channel;

    /**
     * Selects the parent device if there is any.
     *
     * @since API version 1
     */
    selectParent(): void;

    /**
     * Moves this cursor to the given device.
     *
     * @param device
     *           the device that this cursor should point to
     * @since API version 1
     */
    selectDevice(device: Device): void;

    /**
     * Selects the first device in the given channel.
     *
     * @param channel
     *           the channel in which the device should be selected
     * @since API version 1
     */
    selectFirstInChannel(channel: Channel): void;

    /**
     * Selects the last device in the given channel.
     *
     * @param channel
     *           the channel in which the device should be selected
     * @since API version 1
     */
    selectLastInChannel(channel: Channel): void;

    /**
     * Selects the first device in the nested FX slot with the given name.
     *
     * @param chain
     *           the name of the FX slot in which the device should be selected
     * @since API version 1
     */
    selectFirstInSlot(chain: string): void;

    /**
     * Selects the last device in the nested FX slot with the given name.
     *
     * @param chain
     *           the name of the FX slot in which the device should be selected
     * @since API version 1
     */
    selectLastInSlot(chain: string): void;

    /**
     * Selects the first device in the drum pad associated with the given key.
     *
     * @param key
     *           the key associated with the drum pad in which the device should be selected
     * @since API version 1
     */
    selectFirstInKeyPad(key: number): void;

    /**
     * Selects the last device in the drum pad associated with the given key.
     *
     * @param key
     *           the key associated with the drum pad in which the device should be selected
     * @since API version 1
     */
    selectLastInKeyPad(key: number): void;

    /**
     * Selects the first device in the nested layer with the given index.
     *
     * @param index
     *           the index of the nested layer in which the device should be selected
     * @since API version 1
     */
    selectFirstInLayer(index: number): void;

    /**
     * Selects the last device in the nested layer with the given index.
     *
     * @param index
     *           the index of the nested layer in which the device should be selected
     * @since API version 1
     */
    selectLastInLayer(index: number): void;

    /**
     * Selects the first device in the nested layer with the given name.
     *
     * @param name
     *           the name of the nested layer in which the device should be selected
     * @since API version 1
     */
    selectFirstInLayer(name: string): void;

    /**
     * Selects the last device in the nested layer with the given name.
     *
     * @param name
     *           the name of the nested layer in which the device should be selected
     * @since API version 1
     */
    selectLastInLayer(name: string): void;
  }

  // source: com/bitwig/extension/controller/api/CursorDeviceFollowMode.java

  enum CursorDeviceFollowMode {
    FOLLOW_SELECTION = 0,
    FIRST_DEVICE = 1,
    FIRST_INSTRUMENT = 2,
    FIRST_AUDIO_EFFECT = 3,
    FIRST_INSTRUMENT_OR_DEVICE = 4,
  }

  // source: com/bitwig/extension/controller/api/CursorDeviceLayer.java

  /**
   * Instances of this interface represent the cursor item in device layer selections.
   *
   * @since API version 1
   */
  interface CursorDeviceLayer extends CursorChannel, DeviceLayer {}

  // source: com/bitwig/extension/controller/api/CursorDeviceSlot.java

  /**
   * Instances of this interface represent the selected device slot as shown in the Bitwig Studio user
   * interface.
   *
   * @since API version 1
   */
  interface CursorDeviceSlot extends DeviceChain {
    selectSlot(slot: string): void;
  }

  // source: com/bitwig/extension/controller/api/CursorNavigationMode.java

  enum CursorNavigationMode {
    NESTED = 0,
    FLAT = 1,
    GUI = 2,
  }

  // source: com/bitwig/extension/controller/api/CursorRemoteControlsPage.java

  /**
   * Represents a cursor that looks at a {@link RemoteControlsPage}.
   *
   * @since API version 2
   */
  interface CursorRemoteControlsPage extends Cursor, RemoteControlsPage {
    /**
     * Value that reports the names of the devices parameter pages.
     */
    pageNames(): StringArrayValue;

    /**
     * Selects the next page.
     *
     * @param shouldCycle
     *           If true then when the end is reached and there is no next page it selects the first page
     *
     * @since API version 2
     */
    selectNextPage(shouldCycle: boolean): void;

    /**
     * Selects the previous page.
     *
     * @param shouldCycle
     *           If true then when the end is reached and there is no next page it selects the first page
     *
     * @since API version 2
     */
    selectPreviousPage(shouldCycle: boolean): void;

    /**
     * Selects the next page that matches the given expression.
     *
     * @param expression
     *           An expression that can match a page based on how it has been tagged. For now this can only be
     *           the name of a single tag that you would like to match.
     *
     * @param shouldCycle
     *           If true then when the end is reached and there is no next page it selects the first page
     *
     * @since API version 2
     */
    selectNextPageMatching(expression: string, shouldCycle: boolean): void;

    /**
     * Selects the previous page that matches the given expression.
     *
     * @param expression
     *           An expression that can match a page based on how it has been tagged. For now this can only be
     *           the name of a single tag that you would like to match.
     *
     * @param shouldCycle
     *           If true then when the end is reached and there is no next page it selects the first page
     *
     * @since API version 2
     */
    selectPreviousPageMatching(expression: string, shouldCycle: boolean): void;

    /**
     * Value that reports the currently selected parameter page index.
     *
     * @since API version 2
     */
    selectedPageIndex(): SettableIntegerValue;
  }

  // source: com/bitwig/extension/controller/api/CursorTrack.java

  /**
   * Instances of this interface represent the cursor item of track selections.
   *
   * @since API version 1
   */
  interface CursorTrack extends CursorChannel, Track, PinnableCursor {
    /**
     * Makes the cursor track point to it's parent group track, in case it is not already pointing to the root
     * group track.
     *
     * @since API version 1
     */
    selectParent(): void;

    /**
     * Makes the cursor track point to the first child found with the track group that this cursor currently
     * points to. If this cursor is not pointing to a track group or the track group is empty then this has no
     * effect.
     *
     * @since API version 2
     */
    selectFirstChild(): void;

    /**
     * Specifies the behaviour of the functions {@link #selectPrevious()}, {@link #selectNext()},
     * {@link #selectFirst()} and {@link #selectLast()}. Calling those functions can either navigate the cursor
     * within the current nesting level, or over a flat list of either all tracks or only the expanded tracks.
     * Default is CursorNavigationMode.FLAT.
     *
     * @since API version 1
     */
    setCursorNavigationMode(mode: CursorNavigationMode): void;

    createCursorDevice(): PinnableCursorDevice;

    createCursorDevice(name: string): PinnableCursorDevice;

    createCursorDevice(name: string, numSends: number): PinnableCursorDevice;

    /**
     * Creates a {@link CursorDevice} for this cursor track that by default follows a device based on the
     * supplied follow mode.
     *
     * @param id
     *           An id that is used to identify this cursor.
     * @param name
     *           A name that is displayed to the user for this cursor.
     * @param numSends
     *           the number of sends that are simultaneously accessible in nested channels.
     * @param followMode
     *           Mode that defines how this cursor should follow devices.
     *
     * @since API version 2
     */
    createCursorDevice(
      id: string,
      name: string,
      numSends: number,
      followMode: CursorDeviceFollowMode
    ): PinnableCursorDevice;
  }

  // source: com/bitwig/extension/controller/api/Device.java

  import DirectParameterDisplayedValueChangedCallback = com.bitwig.extension.callback.DirectParameterDisplayedValueChangedCallback;
  import DirectParameterNameChangedCallback = com.bitwig.extension.callback.DirectParameterNameChangedCallback;
  import DirectParameterNormalizedValueChangedCallback = com.bitwig.extension.callback.DirectParameterNormalizedValueChangedCallback;
  import StringArrayValueChangedCallback = com.bitwig.extension.callback.StringArrayValueChangedCallback;

  /**
   * This interface represents a device in Bitwig Studio, both internal devices and plugins.
   *
   * @since API version 1
   */
  interface Device extends ObjectProxy {
    /**
     * Returns a representation of the device chain that contains this device. Possible device chain instances
     * are tracks, device layers, drums pads, or FX slots.
     *
     * @return the requested device chain object
     * @since API version 1
     * @deprecated Use {@link #deviceChain()} instead.
     */
    getDeviceChain(): DeviceChain;

    /**
     * Returns a representation of the device chain that contains this device. Possible device chain instances
     * are tracks, device layers, drums pads, or FX slots.
     *
     * @return the requested device chain object
     * @since API version 5
     */
    deviceChain(): DeviceChain;

    /**
     * Value that reports the position of the device within the parent device chain.
     *
     * @since API version 2
     */
    position(): IntegerValue;

    /**
     * Registers an observer that reports the position of the device within the parent device chain.
     *
     * @param callback
     *           a callback function that receives a single integer parameter
     * @since API version 1
     * @deprecated Use {@link #position()} instead.
     */
    addPositionObserver(callback: IntegerValueChangedCallback): void;

    /**
     * Returns an object that provides access to the open state of plugin windows.
     *
     * @return a boolean value object that represents the open state of the editor window, in case the device
     *         features a custom editor window (such as plugins).
     * @since API version 1
     */
    isWindowOpen(): SettableBooleanValue;

    /**
     * Returns an object that provides access to the expanded state of the device.
     *
     * @return a boolean value object that represents the expanded state of the device.
     * @since API version 1
     */
    isExpanded(): SettableBooleanValue;

    /**
     * Returns an object that provides access to the visibility of the device macros section.
     *
     * @return a boolean value object that represents the macro section visibility.
     *
     * @deprecated Use {@link #isRemoteControlsSectionVisible()} instead
     * @since API version 1
     */
    isMacroSectionVisible(): SettableBooleanValue;

    /**
     * Returns an object that provides access to the visibility of the device remote controls section.
     *
     * @return a boolean value object that represents the remote controls section visibility.
     *
     * @since API version 2
     */
    isRemoteControlsSectionVisible(): SettableBooleanValue;

    /**
     * Returns an object that provides access to the visibility of the parameter page mapping editor.
     *
     * @return a boolean value object that represents visibility of the parameter page mapping editor.
     *
     * @deprecated Use {@link #isRemoteControlsSectionVisible()} instead
     * @since API version 1
     */
    isParameterPageSectionVisible(): SettableBooleanValue;

    /**
     * Returns the parameter with the given index in the current parameter page.
     *
     * @param indexInPage
     *           the index of the parameter within the current parameter page.
     * @return an object that provides access to the requested parameter
     * @ @deprecated Use getRemoteControls().getRemoteControlInSlot(indexInPage)
     * @since API version 1
     */
    getParameter(indexInPage: number): Parameter;

    /**
     * Creates a cursor for the selected remote controls page in the device with the supplied number of
     * parameters. This section will follow the current page selection made by the user in the application.
     *
     * @param parameterCount
     *           The number of parameters the remote controls should contain
     *
     * @since API version 2
     */
    createCursorRemoteControlsPage(
      parameterCount: number
    ): CursorRemoteControlsPage;

    /**
     * Creates a cursor for a remote controls page in the device with the supplied number of parameters. This
     * section will be independent from the current page selected by the user in Bitwig Studio's user
     * interface. The supplied filter is an expression that can be used to match pages this section is
     * interested in. The expression is matched by looking at the tags added to the pages. If the expression is
     * empty then no filtering will occur.
     *
     * @param name
     *           A name to associate with this section. This will be used to remember manual mappings made by
     *           the user within this section.
     *
     *
     * @param parameterCount
     *           The number of parameters the remote controls should contain
     *
     * @param filterExpression
     *           An expression used to match pages that the user can navigate through. For now this can only be
     *           the name of a single tag the pages should contain (e.g "drawbars", "dyn", "env", "eq",
     *           "filter", "fx", "lfo", "mixer", "osc", "overview", "perf").
     *
     * @since API version 2
     */
    createCursorRemoteControlsPage(
      name: string,
      parameterCount: number,
      filterExpression: string
    ): CursorRemoteControlsPage;

    /**
     * Returns the parameter with the given index in the envelope parameter page.
     *
     * @param index
     *           the index of the parameter within the envelope parameter page.
     * @return an object that provides access to the requested parameter
     * @ @since API version 1
     * @deprecated The remote controls deprecate this feature. Instead create remote controls with
     *             {@link #createIndependentRemoteControls(String, int, String)}.
     */
    getEnvelopeParameter(index: number): Parameter;

    /**
     * Returns the parameter with the given index in the common parameter page.
     *
     * @param index
     *           the index of the parameter within the common parameter page.
     * @return an object that provides access to the requested parameter
     * @ @since API version 1
     * @deprecated The remote controls deprecate this feature. Instead create remote controls with
     *             {@link #createIndependentRemoteControls(String, int, String)}.
     */
    getCommonParameter(index: number): Parameter;

    /**
     * Returns the modulation source at the given index.
     *
     * @param index
     *           the index of the modulation source
     * @return An object that represents the requested modulation source
     * @ @since API version 1
     * @deprecated The remote controls deprecate this feature. Instead create remote controls with
     *             {@link #createIndependentRemoteControls(String, int, String)}.
     */
    getModulationSource(index: number): ModulationSource;

    /**
     * Returns the macro control at the given index.
     *
     * @param index
     *           the index of the macro control, must be in the range [0..7]
     * @return An object that represents the requested macro control
     * @ @since API version 1
     * @deprecated Devices no longer have a built in fixed macro section. Instead the user can define pages of
     *             mappings and these should be used instead.
     */
    getMacro(index: number): Macro;

    /**
     * Registers an observer that reports if the device is selected.
     *
     * @param callback
     *           a callback function that receives a single boolean parameter.
     * @since API version 1
     * @deprecated Use {@link #exists()} instead.
     */
    addHasSelectedDeviceObserver(callback: BooleanValueChangedCallback): void;

    /**
     * Selects the device in Bitwig Studio.
     *
     * @since API version 1
     */
    selectInEditor(): void;

    /**
     * Value that reports if the device is a plugin.
     *
     * @since API version 2
     */
    isPlugin(): BooleanValue;

    /**
     * Registers an observer that reports if the device is a plugin.
     *
     * @param callback
     *           a callback function that receives a single boolean parameter.
     * @since API version 1
     * @deprecated Use {@link #isPlugin()} instead
     */
    addIsPluginObserver(callback: BooleanValueChangedCallback): void;

    /**
     * Switches to the previous parameter page.
     *
     * @since API version 1
     */
    previousParameterPage(): void;

    /**
     * Switches to the next parameter page.
     *
     * @since API version 1
     */
    nextParameterPage(): void;

    /**
     * Registers an observer that reports if there is a previous parameter page.
     *
     * @param callback
     *           a callback function that receives a single boolean parameter
     * @since API version 1
     */
    addPreviousParameterPageEnabledObserver(
      callback: BooleanValueChangedCallback
    ): void;

    /**
     * Registers an observer that reports if there is a next parameter page.
     *
     * @param callback
     *           a callback function that receives a single boolean parameter
     * @since API version 1
     */
    addNextParameterPageEnabledObserver(
      callback: BooleanValueChangedCallback
    ): void;

    /**
     * Switches to the parameter page at the given page index.
     *
     * @param page
     *           the index of the desired parameter page
     * @since API version 1
     */
    setParameterPage(page: number): void;

    /**
     * Loads the previous preset.
     *
     * @since API version 1
     * @deprecated Use new browser API provided via {@link #createDeviceBrowser(int, int)} instead.
     */
    switchToPreviousPreset(): void;

    /**
     * Loads the next preset.
     *
     * @since API version 1
     * @deprecated Use new browser API provided via {@link #createDeviceBrowser(int, int)} instead.
     */
    switchToNextPreset(): void;

    /**
     * Switches to the previous preset category.
     *
     * @since API version 1
     * @deprecated Use new browser API provided via {@link #createDeviceBrowser(int, int)} instead.
     */
    switchToPreviousPresetCategory(): void;

    /**
     * Switches to the next preset category.
     *
     * @since API version 1
     * @deprecated Use new browser API provided via {@link #createDeviceBrowser(int, int)} instead.
     */
    switchToNextPresetCategory(): void;

    /**
     * Switches to the previous preset creator.
     *
     * @since API version 1
     * @deprecated Use new browser API provided via {@link #createDeviceBrowser(int, int)} instead.
     */
    switchToPreviousPresetCreator(): void;

    /**
     * Switches to the next preset creator.
     *
     * @since API version 1
     * @deprecated Use new browser API provided via {@link #createDeviceBrowser(int, int)} instead.
     */
    switchToNextPresetCreator(): void;

    /**
     * Returns an object used for browsing devices, presets and other content. Committing the browsing session
     * will load or create a device from the selected resource and replace the current device.
     *
     * @param numFilterColumnEntries
     *           the size of the window used to navigate the filter column entries.
     * @param numResultsColumnEntries
     *           the size of the window used to navigate the results column entries.
     * @return the requested device browser object.
     * @since API version 1
     */
    createDeviceBrowser(
      numFilterColumnEntries: number,
      numResultsColumnEntries: number
    ): Browser;

    /**
     * Value that reports the name of the device.
     *
     * @since API version 2
     */
    name(): StringValue;

    /**
     * Registers an observer that reports the name of the device.
     *
     * @param len
     *           the maximum length of the name. Longer names will get truncated.
     * @param textWhenUnassigned
     *           the default name that gets reported when the device is not associated with a Bitwig Studio
     *           device yet.
     * @param callback
     *           a callback function that receives a single name (string) parameter
     * @since API version 1
     * @deprecated Use {@link #name()} instead
     */
    addNameObserver(
      len: number,
      textWhenUnassigned: string,
      callback: StringValueChangedCallback
    ): void;

    /**
     * Value that reports the last loaded preset name.
     *
     * @since API version 2
     */
    presetName(): StringValue;

    /**
     * Registers an observer that reports the last loaded preset name.
     *
     * @param len
     *           the maximum length of the name. Longer names will get truncated.
     * @param textWhenUnassigned
     *           the default name that gets reported when the device is not associated with a Bitwig Studio
     *           device yet.
     * @param callback
     *           a callback function that receives a single name (string) parameter
     * @since API version 1
     * @deprecated Use {@link #presetName()}.addValueObserver(callback) instead.
     */
    addPresetNameObserver(
      len: number,
      textWhenUnassigned: string,
      callback: StringValueChangedCallback
    ): void;

    /**
     * Value that reports the current preset category name.
     *
     * @since API version 2
     */
    presetCategory(): StringValue;

    /**
     * Registers an observer that reports the current preset category name.
     *
     * @param len
     *           the maximum length of the name. Longer names will get truncated.
     * @param textWhenUnassigned
     *           the default name that gets reported when the device is not associated with a Bitwig Studio
     *           device yet.
     * @param callback
     *           a callback function that receives a single name (string) parameter
     * @since API version 1
     * @deprecated use {@link #presetCategory()} instead.
     */
    addPresetCategoryObserver(
      len: number,
      textWhenUnassigned: string,
      callback: StringValueChangedCallback
    ): void;

    /**
     * Value that reports the current preset creator name.
     *
     * @since API version 2
     */
    presetCreator(): StringValue;

    /**
     * Registers an observer that reports the current preset creator name.
     *
     * @param len
     *           the maximum length of the name. Longer names will get truncated.
     * @param textWhenUnassigned
     *           the default name that gets reported when the device is not associated with a Bitwig Studio
     *           device yet.
     * @param callback
     *           a callback function that receives a single name (string) parameter
     * @since API version 1
     * @deprecated Use {@link #presetCreator()} instead.
     */
    addPresetCreatorObserver(
      len: number,
      textWhenUnassigned: string,
      callback: StringValueChangedCallback
    ): void;

    /**
     * Registers an observer that reports the currently selected parameter page.
     *
     * @param valueWhenUnassigned
     *           the default page index that gets reported when the device is not associated with a device
     *           instance in Bitwig Studio yet.
     * @param callback
     *           a callback function that receives a single page index parameter (integer)
     * @since API version 1
     * @deprecated Use {@link #createCursorRemoteControlsPage(int)} instead.
     */
    addSelectedPageObserver(
      valueWhenUnassigned: number,
      callback: IntegerValueChangedCallback
    ): void;

    /**
     * Registers an observer that reports the name of the active modulation source.
     *
     * @param len
     *           the maximum length of the name. Longer names will get truncated.
     * @param textWhenUnassigned
     *           the default name that gets reported when the device is not associated with a Bitwig Studio
     *           device yet.
     * @param callback
     *           a callback function that receives a single name parameter (string)
     * @since API version 1
     * @deprecated Use {@link #createCursorRemoteControlsPage(int)} instead.
     */
    addActiveModulationSourceObserver(
      len: number,
      textWhenUnassigned: string,
      callback: StringValueChangedCallback
    ): void;

    /**
     * Registers an observer that reports the names of the devices parameter pages.
     *
     * @param callback
     *           a callback function that receives a single string array parameter containing the names of the
     *           parameter pages
     * @since API version 1
     * @deprecated Use {@link #createCursorRemoteControlsPage(int)} instead.
     */
    addPageNamesObserver(callback: StringArrayValueChangedCallback): void;

    /**
     * Registers an observer that reports the names of the available presets for the device according to the
     * current configuration of preset category and creator filtering.
     *
     * @param callback
     *           a callback function that receives a single string array parameter containing the names of the
     *           presets for the current category and creator filter.
     * @see #addPresetCategoryObserver
     * @see #addPresetCreatorObserver
     * @see #setPresetCategory
     * @see #setPresetCreator
     * @since API version 1
     * @deprecated Use the new browser API instead.
     */
    addPresetNamesObserver(callback: StringArrayValueChangedCallback): void;

    /**
     * Loads the preset with the index from the list provided by {@link #addPresetNamesObserver}.
     *
     * @since API version 1
     * @deprecated Use the new browser API instead.
     */
    loadPreset(index: number): void;

    /**
     * Registers an observer that reports the names of the available preset categories for the device.
     *
     * @param callback
     *           a callback function that receives a single string array parameter containing the names of the
     *           preset categories
     * @since API version 1
     * @deprecated Use the new browser API instead.
     */
    addPresetCategoriesObserver(
      callback: StringArrayValueChangedCallback
    ): void;

    /**
     * Sets the preset category filter with the index from the array provided by
     * {@link #addPresetCategoriesObserver}.
     *
     * @since API version 1
     * @deprecated Use the new browser API instead.
     */
    setPresetCategory(index: number): void;

    /**
     * Registers an observer that reports the names of the available preset creators for the device.
     *
     * @param callback
     *           a callback function that receives a single string array parameter containing the names of the
     *           preset creators
     * @since API version 1
     * @deprecated Use the new browser API instead.
     */
    addPresetCreatorsObserver(callback: StringArrayValueChangedCallback): void;

    /**
     * Sets the preset creator filter with the index from the list provided by
     * {@link #addPresetCreatorsObserver}.
     *
     * @since API version 1
     * @deprecated Use the new browser API instead.
     */
    setPresetCreator(index: number): void;

    /**
     * Toggles the enabled state of the device.
     *
     * @since API version 1
     * @deprecated Use isEnabled().toggle() instead.
     */
    toggleEnabledState(): void;

    /**
     * Value that reports if the device is enabled.
     *
     * @since API version 2
     */
    isEnabled(): SettableBooleanValue;

    /**
     * Registers an observer that reports if the device is enabled.
     *
     * @param callback
     *           a callback function that receives a single boolean parameter
     * @since API version 1
     * @deprecated Use {@link #isEnabled()} instead.
     */
    addIsEnabledObserver(callback: BooleanValueChangedCallback): void;

    /**
     * Indicates if the device has nested device chain slots. Use {@link #slotNames()}
     * to get a list of available slot names, and navigate to devices in those
     * slots using the {@link CursorDevice} interface.
     *
     * @return a value object that indicates if the device has nested device chains in FX slots.
     * @since API version 1
     */
    hasSlots(): BooleanValue;

    /**
     * Value of the list of available FX slots in this device.
     *
     * @since API version 2
     */
    slotNames(): StringArrayValue;

    /**
     * Registers an observer that gets notified when the list of available FX slots changes.
     *
     * @param callback
     *           a callback function which takes a single string array argument that contains the names of the
     *           slots.
     * @since API version 1
     * @deprecated Use {@link #slotNames()} instead.
     */
    addSlotsObserver(callback: StringArrayValueChangedCallback): void;

    /**
     * Returns an object that represents the selected device slot as shown in the user interface, and that
     * provides access to the contents of slot's device chain.
     *
     * @return the requested slot cursor object
     * @since API version 1
     */
    getCursorSlot(): DeviceSlot;

    /**
     * Indicates if the device is contained by another device.
     *
     * @return a value object that indicates if the device is nested
     * @since API version 1
     */
    isNested(): BooleanValue;

    /**
     * Indicates if the device supports nested layers.
     *
     * @return a value object that indicates if the device supports nested layers.
     * @since API version 1
     */
    hasLayers(): BooleanValue;

    /**
     * Indicates if the device has individual device chains for each note value.
     *
     * @return a value object that indicates if the device has individual device chains for each note value.
     * @since API version 1
     */
    hasDrumPads(): BooleanValue;

    /**
     * Create a bank for navigating the nested layers of the device using a fixed-size window.
     *
     * This bank will work over the following devices:
     *  - Instrument Layer
     *  - Effect Layer
     *  - Instrument Selector
     *  - Effect Selector
     *
     * @param numChannels
     *           the number of channels that the device layer bank should be configured with
     * @return a device layer bank object configured with the desired number of channels
     * @since API version 1
     */
    createLayerBank(numChannels: number): DeviceLayerBank;

    /**
     * Create a bank for navigating the nested layers of the device using a fixed-size window.
     *
     * @param numPads
     *           the number of channels that the drum pad bank should be configured with
     * @return a drum pad bank object configured with the desired number of pads
     * @since API version 1
     */
    createDrumPadBank(numPads: number): DrumPadBank;

    /**
     * Returns a device layer instance that can be used to navigate the layers or drum pads of the device, in
     * case it has any
     *
     * This is the selected layer from the user interface.
     *
     * @return a cursor device layer instance
     * @since API version 1
     */
    createCursorLayer(): CursorDeviceLayer;

    /**
     * Creates a ChainSelector object which will give you control over the current device if it is
     * an Instrument Selector or an Effect Selector.
     *
     * To check if the device is currently a ChainSelector, use {@link ChainSelector.exists()}.
     *
     * If you want to have access to all the chains, use {@link #createLayerBank(int)}.
     *
     * @return a chain selector instance
     * @since API version 6
     */
    createChainSelector(): ChainSelector;

    /**
     * Adds an observer on a list of all parameters for the device.
     *
     * The callback always updates with an array containing all the IDs for the device.
     *
     * @param callback
     *           function with the signature (String[])
     * @since API version 1
     */
    addDirectParameterIdObserver(
      callback: StringArrayValueChangedCallback
    ): void;

    /**
     * Adds an observer for the parameter names (initial and changes) of all parameters for the device.
     *
     * @param maxChars
     *           maximum length of the string sent to the observer.
     * @param callback
     *           function with the signature (String ID, String name)
     * @since API version 1
     */
    addDirectParameterNameObserver(
      maxChars: number,
      callback: DirectParameterNameChangedCallback
    ): void;

    /**
     * Returns an observer that reports changes of parameter display values, i.e. parameter values formatted as
     * a string to be read by the user, for example "-6.02 dB". The returned observer object can be used to
     * configure which parameters should be observed. By default no parameters are observed. It should be
     * avoided to observe all parameters at the same time for performance reasons.
     *
     * @return an observer object that can be used to enable or disable actual observing for certain
     *         parameters.
     * @param maxChars
     *           maximum length of the string sent to the observer.
     * @param callback
     *           function with the signature (String ID, String valueDisplay)
     * @since API version 1
     */
    addDirectParameterValueDisplayObserver(
      maxChars: number,
      callback: DirectParameterDisplayedValueChangedCallback
    ): DirectParameterValueDisplayObserver;

    /**
     * Adds an observer for the parameter display value (initial and changes) of all parameters for the device.
     *
     * @param callback
     *           a callback function with the signature (String ID, float normalizedValue). If the value is not
     *           accessible 'Number.NaN' (not-a-number) is reported, can be checked with 'isNaN(value)'.
     * @since API version 1
     */
    addDirectParameterNormalizedValueObserver(
      callback: DirectParameterNormalizedValueChangedCallback
    ): void;

    /**
     * Sets the parameter with the specified `id` to the given `value` according to the given `resolution`.
     *
     * @param id
     *           the parameter identifier string
     * @param value
     *           the new value normalized to the range [0..resolution-1]
     * @param resolution
     *           the resolution of the new value
     * @since API version 1
     */
    setDirectParameterValueNormalized(
      id: string,
      value: number,
      resolution: number
    ): void;

    /**
     * Increases the parameter with the specified `id` by the given `increment` according to the given
     * `resolution`. To decrease the parameter value pass in a negative increment.
     *
     * @param id
     *           the parameter identifier string
     * @param increment
     *           the amount that the parameter value should be increased by, normalized to the range
     *           [0..resolution-1]
     * @param resolution
     *           the resolution of the new value
     * @since API version 1
     */
    incDirectParameterValueNormalized(
      id: string,
      increment: number,
      resolution: number
    ): void;

    /**
     * Value that reports the file name of the currently loaded sample, in case the device is a sample
     * container device.
     *
     * @since API version 2
     */
    sampleName(): StringValue;

    /**
     * Registers an observer that reports the file name of the currently loaded sample, in case the device is a
     * sample container device.
     *
     * @param maxChars
     *           maximum length of the string sent to the observer.
     * @param textWhenUnassigned
     *           the default name that gets reported when the device is not associated with a Bitwig Studio
     *           device yet.
     * @param callback
     *           a callback function that receives a single string parameter.
     * @deprecated Use {@link #sampleName()} instead.
     */
    addSampleNameObserver(
      maxChars: number,
      textWhenUnassigned: string,
      callback: StringValueChangedCallback
    ): void;

    /**
     * Returns an object that provides bank-wise navigation of sibling devices of the same device chain
     * (including the device instance used to create the siblings bank).
     *
     * @param numDevices
     *           the number of devices that are simultaneously accessible
     * @return the requested device bank object
     * @ @since API version 1
     */
    createSiblingsDeviceBank(numDevices: number): DeviceBank;

    /**
     * Starts browsing for content that can be inserted before this device in Bitwig Studio's popup browser.
     *
     * @since API version 2
     */
    browseToInsertBeforeDevice(): void;

    /**
     * Starts browsing for content that can be inserted before this device in Bitwig Studio's popup browser.
     *
     * @since API version 2
     */
    browseToInsertAfterDevice(): void;

    /**
     * Starts browsing for content that can replace this device in Bitwig Studio's popup browser.
     *
     * @since API version 2
     */
    browseToReplaceDevice(): void;
  }

  // source: com/bitwig/extension/controller/api/DeviceBank.java

  /**
   * This interface is used for navigation of device chains in Bitwig Studio. Instances are configured with a
   * fixed number of devices and provide access to a excerpt of the devices inside a device chain. Various
   * methods are provided for scrolling to different sections of the device chain. It basically acts like a
   * window moving over the devices.
   *
   * To receive an instance of DeviceBank call {@link Track#createDeviceBank}.
   *
   * @see {@link Track#createDeviceBank}
   * @since API version 1
   */
  interface DeviceBank extends Bank<Device> {
    /**
     * Returns the object that was used to instantiate this device bank. Possible device chain instances are
     * tracks, device layers, drums pads, or FX slots.
     *
     * @return the requested device chain object
     * @since API version 1
     */
    getDeviceChain(): DeviceChain;

    /**
     * Returns the device at the given index within the bank.
     *
     * @param indexInBank
     *           the device index within this bank, not the position within the device chain. Must be in the
     *           range [0..sizeOfBank-1].
     * @return the requested device object
     * @since API version 1
     */
    getDevice(indexInBank: number): Device;

    /**
     * Scrolls the device window one page up.
     *
     * @since API version 1
     */
    scrollPageUp(): void;

    /**
     * Scrolls the device window one page down.
     *
     * @since API version 1
     */
    scrollPageDown(): void;

    /**
     * Scrolls the device window one device up.
     *
     * @since API version 1
     */
    scrollUp(): void;

    /**
     * Scrolls the device window one device down.
     *
     * @since API version 1
     */
    scrollDown(): void;

    /**
     * Makes the device with the given position visible in the track bank.
     *
     * @param position
     *           the position of the device within the device chain
     * @since API version 1
     */
    scrollTo(position: number): void;

    /**
     * Registers an observer that reports the current device scroll position.
     *
     * @param callback
     *           a callback function that takes a single integer parameter
     * @param valueWhenUnassigned
     *           the default value that gets reports when the device chain is not yet connected to a Bitwig
     *           Studio document
     * @since API version 1
     * @deprecated Use {@link #scrollPosition()} instead.
     */
    addScrollPositionObserver(
      callback: IntegerValueChangedCallback,
      valueWhenUnassigned: number
    ): void;

    /**
     * Registers an observer that reports if the device window can be scrolled further up.
     *
     * @param callback
     *           a callback function that takes a single boolean parameter
     * @since API version 1
     * @deprecated Use {@link #canScrollBackwards()} instead.
     */
    addCanScrollUpObserver(callback: BooleanValueChangedCallback): void;

    /**
     * Registers an observer that reports if the device window can be scrolled further down.
     *
     * @param callback
     *           a callback function that takes a single boolean parameter
     * @since API version 1
     * @deprecated Use {@link #canScrollForwards()} instead.
     */
    addCanScrollDownObserver(callback: BooleanValueChangedCallback): void;

    /**
     * Registers an observer that reports the total device count of the device chain (not the number of devices
     * accessible through the bank window).
     *
     * @param callback
     *           a callback function that receives a single integer parameter
     * @since API version 1
     * @deprecated Use {@link #deviceCount()}.addValueObserver(callback)
     */
    addDeviceCountObserver(callback: IntegerValueChangedCallback): void;

    /**
     * Browses for content to insert a device at the given index inside this bank.
     *
     * @param index
     *           the index to insert the device at. Must be >= 0 and <= {@link #getSizeOfBank()}.
     *
     * @since API version 2
     */
    browseToInsertDevice(index: number): void;
  }

  // source: com/bitwig/extension/controller/api/DeviceBrowsingSession.java

  /**
   * Instances of this interface are used for browsing devices, including access to all filter columns and the
   * result column as shown in the 'Devices' tab of Bitwig Studio's contextual browser window.
   *
   * @see BrowsingSession
   * @since API version 1
   * @deprecated Use {@link PopupBrowser} instead.
   */
  interface DeviceBrowsingSession extends BrowsingSession {
    /**
     * Returns the category filter as shown in the category column of the browser.
     *
     * @return the requested category filter object.
     * @since API version 1
     */
    getCategoryFilter(): BrowserFilterColumn;

    /**
     * Returns the device type filter as shown in the category column of the browser.
     *
     * @return the requested device type filter object.
     * @since API version 1
     */
    getDeviceTypeFilter(): BrowserFilterColumn;

    /**
     * Returns the file type filter as shown in the category column of the browser.
     *
     * @return the requested file type filter object.
     * @since API version 1
     */
    getFileTypeFilter(): BrowserFilterColumn;
  }

  // source: com/bitwig/extension/controller/api/DeviceChain.java

  /**
   * The foundation of all interfaces that contain devices, such as tracks, device layers, drum pads or FX
   * slots.
   *
   * @since API version 1
   */
  interface DeviceChain extends ObjectProxy {
    /**
     * Selects the device chain in Bitwig Studio, in case it is a selectable object.
     *
     * @since API version 1
     */
    selectInEditor(): void;

    /**
     * Value that reports the name of the device chain, such as the track name or the drum pad
     * name.
     *
     * @since API version 2
     */
    name(): SettableStringValue;

    /**
     * Registers an observer that reports the name of the device chain, such as the track name or the drum pad
     * name.
     *
     * @param numChars
     *           the maximum number of characters used for the reported name
     * @param textWhenUnassigned
     *           the default text that gets reported when the device chain is not associated with an object in
     *           Bitwig Studio yet.
     * @param callback
     *           a callback function that receives a single name parameter (string).
     * @since API version 1
     * @deprecated Use {@link #name()} instead.
     */
    addNameObserver(
      numChars: number,
      textWhenUnassigned: string,
      callback: StringValueChangedCallback
    ): void;

    /**
     * Registers an observer that reports if the device chain is selected in Bitwig Studio editors.
     *
     * @param callback
     *           a callback function that takes a single boolean parameter.
     * @since API version 1
     */
    addIsSelectedInEditorObserver(callback: BooleanValueChangedCallback): void;

    /**
     * @deprecated Use {@link #addIsSelectedInEditorObserver} instead.
     * @since API version 1
     */
    addIsSelectedObserver(callback: BooleanValueChangedCallback): void;

    /**
     * Returns an object that provides bank-wise navigation of devices.
     *
     * @param numDevices
     *           the number of devices should be accessible simultaneously
     * @return the requested device bank object
     * @ @since API version 1
     */
    createDeviceBank(numDevices: number): DeviceBank;

    /**
     * Returns an object used for browsing devices, presets and other content. Committing the browsing session
     * will load or create a device from the selected resource and insert it into the device chain.
     *
     * @param numFilterColumnEntries
     *           the size of the window used to navigate the filter column entries.
     * @param numResultsColumnEntries
     *           the size of the window used to navigate the results column entries.
     * @return the requested device browser object.
     * @since API version 1
     */
    createDeviceBrowser(
      numFilterColumnEntries: number,
      numResultsColumnEntries: number
    ): Browser;

    /**
     * @deprecated Use {@link #selectInEditor()} or {@link Channel#selectInMixer()} instead.
     * @since API version 1
     */
    select(): void;

    /**
     * Starts browsing for content that can be inserted at the start of this device chain.
     *
     * @since API version 2
     */
    browseToInsertAtStartOfChain(): void;

    /**
     * Starts browsing for content that can be inserted at the end of this device chain.
     *
     * @since API version 2
     */
    browseToInsertAtEndOfChain(): void;
  }

  // source: com/bitwig/extension/controller/api/DeviceLayer.java

  /**
   * Instances of this interface represent device layers in Bitwig Studio.
   *
   * @since API version 1
   */
  interface DeviceLayer extends Channel {}

  // source: com/bitwig/extension/controller/api/DeviceLayerBank.java

  /**
   * Devices layers are features of special Bitwig Studio devices, more specifically the Layer Instrument and
   * Layer FX devices, and are also shown as sub-channels in the mixer panel.
   *
   * Instances of device layer bank are configured with a fixed number of channels and represent an excerpt of
   * underlying complete list of channels. Various methods are provided for scrolling to different sections of
   * the underlying list. It basically acts like a one-dimensional window moving over the device layers.
   *
   * To receive an instance of device layer bank call {@link Device#createLayerBank(int numChannels)}.
   *
   * @see {@link Device#createLayerBank}
   * @since API version 1
   */
  interface DeviceLayerBank extends ChannelBank<DeviceLayer> {
    /**
     * Returns the device layer at the given index.
     *
     * @param indexInBank
     *           the device layer index within this bank, not the index within the list of all device layers as
     *           shown in Bitwig Studio layer devices. Must be in the range [0..sizeOfBank-1].
     * @return the device layer object
     * @since API version 1
     * @deprecated Use {@link #getItemAt(int)} instead.
     */
    getChannel(indexInBank: number): DeviceLayer;
  }

  // source: com/bitwig/extension/controller/api/DeviceSlot.java

  /**
   * Instances of this interface represent nested FX slots in devices.
   *
   * @since API version 1
   */
  interface DeviceSlot extends DeviceChain {}

  // source: com/bitwig/extension/controller/api/DirectParameterValueDisplayObserver.java

  /**
   * This interface is used to configure observation of pretty-printed device parameter values.
   *
   * @since API version 1
   */
  interface DirectParameterValueDisplayObserver {
    /**
     * Starts observing the parameters according to the given parameter ID array, or stops observing in case
     * `null` is passed in for the parameter ID array.
     *
     * @param parameterIds
     *           the array of parameter IDs or `null` to stop observing parameter display values.
     * @since API version 1
     */
    setObservedParameterIds(parameterIds: string[]): void;
  }

  // source: com/bitwig/extension/controller/api/DocumentState.java

  /**
   * This interface is used to save custom script settings inside Bitwig Studio documents. The settings are
   * shown to the user in the `Studio IO` panel of Bitwig Studio.
   *
   * @since API version 1
   */
  interface DocumentState extends Settings {}

  // source: com/bitwig/extension/controller/api/DoubleValue.java

  /**
   * Instances of this interface represent double values.
   * @since API version 2
   */
  interface DoubleValue extends Value<DoubleValueChangedCallback> {
    /**
     * Gets the current value.
     *
     * @since API version 2
     */
    get(): number;
  }

  // source: com/bitwig/extension/controller/api/DrumPad.java

  /**
   * Instances of this interface are special kind of channel objects that represent the pads of a drum machine
   * instrument. Drum pads are typically associated to channel objects via note keys.
   *
   * @since API version 1
   */
  interface DrumPad extends Channel {}

  // source: com/bitwig/extension/controller/api/DrumPadBank.java

  /**
   * Drum pads are features of special Bitwig Studio devices (currently only the Bitwig Drum Machine
   * instrument), and are also shown as sub-channels in the mixer panel.
   *
   * Instances of drum pad bank are configured with a fixed number of pads/channels and represent an excerpt of
   * underlying complete list of channels. Various methods are provided for scrolling to different sections of
   * the underlying list. It basically acts like a one-dimensional window moving over the drum pad channels.
   *
   * To receive an instance of drum pad bank call {@link Device#createDrumPadBank(int numChannels)}.
   *
   * @see {@link Device#createDrumPadBank}
   * @since API version 1
   */
  interface DrumPadBank extends ChannelBank<DrumPad> {
    /**
     * Specifies if the Drum Machine should visualize which pads are part of the window. By default indications
     * are enabled.
     *
     * @param shouldIndicate
     *           `true` if visual indications should be enabled, `false` otherwise
     * @since API version 1
     */
    setIndication(shouldIndicate: boolean): void;
  }

  // source: com/bitwig/extension/controller/api/EnumValue.java

  import EnumValueChangedCallback = com.bitwig.extension.callback.EnumValueChangedCallback;

  interface EnumValue extends Value<EnumValueChangedCallback> {
    /**
     * Gets the current value.
     *
     * @since API version 2
     */
    get(): string;
  }

  // source: com/bitwig/extension/controller/api/GenericBrowsingSession.java

  /**
   * Instances of this interface are used for browsing material with bank-wise access to the filter columns.
   *
   * @see com.bitwig.extension.controller.api.BrowsingSession
   * @since API version 1
   * @deprecated Use {@link PopupBrowser} instead.
   */
  interface GenericBrowsingSession extends BrowsingSession {
    /**
     * Value that reports the name of the browsing session.
     *
     * @since API version 2
     */
    name(): StringValue;

    /**
     * Registers an observer that reports the name of the browsing session.
     *
     * @param callback
     *           a callback function that receives a single string argument.
     * @since API version 1
     * @deprecated Use
     */
    addNameObserver(
      maxCharacters: number,
      textWhenUnassigned: string,
      callback: StringValueChangedCallback
    ): void;
  }

  // source: com/bitwig/extension/controller/api/Groove.java

  /**
   * An interface representing the global groove settings of the project.
   *
   * @since API version 1
   */
  interface Groove {
    /**
     * Returns the enabled state of the groove.
     *
     * @return an object that provides access to the groove on/off setting
     * @since API version 1
     */
    getEnabled(): Parameter;

    /**
     * Returns the object that represents the shuffle amount in Bitwig Studio.
     *
     * @return an ranged value object that provides access to the shuffle amount
     * @since API version 1
     */
    getShuffleAmount(): Parameter;

    /**
     * Returns the object that represents the shuffle rate in Bitwig Studio.
     *
     * @return an ranged value object that provides access to the shuffle rate
     * @since API version 1
     */
    getShuffleRate(): Parameter;

    /**
     * Returns the object that represents the accent amount in Bitwig Studio.
     *
     * @return an ranged value object that provides access to the accent amount
     * @since API version 1
     */
    getAccentAmount(): Parameter;

    /**
     * Returns the object that represents the accent rate in Bitwig Studio.
     *
     * @return an ranged value object that provides access to the accent rate
     * @since API version 1
     */
    getAccentRate(): Parameter;

    /**
     * Returns the object that represents the accent phase in Bitwig Studio.
     *
     * @return an ranged value object that provides access to the accent phase
     * @since API version 1
     */
    getAccentPhase(): Parameter;
  }

  // source: com/bitwig/extension/controller/api/IntegerValue.java

  // @ts-ignore
  interface IntegerValue extends Value<IntegerValueChangedCallback> {
    /**
     * Gets the current value.
     *
     * @since API version 2
     */
    get(): number;

    /**
     * Adds an observer that is notified when this value changes. This is intended to aid in backwards
     * compatibility for drivers written to the version 1 API.
     *
     * @param callback
     *           The callback to notify with the new value
     * @param valueWhenUnassigned
     *           The value that the callback will be notified with if this value is not currently assigned to
     *           anything.
     */
    addValueObserver(
      callback: IntegerValueChangedCallback,
      valueWhenUnassigned: number
    ): void;
  }

  // source: com/bitwig/extension/controller/api/Macro.java

  /**
   * Instances of this interface are used to represent macro controls in Bitwig Studio to controllers.
   *
   * @deprecated Macros no longer exist as built in features of all devices. Instead the user can customize
   *             pages of controls.
   * @since API version 1
   */
  interface Macro {
    /**
     * Returns an object that provides access to the control value of the macro.
     *
     * @return a ranged value object.
     * @since API version 1
     */
    getAmount(): Parameter;

    /**
     * Returns an object that provides access to the modulation source of the macro.
     *
     * @return a modulation source object.
     * @since API version 1
     */
    getModulationSource(): ModulationSource;

    /**
     * Registers an observer that reports the label of the macro control.
     *
     * @param numChars
     *           the maximum number of characters of the reported label
     * @param textWhenUnassigned
     *           the default text that is reported when the macro is not connected to a Bitwig Studio macro
     *           control.
     * @param callback
     *           a callback function that receives a single string parameter.
     */
    addLabelObserver(
      numChars: number,
      textWhenUnassigned: string,
      callback: StringValueChangedCallback
    ): void;
  }

  // source: com/bitwig/extension/controller/api/MasterTrack.java

  /**
   * A special kind of track that represents the master track in Bitwig Studio.
   *
   * @since API version 1
   */
  interface MasterTrack extends Track {}

  // source: com/bitwig/extension/controller/api/MidiIn.java

  import ShortMidiDataReceivedCallback = com.bitwig.extension.callback.ShortMidiDataReceivedCallback;
  import SysexMidiDataReceivedCallback = com.bitwig.extension.callback.SysexMidiDataReceivedCallback;

  /**
   * Instances of this interface are used to setup handler functions for incoming MIDI messages from a specific
   * MIDI hardware.
   *
   * @since API version 1
   */
  interface MidiIn {
    /**
     * Registers a callback for receiving short (normal) MIDI messages on this MIDI input port.
     *
     * @param callback
     *           a callback function that receives three integer parameters: 1. the status byte 2. the data1
     *           value 2. the data2 value
     * @ @since API version 1
     */
    setMidiCallback(callback: ShortMidiDataReceivedCallback): void;

    /**
     * Registers a callback for receiving sysex MIDI messages on this MIDI input port.
     *
     * @param callback
     *           a callback function that takes a single string argument
     * @ @since API version 1
     */
    setSysexCallback(callback: SysexMidiDataReceivedCallback): void;

    /**
     * Creates a note input that appears in the track input choosers in Bitwig Studio. This method must be
     * called within the `init()` function of the script. The messages matching the given mask parameter will
     * be fed directly to the application, and are not processed by the script.
     *
     * @param name
     *           the name of the note input as it appears in the track input choosers in Bitwig Studio
     * @param masks
     *           a filter string formatted as hexadecimal value with `?` as wildcard. For example `80????`
     *           would match note-off on channel 1 (0). When this parameter is {@null}, a standard filter will
     *           be used to forward note-related messages on channel 1 (0).
     *
     *           If multiple note input match the same MIDI event then they'll all receive the MIDI event, and
     *           if one of them does not consume events then the events wont' be consumed.
     * @return the object representing the requested note input
     * @since API version 1
     */
    createNoteInput(name: string, ...masks: string[]): NoteInput;
  }

  // source: com/bitwig/extension/controller/api/MidiOut.java

  /**
   * Instances of this interface are used to send MIDI messages to a specific MIDI hardware.
   *
   * @since API version 1
   */
  interface MidiOut {
    /**
     * Sends a MIDI message to the hardware device.
     *
     * @param status
     *           the status byte of the MIDI message
     * @param data1
     *           the data1 part of the MIDI message
     * @param data2
     *           the data2 part of the MIDI message
     * @ @since API version 1
     */
    sendMidi(status: number, data1: number, data2: number): void;

    /**
     * Sends a MIDI SysEx message to the hardware device.
     *
     * @param hexString
     *           the sysex message formatted as hexadecimal value string
     * @ @since API version 1
     */
    sendSysex(hexString: string): void;

    /**
     * Sends a MIDI SysEx message to the hardware device.
     *
     * @param hexString
     *           the sysex message formatted as hexadecimal value string
     * @ @since API version 1
     */
    sendSysex(data: number): void;

    /**
     * Enables or disables sending MIDI beat clock messages to the hardware depending on the given parameter.
     * Typically MIDI devices that run an internal sequencer such as hardware step sequencers would be
     * interested in MIDI clock messages.
     *
     * @param shouldSendClock
     *           `true` in case the hardware should receive MIDI clock messages, `false` otherwise
     * @ @deprecated Users should enable the clock from the settings.
     * @since API version 1
     */
    setShouldSendMidiBeatClock(shouldSendClock: boolean): void;
  }

  // source: com/bitwig/extension/controller/api/Mixer.java

  /**
   * An interface used to access various commands that can be performed on the Bitwig Studio mixer panel.<br/>
   *
   * To get an instance of the mixer interface call {@link ControllerHost#createMixer}.
   *
   * @since API version 1
   */
  interface Mixer {
    /**
     * Gets an object that allows to show/hide the meter section of the mixer panel. Observers can be
     * registered on the returned object for receiving notifications when the meter section switches between
     * shown and hidden state.
     *
     * @return a boolean value object that represents the meter section visibility
     * @since API version 1
     */
    isMeterSectionVisible(): SettableBooleanValue;

    /**
     * Gets an object that allows to show/hide the io section of the mixer panel. Observers can be registered
     * on the returned object for receiving notifications when the io section switches between shown and hidden
     * state.
     *
     * @return a boolean value object that represents the io section visibility
     * @since API version 1
     */
    isIoSectionVisible(): SettableBooleanValue;

    /**
     * Gets an object that allows to show/hide the sends section of the mixer panel. Observers can be
     * registered on the returned object for receiving notifications when the sends section switches between
     * shown and hidden state.
     *
     * @return a boolean value object that represents the sends section visibility
     * @since API version 1
     */
    isSendSectionVisible(): SettableBooleanValue;

    /**
     * Gets an object that allows to show/hide the clip launcher section of the mixer panel. Observers can be
     * registered on the returned object for receiving notifications when the clip launcher section switches
     * between shown and hidden state.
     *
     * @return a boolean value object that represents the clip launcher section visibility
     * @since API version 1
     */
    isClipLauncherSectionVisible(): SettableBooleanValue;

    /**
     * Gets an object that allows to show/hide the devices section of the mixer panel. Observers can be
     * registered on the returned object for receiving notifications when the devices section switches between
     * shown and hidden state.
     *
     * @return a boolean value object that represents the devices section visibility
     * @since API version 1
     */
    isDeviceSectionVisible(): SettableBooleanValue;

    /**
     * Gets an object that allows to show/hide the cross-fade section of the mixer panel. Observers can be
     * registered on the returned object for receiving notifications when the cross-fade section switches
     * between shown and hidden state.
     *
     * @return a boolean value object that represents the cross-fade section visibility
     * @since API version 1
     */
    isCrossFadeSectionVisible(): SettableBooleanValue;

    /**
     * Registers an observer that reports if the meter section is visible (callback argument is `true`) in the
     * mixer panel or not (callback argument is `false`).
     *
     * @param callback
     *           a callback function that receives a single boolean parameter.
     * @see #isMeterSectionVisible()
     * @deprecated call `isMeterSectionVisible().addValueObserver` instead
     * @since API version 1
     */
    addMeterSectionVisibilityObserver(
      callback: BooleanValueChangedCallback
    ): void;

    /**
     * Registers an observer that reports if the IO section is visible (callback argument is `true`) in the
     * mixer panel or not (callback argument is `false`).
     *
     * @param callback
     *           a callback function that receives a single boolean parameter.
     * @see #isIoSectionVisible()
     * @deprecated call `isIoSectionVisible().addValueObserver` instead
     * @since API version 1
     */
    addIoSectionVisibilityObserver(callback: BooleanValueChangedCallback): void;

    /**
     * Registers an observer that reports if the send control section is visible (callback argument is `true`)
     * in the mixer panel or not (callback argument is `false`).
     *
     * @param callback
     *           a callback function that receives a single boolean parameter.
     * @see #isSendSectionVisible()
     * @deprecated call `isSendSectionVisible().addValueObserver` instead
     * @since API version 1
     */
    addSendsSectionVisibilityObserver(
      callback: BooleanValueChangedCallback
    ): void;

    /**
     * Registers an observer that reports if the clip launcher section is visible (callback argument is `true`)
     * in the mixer panel or not (callback argument is `false`).
     *
     * @param callback
     *           a callback function that receives a single boolean parameter.
     * @see #isClipLauncherSectionVisible()
     * @deprecated call `isClipLauncherSectionVisible().addValueObserver` instead
     * @since API version 1
     */
    addClipLauncherSectionVisibilityObserver(
      callback: BooleanValueChangedCallback
    ): void;

    /**
     * Registers an observer that reports if the device section is visible (callback argument is `true`) in the
     * mixer panel or not (callback argument is `false`).
     *
     * @param callback
     *           a callback function that receives a single boolean parameter.
     * @see #isDeviceSectionVisible()
     * @deprecated call `isDeviceSectionVisible().addValueObserver` instead
     * @since API version 1
     */
    addDeviceSectionVisibilityObserver(
      callback: BooleanValueChangedCallback
    ): void;

    /**
     * Registers an observer that reports if the cross-fade section is visible (callback argument is `true`) in
     * the mixer panel or not (callback argument is `false`).
     *
     * @param callback
     *           a callback function that receives a single boolean parameter.
     * @see #isCrossFadeSectionVisible()
     * @deprecated call `isCrossFadeSectionVisible().addValueObserver` instead
     * @since API version 1
     */
    addCrossFadeSectionVisibilityObserver(
      callback: BooleanValueChangedCallback
    ): void;

    /**
     * Toggles the visibility of the meter section in the mixer panel.
     *
     * @see #isMeterSectionVisible()
     * @deprecated call `isMeterSectionVisible().toggle()` instead
     * @since API version 1
     */
    toggleMeterSectionVisibility(): void;

    /**
     * Toggles the visibility of the IO section in the mixer panel.
     *
     * @see #isIoSectionVisible()
     * @deprecated call `isIoSectionVisible().toggle()` instead
     * @since API version 1
     */
    toggleIoSectionVisibility(): void;

    /**
     * Toggles the visibility of the send control section in the mixer panel.
     *
     * @see #isSendSectionVisible()
     * @deprecated call `isSendSectionVisible().toggle()` instead
     * @since API version 1
     */
    toggleSendsSectionVisibility(): void;

    /**
     * Toggles the visibility of the clip launcher section in the mixer panel.
     *
     * @see #isClipLauncherSectionVisible()
     * @deprecated call `isClipLauncherSectionVisible().toggle()` instead
     * @since API version 1
     */
    toggleClipLauncherSectionVisibility(): void;

    /**
     * Toggles the visibility of the device section in the mixer panel.
     *
     * @see #isDeviceSectionVisible()
     * @deprecated call `isDeviceSectionVisible().toggle()` instead
     * @since API version 1
     */
    toggleDeviceSectionVisibility(): void;

    /**
     * Toggles the visibility of the cross-fade section in the mixer panel.
     *
     * @see #isCrossFadeSectionVisible()
     * @deprecated call `isCrossFadeSectionVisible().toggle()` instead
     * @since API version 1
     */
    toggleCrossFadeSectionVisibility(): void;
  }

  // source: com/bitwig/extension/controller/api/ModulationSource.java

  /**
   * This interface represents a modulation source in Bitwig Studio.
   *
   * @since API version 1
   */
  interface ModulationSource {
    /**
     * Value which reports when the modulation source is in mapping mode.
     *
     * @since API version 2
     */
    isMapping(): BooleanValue;

    /**
     * Registers an observer which reports when the modulation source is in mapping mode.
     *
     * @param callback
     *           a callback function that receives a single boolean parameter
     * @since API version 1
     * @deprecated Use {@link #isMapping()} instead.
     */
    addIsMappingObserver(callback: BooleanValueChangedCallback): void;

    /**
     * Toggles the modulation source between mapping mode and normal control functionality.
     *
     * @since API version 1
     */
    toggleIsMapping(): void;

    /**
     * Value the reports the name of the modulation source.
     *
     * @since API version 2
     */
    name(): StringValue;

    /**
     * Registers an observer the reports the name of the modulation source.
     *
     * @param numChars
     *           the maximum number of character the reported name should be long
     * @param textWhenUnassigned
     *           the default text that gets reported if the modulation source is not connected to to a
     *           modulation source in Bitwig Studio yet
     * @param callback
     *           a callback function that receives a single string parameter
     * @since API version 1
     * @deprecated Use {@link #name()} instead.
     */
    addNameObserver(
      numChars: number,
      textWhenUnassigned: string,
      callback: StringValueChangedCallback
    ): void;

    /**
     * Value which reports if the modulation source is mapped to any destination(s).
     *
     * @since API version 2
     */
    isMapped(): BooleanValue;

    /**
     * Registers an observer which reports if the modulation source is mapped to any destination(s).
     *
     * @param callback
     *           a callback function that receives a single boolean parameter
     * @since API version 1
     * @deprecated Use {@link #isMapped()} instead.
     */
    addIsMappedObserver(callback: BooleanValueChangedCallback): void;
  }

  // source: com/bitwig/extension/controller/api/MultiSampleBrowsingSession.java

  /**
   * Instances of this interface are used for browsing multi-samples, including access to all filter columns and
   * the result column as shown in the 'Multi-Samples' tab of Bitwig Studio's contextual browser window.
   *
   * @see BrowsingSession
   * @since API version 1
   * @deprecated Use {@link PopupBrowser} instead.
   */
  interface MultiSampleBrowsingSession extends BrowsingSession {
    /**
     * Returns the file type filter as shown in the category column of the browser.
     *
     * @return the requested file type filter object.
     * @since API version 1
     */
    getFileTypeFilter(): BrowserFilterColumn;
  }

  // source: com/bitwig/extension/controller/api/MusicBrowsingSession.java

  /**
   * Instances of this interface are used for browsing music files, including access to all filter columns and
   * the result column as shown in the 'Music' tab of Bitwig Studio's contextual browser window.
   *
   * @see BrowsingSession
   * @since API version 1
   * @deprecated Use {@link PopupBrowser} instead.
   */
  interface MusicBrowsingSession extends BrowsingSession {
    /**
     * Returns the file type filter as shown in the category column of the browser.
     *
     * @return the requested file type filter object.
     * @since API version 1
     */
    getFileTypeFilter(): BrowserFilterColumn;
  }

  // source: com/bitwig/extension/controller/api/NoteInput.java

  enum NoteExpression {
    NONE = 0,
    PITCH_DOWN = 1,
    PITCH_UP = 2,
    GAIN_DOWN = 3,
    GAIN_UP = 4,
    PAN_LEFT = 5,
    PAN_RIGHT = 6,
    TIMBRE_DOWN = 7,
    TIMBRE_UP = 8,
  }

  /**
   * Instances of this interface implement note input functionality used for recording notes in Bitwig Studio
   * and for playing the instruments in tracks on hardware keyboards. In Bitwig Studio the note inputs are shown
   * in the input choosers of Bitwig Studio tracks.
   *
   * @since API version 1
   */
  interface NoteInput {
    /**
     * Specifies if the note input should consume MIDI notes, or in other words if it should prevent forwarding
     * incoming notes to the MIDI callback registered in {@link MidiIn#setMidiCallback}. This setting is `true`
     * by default.
     *
     * @param shouldConsumeEvents
     *           `true` if note events should be consumed, `false` of the events should be additionally sent to
     *           the callback registered via {@link MidiIn#setMidiCallback}
     * @since API version 1
     */
    setShouldConsumeEvents(shouldConsumeEvents: boolean): void;

    /**
     * Specifies a translation table which defines the actual key value (0-127) of notes arriving in Bitwig
     * Studio for each note key potentially received from the hardware. This is used for note-on/off and
     * polyphonic aftertouch events. Specifying a value of `-1` for a key means that notes with the key value
     * will be filtered out.
     *
     * Typically this method is used to implement transposition or scale features in controller scripts. By
     * default an identity transform table is configured, which means that all incoming MIDI notes keep their
     * original key value when being sent into Bitwig Studio.
     *
     * @param table
     *           an array which should contain 128 entries. Each entry should be a note value in the range
     *           [0..127] or -1 in case of filtering.
     * @since API version 1
     */
    setKeyTranslationTable(table: number[]): void;

    /**
     * Specifies a translation table which defines the actual velocity value (0-127) of notes arriving in
     * Bitwig Studio for each note velocity potentially received from the hardware. This is used for note-on
     * events only.
     *
     * Typically this method is used to implement velocity curves or fixed velocity mappings in controller
     * scripts. By default an identity transform table is configured, which means that all incoming MIDI notes
     * keep their original velocity when being sent into Bitwig Studio.
     *
     * @param table
     *           an array which should contain 128 entries. Each entry should be a note value in the range
     *           [0..127] or -1 in case of filtering.
     * @since API version 1
     */
    setVelocityTranslationTable(table: number[]): void;

    /**
     * Assigns polyphonic aftertouch MIDI messages to the specified note expression. Multi-dimensional control
     * is possible by calling this method several times with different MIDI channel parameters. If a key
     * translation table is configured by calling {@link #setKeyTranslationTable}, that table is used for
     * polyphonic aftertouch as well.
     *
     * @param channel
     *           the MIDI channel to map, range [0..15]
     * @param expression
     *           the note-expression to map for the given MIDI channel
     * @param pitchRange
     *           the pitch mapping range in semitones, values must be in the range [1..24]. This parameter is
     *           ignored for non-pitch expressions.
     * @ @since API version 1
     */
    assignPolyphonicAftertouchToExpression(
      channel: number,
      expression: NoteExpression,
      pitchRange: number
    ): void;

    /**
     * Enables use of Expressive MIDI mode. (note-per-channel)
     *
     * @param useExpressiveMidi
     *           enabled/disable the MPE mode for this note-input
     * @param baseChannel
     *           which channel (must be either 0 or 15) which is used as the base for this note-input
     * @param pitchBendRange
     *           initial pitch bend range used
     */
    setUseExpressiveMidi(
      useExpressiveMidi: boolean,
      baseChannel: number,
      pitchBendRange: number
    ): void;

    /**
     * Enables use of Multidimensional Polyphonic Expression mode. (note-per-channel)
     *
     * @param useMPE
     *           enabled/disable the MPE mode for this note-input
     * @param baseChannel
     *           which channel (must be either 0 or 15) which is used as the base for this note-input
     */
    setUseMultidimensionalPolyphonicExpression(
      useMPE: boolean,
      baseChannel: number
    ): void;

    /**
     * Sends MIDI data directly to the note input. This will bypass both the event filter and translation
     * tables. The MIDI channel of the message will be ignored.
     *
     * @param status
     *           the status byte of the MIDI message
     * @param data0
     *           the data0 part of the MIDI message
     * @param data1
     *           the data1 part of the MIDI message
     * @since API version 1
     */
    sendRawMidiEvent(status: number, data0: number, data1: number): void;
  }

  // source: com/bitwig/extension/controller/api/NoteLane.java

  /**
   * Instances of this interface are used to access the notes for a specific note key.
   *
   * @since API version 1
   */
  interface NoteLane {
    /**
     * Value which represents the id of this lane. is either the note pitch represented by this lane, or in
     * case of audio a lane index (currently always 0 in that case).
     *
     * @since API version 2
     */
    noteLaneId(): IntegerValue;

    /**
     * Registers an observer for the note value, which is either the note pitch represented by this lane, or in
     * case of audio a lane index (currently always 0 in that case).
     *
     * @since API version 1
     * @deprecated Use {@link #noteLaneId()} instead.
     */
    addNoteValueObserver(callback: IntegerValueChangedCallback): void;

    /**
     * Value  that reports the name of the note lane. Typically the name of a note lane is
     * either equal to the title of an associated drum pad, or reflects the pitch of the note, e.g. "C#3".
     */
    name(): StringValue;

    /**
     * Registers an observer that reports the name of the note lane. Typically the name of a note lane is
     * either equal to the title of an associated drum pad, or reflects the pitch of the note, e.g. "C#3".
     *
     * @param numChars
     *           the maximum number of characters used for the reported name
     * @param textWhenUnassigned
     *           the default name that gets reported when the lane is not yet associated with a note lane in
     *           Bitwig Studio
     * @param callback
     *           a callback function that receives a single string argument
     * @since API version 1
     * @deprecated Use {@link #name()} instead.
     */
    addNameObserver(
      numChars: number,
      textWhenUnassigned: string,
      callback: StringValueChangedCallback
    ): void;

    /**
     * Value the color of the note lane. By default the reported color will be the
     * track color, or in case an associated drum pad has a custom color it will be the color of that pad
     */
    color(): SettableColorValue;

    /**
     * Registers an observer that reports the color of the note lane. By default the reported color will be the
     * track color, or in case an associated drum pad has a custom color it will be the color of that pad.
     *
     * @param callback
     *           a callback function that receives three arguments which from an RGB color: 1. the red
     *           dimension of the color value, 2. the green dimension of the color value, and 3. the blue
     *           dimension of the color value
     * @since API version 1
     * @deprecated Use {@link #color()} instead.
     */
    addColorObserver(callback: ColorValueChangedCallback): void;

    /**
     * Plays a note with the key of the note lane and the provided velocity parameter.
     *
     * @param velocity
     *           the velocity the note should be played with
     * @since API version 1
     */
    play(velocity: number): void;
  }

  // source: com/bitwig/extension/controller/api/NotificationSettings.java

  /**
   * Bitwig Studio supports automatic visual feedback from controllers that shows up as popup notifications. For
   * example when the selected track or the current device preset was changed on the controller, these
   * notifications are shown, depending on the configuration.
   *
   * It depends both on the users preference and the capabilities of the controller hardware if a certain
   * notification should be shown. This interface provides functions for enabling/disabling the various kinds of
   * automatic notifications from the hardware point of view. Typically, controllers that include an advanced
   * display don't need to show many notifications additionally on screen. For other controllers that do not
   * include a display it might be useful to show all notifications. By default all notifications are disabled.
   *
   * In addition, the user can enable or disable all notifications the have been enabled using this interface in
   * the preferences dialog of Bitwig Studio.
   *
   * @since API version 1
   */
  interface NotificationSettings {
    /**
     * Returns an object that reports if user notifications are enabled and that allows to enable/disable user
     * notifications from the control surface. If user notifications are disabled, no automatic notifications
     * will be shown in the Bitwig Studio user interface. If user notifications are enabled, all automatic
     * notifications will be shown that are enabled using the methods of this interface.
     *
     * @return a boolean value object
     * @since API version 1
     */
    getUserNotificationsEnabled(): SettableBooleanValue;

    /**
     * Specifies if user notification related to selection changes should be shown. Please note that this
     * setting only applies when user notifications are enabled in general, otherwise no notification are
     * shown. By default this setting is `false`.
     *
     * @param shouldShowNotifications
     *           `true` in case selection notifications should be shown, `false` otherwise.
     * @see #getUserNotificationsEnabled()
     * @since API version 1
     */
    setShouldShowSelectionNotifications(shouldShowNotifications: boolean): void;

    /**
     * Specifies if user notification related to selection changes should be shown. Please note that this
     * setting only applies when user notifications are enabled in general, otherwise no notification are
     * shown. By default this setting is `false`.
     *
     * @param shouldShowNotifications
     *           `true` in case selection notifications should be shown, `false` otherwise.
     * @see #getUserNotificationsEnabled()
     * @since API version 1
     */
    setShouldShowChannelSelectionNotifications(
      shouldShowNotifications: boolean
    ): void;

    /**
     * Specifies if user notification related to selection changes should be shown. Please note that this
     * setting only applies when user notifications are enabled in general, otherwise no notification are
     * shown. By default this setting is `false`.
     *
     * @param shouldShowNotifications
     *           `true` in case selection notifications should be shown, `false` otherwise.
     * @see #getUserNotificationsEnabled()
     * @since API version 1
     */
    setShouldShowTrackSelectionNotifications(
      shouldShowNotifications: boolean
    ): void;

    /**
     * Specifies if user notification related to selection changes should be shown. Please note that this
     * setting only applies when user notifications are enabled in general, otherwise no notification are
     * shown. By default this setting is `false`.
     *
     * @param shouldShowNotifications
     *           `true` in case selection notifications should be shown, `false` otherwise.
     * @see #getUserNotificationsEnabled()
     * @since API version 1
     */
    setShouldShowDeviceSelectionNotifications(
      shouldShowNotifications: boolean
    ): void;

    /**
     * Specifies if user notification related to selection changes should be shown. Please note that this
     * setting only applies when user notifications are enabled in general, otherwise no notification are
     * shown. By default this setting is `false`.
     *
     * @param shouldShowNotifications
     *           `true` in case selection notifications should be shown, `false` otherwise.
     * @see #getUserNotificationsEnabled()
     * @since API version 1
     */
    setShouldShowDeviceLayerSelectionNotifications(
      shouldShowNotifications: boolean
    ): void;

    /**
     * Specifies if user notification related to selection changes should be shown. Please note that this
     * setting only applies when user notifications are enabled in general, otherwise no notification are
     * shown.
     *
     * @param shouldShowNotifications
     *           `true` in case selection notifications should be shown, `false` otherwise.
     * @see #getUserNotificationsEnabled()
     * @since API version 1
     */
    setShouldShowPresetNotifications(shouldShowNotifications: boolean): void;

    /**
     * Specifies if user notification related to selection changes should be shown. Please note that this
     * setting only applies when user notifications are enabled in general, otherwise no notification are
     * shown. By default this setting is `false`.
     *
     * @param shouldShowNotifications
     *           `true` in case selection notifications should be shown, `false` otherwise.
     * @see #getUserNotificationsEnabled()
     * @since API version 1
     */
    setShouldShowMappingNotifications(shouldShowNotifications: boolean): void;

    /**
     * Specifies if user notification related to selection changes should be shown. Please note that this
     * setting only applies when user notifications are enabled in general, otherwise no notification are
     * shown. By default this setting is `false`.
     *
     * @param shouldShowNotifications
     *           `true` in case selection notifications should be shown, `false` otherwise.
     * @see #getUserNotificationsEnabled()
     * @since API version 1
     */
    setShouldShowValueNotifications(shouldShowNotifications: boolean): void;
  }

  // source: com/bitwig/extension/controller/api/ObjectArrayValue.java

  import ObjectValueChangedCallback = com.bitwig.extension.callback.ObjectValueChangedCallback;

  /**
   * @since API version 2
   */
  interface ObjectArrayValue<ObjectType>
    extends Value<ObjectValueChangedCallback<ObjectType[]>> {
    /**
     * @since API version 2
     */
    get(): ObjectType[];

    /**
     * @since API version 2
     */
    get(index: number): ObjectType;
  }

  // source: com/bitwig/extension/controller/api/ObjectProxy.java

  /**
   * Interface for an object that acts as a proxy for the actual object in Bitwig Studio (for example a track, a
   * device etc).
   *
   * @since API version 2
   */
  interface ObjectProxy extends Subscribable {
    /**
     * Returns a value object that indicates if the object being proxied exists, or if it has content.
     */
    exists(): BooleanValue;

    /**
     * Creates a {@link BooleanValue} that determines this proxy is considered equal to another proxy. For this
     * to be the case both proxies need to be proxying the same target object.
     *
     * @since API version 3
     */
    createEqualsValue(other: ObjectProxy): BooleanValue;
  }

  // source: com/bitwig/extension/controller/api/Parameter.java

  /**
   * Instances of this interface represent ranged parameters that can be controlled with automation in Bitwig
   * Studio.
   *
   * @since API version 1
   */
  interface Parameter extends SettableRangedValue, ObjectProxy {
    /**
     * Gets the current value of this parameter.
     *
     * @since API version 2
     */
    value(): SettableRangedValue;

    /**
     * Gets the modulated value of this parameter.
     *
     * @since API version 2
     */
    modulatedValue(): RangedValue;

    /**
     * The name of the parameter.
     * @since API version 2
     */
    name(): StringValue;

    /**
     * Adds an observer which reports changes to the name of the automated parameter. The callback will get
     * called at least once immediately after calling this method for reporting the current name.
     *
     * @param maxChars
     *           maximum length of the string sent to the observer
     * @param textWhenUnassigned
     *           the default text to use
     * @param callback
     *           a callback function that receives a single string parameter
     * @since API version 1
     * @deprecated
     */
    addNameObserver(
      maxChars: number,
      textWhenUnassigned: string,
      callback: StringValueChangedCallback
    ): void;

    /**
     * Adds an observer which sends a formatted text representation of the value whenever the value changes.
     * The callback will get called at least once immediately after calling this method for reporting the
     * current state.
     *
     * @param maxChars
     *           maximum length of the string sent to the observer
     * @param textWhenUnassigned
     *           the default text to use
     * @param callback
     *           a callback function that receives a single string parameter
     * @since API version 1
     * @deprecated Use {@link #value()#displayedValue()} instead
     */
    addValueDisplayObserver(
      maxChars: number,
      textWhenUnassigned: string,
      callback: StringValueChangedCallback
    ): void;

    /**
     * Resets the value to its default.
     *
     * @since API version 1
     */
    reset(): void;

    /**
     * Touch (or un-touch) the value for automation recording.
     *
     * @param isBeingTouched
     *           `true` for touching, `false` for un-touching
     * @since API version 1
     */
    touch(isBeingTouched: boolean): void;

    /**
     * Specifies if this value should be indicated as mapped in Bitwig Studio, which is visually shown as
     * colored dots or tinting on the parameter controls.
     *
     * @param shouldIndicate
     *           `true` in case visual indications should be shown in Bitwig Studio, `false` otherwise
     * @since API version 1
     */
    setIndication(shouldIndicate: boolean): void;

    /**
     * Specifies a label for the mapped hardware parameter as shown in Bitwig Studio, for example in menu items
     * for learning controls.
     *
     * @param label
     *           the label to be shown in Bitwig Studio
     * @since API version 1
     */
    setLabel(label: string): void;

    /**
     * Restores control of this parameter to automation playback.
     *
     * @since API version 1
     */
    restoreAutomationControl(): void;
  }

  // source: com/bitwig/extension/controller/api/ParameterBank.java

  /**
   * Defines a bank of parameters.
   *
   * @since API version 2
   */
  interface ParameterBank {
    /**
     * Gets the number of slots that these remote controls have.
     *
     * @since API version 2
     */
    getParameterCount(): number;

    /**
     * Returns the parameter at the given index within the bank.
     *
     * @param indexInBank
     *           the parameter index within this bank. Must be in the range [0..getParameterCount()-1].
     * @return the requested parameter
     * @since API version 2
     */
    getParameter(indexInBank: number): Parameter;
  }

  // source: com/bitwig/extension/controller/api/PinnableCursor.java

  /**
   * Interface that defines a cursor that can be "pinned".
   *
   * @since API version 2
   */
  interface PinnableCursor extends Cursor {
    /**
     * Determines if this cursor is currently pinned or not. If the cursor is pinned then it won't follow the
     * selection the user makes in the application.
     *
     * @since API version 2
     */
    isPinned(): SettableBooleanValue;
  }

  // source: com/bitwig/extension/controller/api/PinnableCursorDevice.java

  /**
   * Cursor that can be pinned to the current device or follow the selection.
   *
   * @since API version 2
   */
  interface PinnableCursorDevice extends CursorDevice, PinnableCursor {}

  // source: com/bitwig/extension/controller/api/PlayingNote.java

  /**
   * @since API version 2
   */
  interface PlayingNote {
    /**
     * @since API version 2
     */
    pitch(): number;

    /**
     * @since API version 2
     */
    velocity(): number;
  }

  // source: com/bitwig/extension/controller/api/PlayingNoteArrayValue.java

  /**
   * @since API version 2
   */
  interface PlayingNoteArrayValue extends ObjectArrayValue<PlayingNote> {
    isNotePlaying(note: number): boolean;
  }

  // source: com/bitwig/extension/controller/api/PopupBrowser.java

  /**
   * Object that represents the popup browser in Bitwig Studio.
   *
   * @since API version 2
   */
  interface PopupBrowser extends ObjectProxy {
    /**
     * The title of the popup browser.
     *
     * @since API version 2
     */
    title(): StringValue;

    /**
     * Value that reports the possible content types that can be inserted by the popup browser. These are
     * represented by the tabs in Bitwig Studio's popup browser.
     *
     * (e.g "Device", "Preset", "Sample" etc.)
     *
     * @since API version 2
     */
    contentTypeNames(): StringArrayValue;

    /**
     * Value that represents the selected content type.
     *
     * @since API version 2
     */
    selectedContentTypeName(): StringValue;

    /**
     * Value that represents the index of the selected content type within the content types supported.
     *
     * @since API version 2
     */
    selectedContentTypeIndex(): SettableIntegerValue;

    /**
     * The smart collections column of the browser.
     *
     * @since API version 2
     */
    smartCollectionColumn(): BrowserFilterColumn;

    /**
     * The location column of the browser.
     *
     * @since API version 2
     */
    locationColumn(): BrowserFilterColumn;

    /**
     * The device column of the browser.
     *
     * @since API version 2
     */
    deviceColumn(): BrowserFilterColumn;

    /**
     * The category column of the browser.
     *
     * @since API version 2
     */
    categoryColumn(): BrowserFilterColumn;

    /**
     * The tag column of the browser.
     *
     * @since API version 2
     */
    tagColumn(): BrowserFilterColumn;

    /**
     * The device type column of the browser.
     *
     * @since API version 2
     */
    deviceTypeColumn(): BrowserFilterColumn;

    /**
     * The file type column of the browser.
     *
     * @since API version 2
     */
    fileTypeColumn(): BrowserFilterColumn;

    /**
     * The creator column of the browser.
     *
     * @since API version 2
     */
    creatorColumn(): BrowserFilterColumn;

    /**
     * Column that represents the results of the search.
     *
     * @since API version 2
     */
    resultsColumn(): BrowserResultsColumn;

    /**
     * Value that indicates if the browser is able to audition material in place while browsing.
     *
     * @since API version 2
     */
    canAudition(): BooleanValue;

    /**
     * Value that decides if the browser is currently auditioning material in place while browsing or not.
     *
     * @since API version 2
     */
    shouldAudition(): SettableBooleanValue;

    /**
     * Selects the next file.
     *
     * @since API version 2
     */
    selectNextFile(): void;

    /**
     * Selects the previous file.
     *
     * @since API version 2
     */
    selectPreviousFile(): void;

    /**
     * Selects the first file.
     *
     * @since API version 2
     */
    selectFirstFile(): void;

    /**
     * Selects the last file.
     *
     * @since API version 2
     */
    selectLastFile(): void;

    /**
     * Cancels the popup browser.
     *
     * @since API version 2
     */
    cancel(): void;

    /**
     * Commits the selected item in the popup browser.
     *
     * @since API version 2
     */
    commit(): void;
  }

  // source: com/bitwig/extension/controller/api/Preferences.java

  /**
   * This interface is used to store custom controller settings into the Bitwig Studio preferences. The settings
   * are shown to the user in the controller preferences dialog of Bitwig Studio.
   *
   * @since API version 1
   */
  interface Preferences extends Settings {}

  // source: com/bitwig/extension/controller/api/PresetBrowsingSession.java

  /**
   * Instances of this interface are used for browsing presets, including access to all filter columns and the
   * result column as shown in the 'Presets' tab of Bitwig Studio's contextual browser window.
   *
   * @see BrowsingSession
   * @since API version 1
   * @deprecated Use {@link PopupBrowser} instead.
   */
  interface PresetBrowsingSession extends BitwigBrowsingSession {
    /**
     * Returns the category filter as shown in the category column of the browser.
     *
     * @return the requested category filter object.
     * @since API version 1
     */
    getCategoryFilter(): BrowserFilterColumn;

    /**
     * Returns the preset type filter as shown in the category column of the browser.
     *
     * @return the requested preset type filter object.
     * @since API version 1
     */
    getPresetTypeFilter(): BrowserFilterColumn;

    /**
     * Returns the file type filter as shown in the category column of the browser.
     *
     * @return the requested file type filter object.
     * @since API version 1
     */
    getFileTypeFilter(): BrowserFilterColumn;
  }

  // source: com/bitwig/extension/controller/api/PrimaryDevice.java

  enum ChainLocation {
    FIRST = 0,
    NEXT = 1,
    PREVIOUS = 2,
    LAST = 3,
  }

  enum DeviceType {
    ANY = 0,
  }

  /**
   * A special kind of device that represents the primary device of a track.
   *
   * @since API version 1
   * @deprecated This is now replaced by creating a named {@link CursorDevice}.
   * @see Track#createCursorDevice(String)
   */
  interface PrimaryDevice extends Device {
    /**
     * Makes the device with the given type and location the new primary device.
     *
     * @param deviceType
     *           the requested device type of the new primary device
     * @param chainLocation
     *           the requested chain location of the new primary device
     * @since API version 1
     */
    switchToDevice(deviceType: DeviceType, chainLocation: ChainLocation): void;

    /**
     * Registers an observer that reports if navigation to another device with the provided characteristics is
     * possible.
     *
     * @param deviceType
     *           the requested device type of the new primary device
     * @param chainLocation
     *           the requested chain location of the new primary device
     * @param callback
     *           a callback function the receives a single boolean parameter
     * @since API version 1
     * @deprecated This method never did anything. Please do not use.
     */
    addCanSwitchToDeviceObserver(
      deviceType: DeviceType,
      chainLocation: ChainLocation,
      callback: BooleanValueChangedCallback
    ): void;
  }

  // source: com/bitwig/extension/controller/api/Project.java

  /**
   * An interface for representing the current project.
   *
   * @since API version 1
   */
  interface Project extends ObjectProxy {
    /**
     * Returns an object that represents the root track group of the active Bitwig Studio project.
     *
     * @return the root track group of the currently active project
     * @since API version 1
     */
    getRootTrackGroup(): Track;

    /**
     * Returns an object that represents the top level track group as shown in the arranger/mixer of the active
     * Bitwig Studio project.
     *
     * @return the shown top level track group of the currently active project
     * @since API version 1
     */
    getShownTopLevelTrackGroup(): Track;

    /**
     * Creates a new scene (using an existing empty scene if possible) from the clips that are currently
     * playing in the clip launcher.
     *
     * @since API version 1
     */
    createSceneFromPlayingLauncherClips(): void;
  }

  // source: com/bitwig/extension/controller/api/RangedValue.java

  /** Instances of this interface represent numeric values that have an upper and lower limit.
   *
   * @since API version 1
   * */
  // @ts-ignore
  interface RangedValue extends Value<DoubleValueChangedCallback> {
    /**
     * The current value normalized between 0..1 where 0 represents the minimum value and 1 the maximum.
     *
     * @since API version 2
     */
    get(): number;

    /**
     * Gets the current value.
     *
     * @since API version 2
     */
    getRaw(): number;

    /**
     * Value that represents a formatted text representation of the value whenever the value changes.
     *
     * @since API version 2
     */
    displayedValue(): StringValue;

    /**
     * Adds an observer which receives the normalized value of the parameter as an integer number within the
     * range [0..range-1].
     *
     * @param range
     *           the range used to scale the value when reported to the callback
     * @param callback
     *           a callback function that receives a single double parameter
     * @since API version 1
     */
    addValueObserver(
      range: number,
      callback: IntegerValueChangedCallback
    ): void;

    /**
     * Add an observer which receives the internal raw of the parameter as floating point.
     *
     * @param callback
     *           a callback function that receives a single numeric parameter with double precision.
     * @since API version 1
     */
    addRawValueObserver(callback: DoubleValueChangedCallback): void;
  }

  // source: com/bitwig/extension/controller/api/RemoteConnection.java

  import NoArgsCallback = com.bitwig.extension.callback.NoArgsCallback;

  /**
   * Instances of this interface are reported to the supplied script callback when connecting to a remote TCP
   * socket via {@link ControllerHost#connectToRemoteHost}.
   *
   * @see {@link ControllerHost#connectToRemoteHost}
   * @since API version 1
   */
  interface RemoteConnection {
    /**
     * Disconnects from the remote host.
     *
     * @since API version 1
     */
    disconnect(): void;

    /**
     * Registers a callback function that gets called when the connection gets lost or disconnected.
     *
     * @param callback
     *           a callback function that doesn't receive any parameters
     * @since API version 1
     */
    setDisconnectCallback(callback: NoArgsCallback): void;

    /**
     * Sets the callback used for receiving data.
     *
     * The remote connection needs a header for each message sent to it containing a 32-bit big-endian integer
     * saying how big the rest of the message is. The data delivered to the script will not include this
     * header.
     *
     * @param callback
     *           a callback function with the signature `(byte[] data)`
     * @since API version 1
     */
    setReceiveCallback(callback: DataReceivedCallback): void;

    /**
     * Sends data to the remote host.
     *
     * @param data
     *           the byte array containing the data to be sent. When creating a numeric byte array in
     *           JavaScript, the byte values must be signed (in the range -128..127).
     * @throws IOException
     * @since API version 1
     */
    send(data: number): void;
  }

  // source: com/bitwig/extension/controller/api/RemoteControl.java

  /**
   * Represents a remote control in Bitwig Studio.
   *
   * @sice API version 2
   */
  interface RemoteControl extends Parameter {
    name(): SettableStringValue;
  }

  // source: com/bitwig/extension/controller/api/RemoteControlsPage.java

  /**
   * Represents a page of remote controls in a device.
   *
   * @since API version 2
   */
  interface RemoteControlsPage extends ParameterBank {
    getParameter(indexInBank: number): RemoteControl;

    /*
   @since API version 4
    */
    getName(): StringValue;
  }

  // source: com/bitwig/extension/controller/api/RemoteSocket.java

  /**
   * Instances of this interface represent a TCP socket that other network clients can connect to, typically
   * created by calling {@link ControllerHost#createRemoteConnection}.
   *
   * @see {@link ControllerHost#createRemoteConnection}
   * @since API version 1
   */
  interface RemoteSocket {
    /**
     * Sets a callback which receives a remote connection object for each incoming connection.
     *
     * @param callback
     *           a callback function which receives a single {@link RemoteConnection} argument
     * @since API version 1
     */
    setClientConnectCallback(callback: ConnectionEstablishedCallback): void;

    /**
     * Gets the actual port used for the remote socket, which might differ from the originally requested port
     * when calling {@link ControllerHost#createRemoteConnection(String name, int port)} in case the requested port was
     * already used.
     *
     * @return the actual port used for the remote socket
     * @since API version 1
     */
    getPort(): number;
  }

  // source: com/bitwig/extension/controller/api/SampleBrowsingSession.java

  /**
   * Instances of this interface are used for browsing samples, including access to all filter columns and the
   * result column as shown in the 'Samples' tab of Bitwig Studio's contextual browser window.
   *
   * @see BrowsingSession
   * @since API version 1
   * @deprecated Use {@link PopupBrowser} instead.
   */
  interface SampleBrowsingSession extends BrowsingSession {
    /**
     * Returns the file type filter as shown in the category column of the browser.
     *
     * @return the requested file type filter object.
     * @since API version 1
     */
    getFileTypeFilter(): BrowserFilterColumn;
  }

  // source: com/bitwig/extension/controller/api/Scene.java

  /**
   * Instances of this interface represent scenes in Bitwig Studio.
   *
   * @since API version 1
   */
  interface Scene extends ClipLauncherSlotOrScene {
    /**
     * Returns an object that provides access to the name of the scene.
     *
     * @return a string value object that represents the scene name.
     * @since API version 1
     * @deprecated Use {@link #name()} instead.
     */
    getName(): SettableStringValue;

    /**
     * Returns an object that provides access to the name of the scene.
     *
     * @return a string value object that represents the scene name.
     * @since API version 2
     */
    name(): SettableStringValue;

    /**
     * Value that reports the number of clips in the scene.
     *
     * @since API version 2
     */
    clipCount(): IntegerValue;

    /**
     * Registers an observer that reports the number of clips in the scene.
     *
     * @param callback
     *           a callback function that receives a single integer parameter
     * @since API version 1
     * @deprecated Use {@link #clipCount()}.addValueObserver(callback).
     */
    addClipCountObserver(callback: IntegerValueChangedCallback): void;

    /**
     * Registers an observer that reports the position of the scene within the list of Bitwig Studio scenes.
     *
     * @param callback
     *           a callback function that receives a single integer parameter
     * @since API version 1
     * @deprecated Use {@link #sceneIndex()} instead.
     */
    addPositionObserver(callback: IntegerValueChangedCallback): void;

    /**
     * Registers an observer that reports if the scene is selected in Bitwig Studio.
     *
     * @param callback
     *           a callback function that takes a single boolean parameter.
     * @since API version 1
     */
    addIsSelectedInEditorObserver(callback: BooleanValueChangedCallback): void;

    /**
     * Selects the scene in Bitwig Studio.
     *
     * @since API version 1
     */
    selectInEditor(): void;

    /**
     * Makes the scene visible in the Bitwig Studio user interface.
     *
     * @since API version 1
     */
    showInEditor(): void;
  }

  // source: com/bitwig/extension/controller/api/SceneBank.java

  /**
   * A scene bank provides access to a range of scenes in Bitwig Studio. Instances of scene bank are configured
   * with a fixed number of scenes and represent an excerpt of a larger list of scenes. Various methods are
   * provided for scrolling to different sections of the scene list. It basically acts like a window moving over
   * the list of underlying scenes.
   *
   * To receive an instance of scene bank call
   * {@link com.bitwig.extension.controller.api.ControllerHost#createSceneBank}.
   *
   * @see {@link com.bitwig.extension.controller.api.ControllerHost#createSceneBank}
   * @since API version 1
   */
  interface SceneBank extends ClipLauncherSlotOrSceneBank<Scene> {
    /**
     * Returns the scene at the given index within the bank.
     *
     * @param indexInBank
     *           the scene index within this bank, not the index within the list of all Bitwig Studio scenes.
     *           Must be in the range [0..sizeOfBank-1].
     * @return the requested scene object
     * @since API version 1
     */
    getScene(indexInBank: number): Scene;

    /**
     * Scrolls the scenes one page up.
     *
     * @since API version 1
     * @deprecated Use {@link #scrollPageBackwards()} instead.
     */
    scrollPageUp(): void;

    /**
     * Scrolls the scenes one page down.
     *
     * @since API version 1
     * @deprecated Use {@link #scrollPageForwards()} instead.
     */
    scrollPageDown(): void;

    /**
     * Scrolls the scenes one scene up.
     *
     * @since API version 1
     * @deprecated Use {@link #scrollBackwards()} instead.
     */
    scrollUp(): void;

    /**
     * Scrolls the scenes one scene down.
     *
     * @since API version 1
     * @deprecated Use {@link #scrollForwards()} instead.
     */
    scrollDown(): void;

    /**
     * Makes the scene with the given position visible in the track bank.
     *
     * @param position
     *           the position of the scene within the underlying full list of scenes
     * @since API version 1
     * @deprecated Use {@link #scrollPosition()} instead.
     */
    scrollTo(position: number): void;

    /**
     * Registers an observer that reports the current scene scroll position.
     *
     * @param callback
     *           a callback function that takes a single integer parameter
     * @param valueWhenUnassigned
     *           the default value that gets reports when the track bank is not yet connected to a Bitwig
     *           Studio document
     * @since API version 1
     * @deprecated Use {@link #scrollPosition()} instead
     */
    addScrollPositionObserver(
      callback: IntegerValueChangedCallback,
      valueWhenUnassigned: number
    ): void;

    /**
     * Registers an observer that reports if the scene window can be scrolled further up.
     *
     * @param callback
     *           a callback function that takes a single boolean parameter
     * @since API version 1
     * @deprecated Use {@link #canScrollBackwards()} instead.
     */
    addCanScrollUpObserver(callback: BooleanValueChangedCallback): void;

    /**
     * Registers an observer that reports if the scene window can be scrolled further down.
     *
     * @param callback
     *           a callback function that takes a single boolean parameter
     * @since API version 1
     * @deprecated Use {@link #canScrollForwards()} instead.
     */
    addCanScrollDownObserver(callback: BooleanValueChangedCallback): void;

    /**
     * Registers an observer that reports the underlying total scene count (not the number of scenes available
     * in the bank window).
     *
     * @param callback
     *           a callback function that receives a single integer parameter
     * @since API version 1
     * @deprecated Use {@link #itemCount()} instead.
     */
    addSceneCountObserver(callback: IntegerValueChangedCallback): void;

    /**
     * Launches the scene with the given bank index.
     *
     * @param indexInWindow
     *           the scene index within the bank, not the position of the scene withing the underlying full
     *           list of scenes.
     * @since API version 1
     */
    launchScene(indexInWindow: number): void;
  }

  // source: com/bitwig/extension/controller/api/Scrollable.java

  /**
   * Interface for something that can be scrolled.
   *
   * @since API version 2
   */
  interface Scrollable {
    /**
     * Value that reports the current scene scroll position.
     *
     * @since API version 2
     */
    scrollPosition(): SettableIntegerValue;

    /**
     * Scrolls by a number of steps.
     *
     * @param amount
     *           The number of steps to scroll by (positive is forwards and negative is backwards).
     */
    scrollBy(amount: number): void;

    /**
     * Scrolls forwards by one step. This is the same as calling {@link #scrollBy(int)} with 1
     *
     * @since API version 2
     */
    scrollForwards(): void;

    /**
     * Scrolls forwards by one step. This is the same as calling {@link #scrollBy(int)} with -1
     *
     * @since API version 2
     */
    scrollBackwards(): void;

    /**
     * Scrolls by a number of pages.
     *
     * @param amount
     *           The number of pages to scroll by (positive is forwards and negative is backwards).
     */
    scrollByPages(amount: number): void;

    /**
     * Scrolls forwards by one page.
     *
     * @since API version 2
     */
    scrollPageForwards(): void;

    /**
     * Scrolls backwards by one page.
     *
     * @since API version 2
     */
    scrollPageBackwards(): void;

    /**
     * Value that reports if it is possible to scroll the bank backwards or not.
     *
     * @since API version 2
     */
    canScrollBackwards(): BooleanValue;

    /**
     * Value that reports if it is possible to scroll the bank forwards or not.
     *
     * @since API version 2
     */
    canScrollForwards(): BooleanValue;
  }

  // source: com/bitwig/extension/controller/api/Send.java

  interface Send extends Parameter {
    /**
     * Value that reports the color of the channel that this send sends to.
     *
     * @since API version 2
     */
    sendChannelColor(): SettableColorValue;
  }

  // source: com/bitwig/extension/controller/api/SendBank.java

  interface SendBank extends Bank<Send> {}

  // source: com/bitwig/extension/controller/api/SettableBeatTimeValue.java

  interface SettableBeatTimeValue extends BeatTimeValue, SettableDoubleValue {
    /**
     * The same as the set method.
     *
     * @deprecated This is included only for backwards compatibility with API version 1 where this inherited
     *             from {@link RangedValue} instead of {@link DoubleValue}.
     */
    setRaw(value: number): void;

    /**
     * The same as the inc method.
     *
     * @deprecated This is included only for backwards compatibility with API version 1 where this inherited
     *             from {@link RangedValue} instead of {@link DoubleValue}.
     */
    incRaw(amount: number): void;
  }

  // source: com/bitwig/extension/controller/api/SettableBooleanValue.java

  /**
   * Instances of this interface represent boolean values.
   *
   * @since API version 1
   */
  interface SettableBooleanValue extends BooleanValue {
    /**
     * Sets the internal value.
     *
     * @param value
     *           the new boolean value.
     * @since API version 1
     */
    set(value: boolean): void;

    /**
     * Toggles the current state. In case the current value is `false`, the new value will be `true` and the
     * other way round.
     *
     * @since API version 1
     */
    toggle(): void;
  }

  // source: com/bitwig/extension/controller/api/SettableColorValue.java

  interface SettableColorValue extends ColorValue {
    /**
     * Sets the internal value.
     *
     * @since API version 2
     */
    set(red: number, green: number, blue: number): void;

    /**
     * Sets the internal value.
     *
     * @since API version 5
     */
    set(red: number, green: number, blue: number, alpha: number): void;
  }

  // source: com/bitwig/extension/controller/api/SettableDoubleValue.java

  interface SettableDoubleValue extends DoubleValue {
    /**
     * Sets the internal value.
     *
     * @param value
     *           the new integer value.
     * @since API version 1
     */
    set(value: number): void;

    /**
     * Increases/decrease the internal value by the given amount.
     *
     * @param amount
     *           the integer amount to increase
     * @since API version 1
     */
    inc(amount: number): void;
  }

  // source: com/bitwig/extension/controller/api/SettableEnumValue.java

  /**
   * Instances of this interface represent enumeration values. Enum values work similar to string values, but
   * are limited to a fixed set of value options.
   *
   * @since API version 1
   */
  interface SettableEnumValue extends EnumValue {
    /**
     * Sets the value to the enumeration item with the given name.
     *
     * @param name
     *           the name of the new enum item
     * @since API version 1
     */
    set(value: string): void;
  }

  // source: com/bitwig/extension/controller/api/SettableIntegerValue.java

  /**
   * Instances of this interface represent integer values.
   *
   * @since API version 1
   */
  interface SettableIntegerValue extends IntegerValue {
    /**
     * Sets the internal value.
     *
     * @param value
     *           the new integer value.
     * @since API version 1
     */
    set(value: number): void;

    /**
     * Increases/decrease the internal value by the given amount.
     *
     * @param amount
     *           the integer amount to increase
     * @since API version 1
     */
    inc(amount: number): void;
  }

  // source: com/bitwig/extension/controller/api/SettableRangedValue.java

  /**
   * Instances of this interface represent numeric values that have an upper and lower limit.
   *
   * @since API version 1
   */
  interface SettableRangedValue extends RangedValue {
    /**
     * Sets the value in an absolute fashion as a value between 0 .. 1 where 0 represents the minimum value and
     * 1 the maximum. The value may not be set immediately if the user has configured a take over strategy for
     * the controller.
     *
     * @param value
     *           absolute value [0 .. 1]
     * @since API version 2
     */
    set(value: number): void;

    /**
     * Sets the value in an absolute fashion as a value between 0 .. 1 where 0 represents the minimum value and
     * 1 the maximum. The value change is applied immediately and does not care about what take over mode the
     * user has selected. This is useful if the value does not need take over (e.g. a motorized slider).
     *
     * @param value
     *           absolute value [0 .. 1]
     *
     * @since API version 4
     */
    setImmediately(value: number): void;

    /**
     * Sets the value in an absolute fashion. The value will be scaled according to the given resolution.
     *
     * Typically the resolution would be specified as the amount of steps the hardware control provides (for
     * example 128) and just pass the integer value as it comes from the MIDI device. The host application will
     * take care of scaling it.
     *
     * @param value
     *           integer number in the range [0 .. resolution-1]
     * @param resolution
     *           the resolution used for scaling @ if passed-in parameters are null
     * @since API version 1
     */
    set(value: number, resolution: number): void;

    /**
     * Increments or decrements the value by a normalized amount assuming the whole range of the value is 0 ..
     * 1. For example to increment by 10% you would use 0.1 as the increment.
     *
     * @since API version 2
     */
    inc(increment: number): void;

    /**
     * Increments or decrements the value according to the given increment and resolution parameters.
     *
     * Typically the resolution would be specified as the amount of steps the hardware control provides (for
     * example 128) and just pass the integer value as it comes from the MIDI device. The host application will
     * take care of scaling it.
     *
     * @param increment
     *           the amount that the current value is increased by
     * @param resolution
     *           the resolution used for scaling
     * @since API version 1
     */
    inc(increment: number, resolution: number): void;

    /**
     * Set the internal (raw) value.
     *
     * @param value
     *           the new value with double precision. Range is undefined.
     * @since API version 1
     */
    setRaw(value: number): void;

    /**
     * Increments / decrements the internal (raw) value.
     *
     * @param delta
     *           the amount that the current internal value get increased by.
     * @since API version 1
     */
    incRaw(delta: number): void;
  }

  // source: com/bitwig/extension/controller/api/SettableStringArrayValue.java

  /**
   * @since API version 2
   */
  interface SettableStringArrayValue extends StringArrayValue {
    /**
     * Sets the internal value.
     *
     * @param value
     *           the new value.
     * @since API version 2
     */
    set(value: string[]): void;
  }

  // source: com/bitwig/extension/controller/api/SettableStringValue.java

  /**
   * Instances of this interface implement the {@link Value} interface for string values.
   *
   * @since API version 1
   */
  interface SettableStringValue extends StringValue {
    /**
     * Sets the value object to the given string.
     *
     * @param value
     *           the new value string
     * @since API version 1
     */
    set(value: string): void;
  }

  // source: com/bitwig/extension/controller/api/Setting.java

  /**
   * A common base interface for labeled and categorized settings.
   *
   * @since API version 1
   */
  interface Setting {
    /**
     * Returns the category name of the setting.
     *
     * @return a string value containing the category name
     * @since API version 1
     */
    getCategory(): string;

    /**
     * Returns the label text of the setting.
     *
     * @return a string value containing the label text
     * @since API version 1
     */
    getLabel(): string;

    /**
     * Marks the settings as enabled in Bitwig Studio. By default the setting is enabled.
     *
     * @since API version 1
     */
    enable(): void;

    /**
     * Marks the settings as disabled in Bitwig Studio. By default the setting is enabled.
     *
     * @since API version 1
     */
    disable(): void;

    /**
     * Shows the setting in Bitwig Studio. By default the setting is shown.
     *
     * @since API version 1
     */
    show(): void;

    /**
     * Hides the setting in Bitwig Studio. By default the setting is shown.
     *
     * @since API version 1
     */
    hide(): void;
  }

  // source: com/bitwig/extension/controller/api/Settings.java

  import Color = com.bitwig.extension.api.Color;

  /**
   * This interface builds the foundation for storing custom settings in Bitwig Studio documents or in the
   * Bitwig Studio preferences.
   *
   * @since API version 1
   */
  interface Settings {
    /**
     * Returns a signal setting object, which is shown a push button with the given label in Bitwig Studio.
     *
     * @param label
     *           the name of the setting, must not be `null`
     * @param category
     *           the name of the category, may not be `null`
     * @param action
     *           the action string as displayed on the related Bitwig Studio button, must not be `null`
     * @return the object that encapsulates the requested signal
     * @ @since API version 1
     */
    getSignalSetting(label: string, category: string, action: string): Signal;

    /**
     * Returns a numeric setting that is shown a number field in Bitwig Studio.
     *
     * @param label
     *           the name of the setting, must not be `null`
     * @param category
     *           the name of the category, may not be `null`
     * @param minValue
     *           the minimum value that the user is allowed to enter
     * @param maxValue
     *           the minimum value that the user is allowed to enter
     * @param stepResolution
     *           the step resolution used for the number field
     * @param unit
     *           the string that should be used to display the unit of the number
     * @param initialValue
     *           the initial numeric value of the setting
     * @return the object that encapsulates the requested numeric setting
     * @ @since API version 1
     */
    getNumberSetting(
      label: string,
      category: string,
      minValue: number,
      maxValue: number,
      stepResolution: number,
      unit: string,
      initialValue: number
    ): SettableRangedValue;

    /**
     * Returns an enumeration setting that is shown either as a chooser or as a button group in Bitwig Studio,
     * depending on the number of provided options.
     *
     * @param label
     *           the name of the setting, must not be `null`
     * @param category
     *           the name of the category, may not be `null`
     * @param options
     *           the string array that defines the allowed options for the button group or chooser
     * @param initialValue
     *           the initial string value, must be one of the items specified with the option argument
     * @return the object that encapsulates the requested enum setting
     * @ @since API version 1
     */
    getEnumSetting(
      label: string,
      category: string,
      options: string[],
      initialValue: string
    ): SettableEnumValue;

    /**
     * Returns a textual setting that is shown as a text field in the Bitwig Studio user interface.
     *
     * @param label
     *           the name of the setting, must not be `null`
     * @param category
     *           the name of the category, may not be `null`
     * @param numChars
     *           the maximum number of character used for the text value
     * @param initialText
     *           the initial text value of the setting
     * @return the object that encapsulates the requested string setting
     * @ @since API version 1
     */
    getStringSetting(
      label: string,
      category: string,
      numChars: number,
      initialText: string
    ): SettableStringValue;

    /**
     * Returns a color setting that is shown in the Bitwig Studio user interface.
     *
     * @param label
     *           the name of the setting, must not be `null`
     * @param category
     *           the name of the category, may not be `null`
     * @param initialColor
     *           the initial color value of the setting
     * @return the object that encapsulates the requested string setting
     * @since API version 5
     */
    getColorSetting(
      label: string,
      category: string,
      initialColor: Color
    ): SettableColorValue;
  }

  // source: com/bitwig/extension/controller/api/Signal.java

  /**
   * A generic interface used to implement actions or events that are not associated with a value.
   *
   * @since API version 1
   */
  interface Signal {
    /**
     * Registers an observer that gets notified when the signal gets fired.
     *
     * @param callback
     *           a callback function that does not receive any argument.
     * @since API version 1
     */
    addSignalObserver(callback: NoArgsCallback): void;

    /**
     * Fires the action or event represented by the signal object.
     *
     * @since API version 1
     */
    fire(): void;
  }

  // source: com/bitwig/extension/controller/api/SoloValue.java

  /**
   * Instances of this interface represent the state of a solo button.
   *
   * @since API version 1
   */
  // @ts-ignore
  interface SoloValue extends SettableBooleanValue {
    /**
     * Toggles the current solo state.
     *
     * @param exclusive
     *           specifies if solo on other channels should be disabled automatically ('true') or not
     *           ('false').
     * @since API version 1
     */
    toggle(exclusive: boolean): void;
  }

  // source: com/bitwig/extension/controller/api/SourceSelector.java

  /**
   * Instance of this class represent sources selectors in Bitwig Studio, which are shown as choosers in the
   * user interface and contain entries for either note inputs or audio inputs or both.
   *
   * The most prominent source selector in Bitwig Studio is the one shown in the track IO section, which can be
   * accessed via the API by calling {@link Track#getSourceSelector()}.
   *
   * @since API version 1
   */
  interface SourceSelector extends ObjectProxy {
    /**
     * Returns an object that indicates if the source selector has note inputs enabled.
     *
     * @return a boolean value object
     * @since API version 1
     * @deprecated Use {@link #hasNoteInputSelected()} instead.
     */
    getHasNoteInputSelected(): SettableBooleanValue;

    /**
     * Returns an object that indicates if the source selector has note inputs enabled.
     *
     * @return a boolean value object
     * @since API version 5
     */
    hasNoteInputSelected(): SettableBooleanValue;

    /**
     * Returns an object that indicates if the source selector has audio inputs enabled.
     *
     * @return a boolean value object
     * @since API version 1
     * @deprecated Use {@link #hasAudioInputSelected()} instead.
     */
    getHasAudioInputSelected(): SettableBooleanValue;

    /**
     * Returns an object that indicates if the source selector has audio inputs enabled.
     *
     * @return a boolean value object
     * @since API version 5
     */
    hasAudioInputSelected(): SettableBooleanValue;
  }

  // source: com/bitwig/extension/controller/api/StringArrayValue.java

  /**
   * @since API version 2
   */
  // @ts-ignore
  interface StringArrayValue extends ObjectArrayValue<string> {
    /**
     * Gets the current value.
     *
     * @since API version 2
     */
    get(): string[];
  }

  // source: com/bitwig/extension/controller/api/StringValue.java

  interface StringValue extends Value<StringValueChangedCallback> {
    /**
     * Gets the current value.
     *
     * @since API version 2
     */
    get(): string;

    /**
     * Gets the current value and tries to intelligently limit it to the supplied length in the best way
     * possible.
     *
     * @since API version 2
     */
    getLimited(maxLength: number): string;
  }

  // source: com/bitwig/extension/controller/api/Subscribable.java

  /**
   * Interface for an object that can be 'subscribed' or not. A subscribed object will notify any observers when
   * changes occur to it. When it is unsubscribed the observers will no longer be notified. A driver can use
   * this to say which objects it is interested in and which ones it is not (for example in one mode the driver
   * may not be interested in track meters) at runtime. This allows the driver to improve efficiency by only
   * getting notified about changes that are really relevant to it. By default a driver is subscribed to
   * everything.
   *
   * @since API version 2
   */
  interface Subscribable {
    /**
     * Determines if this object is currently 'subscribed'. In the subscribed state it will notify any
     * observers registered on it.
     */
    isSubscribed(): boolean;

    /** Sets whether the driver currently considers this object 'active' or not. */
    setIsSubscribed(value: boolean): void;

    /**
     * Subscribes the driver to this object. This is equivalent to calling {@link #setIsSubscribed(boolean)}
     * with true.
     */
    subscribe(): void;

    /**
     * Unsubscribes the driver from this object. This is equivalent to calling {@link #setIsSubscribed(boolean)}
     * with false.
     */
    unsubscribe(): void;
  }

  // source: com/bitwig/extension/controller/api/TimeSignatureValue.java

  /**
   * Instances of this interface represent time signature values.
   *
   * @since API version 1
   */
  interface TimeSignatureValue extends Value<StringValueChangedCallback> {
    /**
     * Gets the current value.
     *
     * @since API version 2
     */
    get(): string;

    /**
     * Updates the time signature according to the given string.
     *
     * @param name
     *           a textual representation of the new time signature value, formatted as
     *           `numerator/denominator[, ticks]`
     * @since API version 1
     */
    set(name: string): void;

    /**
     * Returns an object that provides access to the time signature numerator.
     *
     * @return an integer value object that represents the time signature numerator.
     * @since API version 1
     * @deprecated Use {@link #numerator()} instead.
     */
    getNumerator(): SettableIntegerValue;

    /**
     * Returns an object that provides access to the time signature numerator.
     *
     * @return an integer value object that represents the time signature numerator.
     * @since API version 5
     */
    numerator(): SettableIntegerValue;

    /**
     * Returns an object that provides access to the time signature denominator.
     *
     * @return an integer value object that represents the time signature denominator.
     * @since API version 1
     * @deprecated Use {@link #denominator()} instead.
     */
    getDenominator(): SettableIntegerValue;

    /**
     * Returns an object that provides access to the time signature denominator.
     *
     * @return an integer value object that represents the time signature denominator.
     * @since API version 5
     */
    denominator(): SettableIntegerValue;

    /**
     * Returns an object that provides access to the time signature tick subdivisions.
     *
     * @return an integer value object that represents the time signature ticks.
     * @since API version 1
     * @deprecated Use {@link #ticks()} instead.
     */
    getTicks(): SettableIntegerValue;

    /**
     * Returns an object that provides access to the time signature tick subdivisions.
     *
     * @return an integer value object that represents the time signature ticks.
     * @since API version 5
     */
    ticks(): SettableIntegerValue;
  }

  // source: com/bitwig/extension/controller/api/Track.java

  /**
   * Instances of this interface represent tracks in Bitwig Studio.
   *
   * @since API version 1
   */
  interface Track extends Channel {
    /**
     * Value that reports the position of the track within the list of Bitwig Studio tracks.
     *
     * @since API version 2
     */
    position(): IntegerValue;

    /**
     * Registers an observer that reports the position of the track within the list of Bitwig Studio tracks.
     *
     * @param callback
     *           a callback function that receives a single integer parameter
     * @since API version 1
     * @deprecated Use {@link #position()} instead.
     */
    addPositionObserver(callback: IntegerValueChangedCallback): void;

    /**
     * Returns an object that can be used to access the clip launcher slots of the track.
     *
     * @return an object that represents the clip launcher slots of the track
     * @since API version 1
     * @deprecated Use {@link #clipLauncherSlotBank()} instead.
     */
    getClipLauncherSlots(): ClipLauncherSlotBank;

    /**
     * Returns an object that can be used to access the clip launcher slots of the track.
     *
     * @return an object that represents the clip launcher slots of the track
     * @since API version 2
     */
    clipLauncherSlotBank(): ClipLauncherSlotBank;

    /**
     * @deprecated Use {@link #getClipLauncherSlots()} instead.
     * @since API version 1
     */
    getClipLauncher(): ClipLauncherSlotBank;

    /**
     * Registers an observer that reports if the clip launcher slots are queued for stop.
     *
     * @param callback
     *           a callback function that receives a single boolean argument.
     * @since API version 1
     * @deprecated Use {@link #isQueuedForStop()} instead.
     */
    addIsQueuedForStopObserver(callback: BooleanValueChangedCallback): void;

    /**
     * Returns an object that provides access to the arm state of the track.
     *
     * @return a boolean value object
     * @since API version 1
     * @deprecated Use {@link #arm()} instead.
     */
    getArm(): SettableBooleanValue;

    /**
     * Returns an object that provides access to the arm state of the track.
     *
     * @return a boolean value object
     * @since API version 5
     */
    arm(): SettableBooleanValue;

    /**
     * Returns an object that provides access to the monitoring state of the track.
     *
     * @return a boolean value object
     * @since API version 1
     * @deprecated Use {@link #monitor()} instead.
     */
    getMonitor(): SettableBooleanValue;

    /**
     * Returns an object that provides access to the monitoring state of the track.
     *
     * @return a boolean value object
     * @since API version 5
     */
    monitor(): SettableBooleanValue;

    /**
     * Returns an object that provides access to the auto-monitoring state of the track.
     *
     * @return a boolean value object
     * @since API version 1
     * @deprecated Use {@link #autoMonitor()} instead.
     */
    getAutoMonitor(): SettableBooleanValue;

    /**
     * Returns an object that provides access to the auto-monitoring state of the track.
     *
     * @return a boolean value object
     * @since API version 5
     */
    autoMonitor(): SettableBooleanValue;

    /**
     * Returns an object that provides access to the cross-fade mode of the track.
     *
     * @return an enum value object that has three possible states: "A", "B", or "AB"
     * @since API version 1
     * @deprecated Use {@link #crossFadeMode()} instead.
     */
    getCrossFadeMode(): SettableEnumValue;

    /**
     * Returns an object that provides access to the cross-fade mode of the track.
     *
     * @return an enum value object that has three possible states: "A", "B", or "AB"
     * @since API version 5
     */
    crossFadeMode(): SettableEnumValue;

    /**
     * Value that reports if this track is currently stopped. When a track is stopped it is not playing content
     * from the arranger or clip launcher.
     *
     * @since API version 2
     */
    isStopped(): BooleanValue;

    /**
     * Returns a value object that provides access to the clip launcher playback state of the track.
     *
     * @return a boolean value object that indicates if the clip launcher is stopped for this track
     * @see #getIsMatrixQueuedForStop()
     * @since API version 1
     * @deprecated Use {@link #isStopped()} instead.
     */
    getIsMatrixStopped(): BooleanValue;

    /**
     * Returns a value object that provides access to the clip launcher's queue-for-stop state on this track. A
     * clip is considered to be queued for stop when playback has been requested to be stopped on that clip,
     * but the playback has not stopped yet due to the current launch quantization settings.
     *
     * @return a boolean value object that indicates if the clip launcher slots have been queued for stop
     * @see #getIsMatrixStopped()
     * @since API version 1
     * @deprecated Use {@link #isQueuedForStop()} instead.
     */
    getIsMatrixQueuedForStop(): BooleanValue;

    /**
     * Value that reports if the clip launcher slots are queued for stop.
     *
     * @since API version 2
     */
    isQueuedForStop(): BooleanValue;

    /**
     * Returns the source selector for the track, which is shown in the IO section of the track in Bitwig
     * Studio and lists either note or audio sources or both depending on the track type.
     *
     * @return a source selector object
     * @since API version 1
     * @deprecated Use {@link #sourceSelector()} instead.
     */
    getSourceSelector(): SourceSelector;

    /**
     * Returns the source selector for the track, which is shown in the IO section of the track in Bitwig
     * Studio and lists either note or audio sources or both depending on the track type.
     *
     * @return a source selector object
     * @since API version 5
     */
    sourceSelector(): SourceSelector;

    /**
     * Stops playback of the track.
     *
     * @since API version 1
     */
    stop(): void;

    /**
     * Calling this method causes the arrangement sequencer to take over playback.
     *
     * @since API version 1
     */
    returnToArrangement(): void;

    /**
     * Updates the name of the track.
     *
     * @param name
     *           the new track name
     * @since API version 1
     */
    setName(name: string): void;

    /**
     * Registers an observer that reports names for note key values on this track. The track might provide
     * special names for certain keys if it contains instruments that support that features, such as the Bitwig
     * Drum Machine.
     *
     * @param callback
     *           a callback function that receives two arguments: 1. the key value in the range [0..127], and
     *           2. the name string
     * @since API version 1
     */
    addPitchNamesObserver(callback: IndexedStringValueChangedCallback): void;

    /**
     * Plays a note on the track with a default duration and the given key and velocity.
     *
     * @param key
     *           the key value of the played note
     * @param velocity
     *           the velocity of the played note
     * @since API version 1
     */
    playNote(key: number, velocity: number): void;

    /**
     * Starts playing a note on the track with the given key and velocity.
     *
     * @param key
     *           the key value of the played note
     * @param velocity
     *           the velocity of the played note
     * @since API version 1
     */
    startNote(key: number, velocity: number): void;

    /**
     * Stops playing a currently played note.
     *
     * @param key
     *           the key value of the playing note
     * @param velocity
     *           the note-off velocity
     * @since API version 1
     */
    stopNote(key: number, velocity: number): void;

    /**
     * Sends a MIDI message to the hardware device.
     *
     * @param status
     *           the status byte of the MIDI message
     * @param data1
     *           the data1 part of the MIDI message
     * @param data2
     *           the data2 part of the MIDI message
     * @ @since API version 2
     */
    sendMidi(status: number, data1: number, data2: number): void;

    /**
     * Value that reports the track type. Possible reported track types are `Group`, `Instrument`, `Audio`,
     * `Hybrid`, `Effect` or `Master`.
     *
     * @since API version 2
     */
    trackType(): StringValue;

    /**
     * Registers an observer that reports the track type. Possible reported track types are `Group`,
     * `Instrument`, `Audio`, `Hybrid`, `Effect` or `Master`.
     *
     * @param numChars
     *           the maximum number of characters used for the reported track type
     * @param textWhenUnassigned
     *           the default text that gets reported when the track is not yet associated with a Bitwig Studio
     *           track.
     * @param callback
     *           a callback function that receives a single track type parameter (string).
     * @since API version 1
     * @deprecated Use {@link #trackType()} instead.
     */
    addTrackTypeObserver(
      numChars: number,
      textWhenUnassigned: string,
      callback: StringValueChangedCallback
    ): void;

    /**
     * Value that reports if the track may contain child tracks, which is the case for group tracks.
     *
     * @since API version 2
     */
    isGroup(): BooleanValue;

    /**
     * Registers an observer that reports if the track may contain child tracks, which is the case for group
     * tracks.
     *
     * @param callback
     *           a callback function that receives a single boolean parameter.
     * @since API version 1
     * @deprecated Use {@link #isGroup()} instead.
     */
    addIsGroupObserver(callback: BooleanValueChangedCallback): void;

    /**
     * Returns an object that indicates if the track may contain notes.
     *
     * @return a boolean value object
     * @since API version 1
     * @deprecated Use {@link #canHoldNoteData()} instead.
     */
    getCanHoldNoteData(): SettableBooleanValue;

    /**
     * Returns an object that indicates if the track may contain notes.
     *
     * @return a boolean value object
     * @since API version 5
     */
    canHoldNoteData(): SettableBooleanValue;

    /**
     * Returns an object that indicates if the track may contain audio events.
     *
     * @return a boolean value object
     * @since API version 1
     * @deprecated Use {@link #canHoldAudioData()} instead.
     */
    getCanHoldAudioData(): SettableBooleanValue;

    /**
     * Returns an object that indicates if the track may contain audio events.
     *
     * @return a boolean value object
     * @since API version 5
     */
    canHoldAudioData(): SettableBooleanValue;

    /**
     * Returns an object that provides access to the cursor item of the track's device selection as shown in
     * the Bitwig Studio user interface.
     *
     * @return the requested device selection cursor object
     * @ @since API version 1
     */
    createCursorDevice(): CursorDevice;

    /**
     * Creates a named device selection cursor that is independent from the device selection in the Bitwig
     * Studio user interface, assuming the name parameter is not null. When `name` is `null` the result is
     * equal to calling {@link Track#createCursorDevice}.
     *
     * @param name
     *           the name of the custom device selection cursor, for example "Primary", or `null` to refer to
     *           the device selection cursor in the arranger cursor track as shown in the Bitwig Studio user
     *           interface.
     * @return the requested device selection cursor object
     * @ @see Track#createCursorDevice
     * @see #createCursorDevice(String name, int numSends)
     * @since API version 1
     */
    createCursorDevice(name: string): CursorDevice;

    /**
     * Creates a named device selection cursor that is independent from the device selection in the Bitwig
     * Studio user interface, assuming the name parameter is not null. When `name` is `null` the result is
     * equal to calling {@link Track#createCursorDevice}.
     *
     * @param name
     *           the name of the custom device selection cursor, for example "Primary", or `null` to refer to
     *           the device selection cursor in the arranger cursor track as shown in the Bitwig Studio user
     *           interface.
     * @param numSends
     *           the number of sends that are simultaneously accessible in nested channels.
     * @return the requested device selection cursor object
     * @ @see Track#createCursorDevice
     * @since API version 1
     */
    createCursorDevice(name: string, numSends: number): CursorDevice;

    /**
     * Gets the channels primary device.
     *
     * @return an object that provides access to the channels primary device.
     * @deprecated Use {@link #createCursorDevice(String) createCursorDevice("Primary")} instead.
     * @since API version 1
     */
    getPrimaryDevice(): Device;

    /**
     * @deprecated Use {@link #createCursorDevice(String) createCursorDevice("Primary")} instead.
     * @since API version 1
     */
    getPrimaryInstrument(): Device;

    /**
     * Returns a track bank with the given number of child tracks, sends and scenes. The track bank will only
     * have content if the connected track is a group track.<br/>
     *
     * A track bank can be seen as a fixed-size window onto the list of tracks in the connected track group
     * including their sends and scenes, that can be scrolled in order to access different parts of the track
     * list. For example a track bank configured for 8 tracks can show track 1-8, 2-9, 3-10 and so on.<br/>
     *
     * The idea behind the `bank pattern` is that hardware typically is equipped with a fixed amount of channel
     * strips or controls, for example consider a mixing console with 8 channels, but Bitwig Studio documents
     * contain a dynamic list of tracks, most likely more tracks than the hardware can control simultaneously.
     * The track bank returned by this function provides a convenient interface for controlling which tracks
     * are currently shown on the hardware.<br/>
     *
     * Creating a track bank using this method will consider all tracks in the document, including effect
     * tracks and the master track. Use {@link #createMainTrackBank} or {@link #createEffectTrackBank} in case
     * you are only interested in tracks of a certain kind.
     *
     * @param numTracks
     *           the number of child tracks spanned by the track bank
     * @param numSends
     *           the number of sends spanned by the track bank
     * @param numScenes
     *           the number of scenes spanned by the track bank
     * @param hasFlatTrackList
     *           specifies whether the track bank should operate on a flat list of all nested child tracks or
     *           only on the direct child tracks of the connected group track.
     * @return an object for bank-wise navigation of tracks, sends and scenes
     * @see #createMainTrackBank
     * @see #createEffectTrackBank
     * @since API version 1
     */
    createTrackBank(
      numTracks: number,
      numSends: number,
      numScenes: number,
      hasFlatTrackList: boolean
    ): TrackBank;

    /**
     * Returns a track bank with the given number of child tracks, sends and scenes. Only audio tracks,
     * instrument tracks and hybrid tracks are considered. The track bank will only have content if the
     * connected track is a group track. For more information about track banks and the `bank pattern` in
     * general, see the documentation for {@link #createTrackBank}.
     *
     * @param numTracks
     *           the number of child tracks spanned by the track bank
     * @param numSends
     *           the number of sends spanned by the track bank
     * @param numScenes
     *           the number of scenes spanned by the track bank
     * @param hasFlatTrackList
     *           specifies whether the track bank should operate on a flat list of all nested child tracks or
     *           only on the direct child tracks of the connected group track.
     * @return an object for bank-wise navigation of tracks, sends and scenes
     * @see #createTrackBank
     * @see #createEffectTrackBank
     * @since API version 1
     */
    createMainTrackBank(
      numTracks: number,
      numSends: number,
      numScenes: number,
      hasFlatTrackList: boolean
    ): TrackBank;

    /**
     * Returns a track bank with the given number of child effect tracks and scenes. Only effect tracks are
     * considered. The track bank will only have content if the connected track is a group track. For more
     * information about track banks and the `bank pattern` in general, see the documentation for
     * {@link #createTrackBank}.
     *
     * @param numTracks
     *           the number of child tracks spanned by the track bank
     * @param numScenes
     *           the number of scenes spanned by the track bank
     * @param hasFlatTrackList
     *           specifies whether the track bank should operate on a flat list of all nested child tracks or
     *           only on the direct child tracks of the connected group track.
     * @return an object for bank-wise navigation of tracks, sends and scenes
     * @see #createTrackBank
     * @see #createMainTrackBank
     * @since API version 1
     */
    createEffectTrackBank(
      numTracks: number,
      numScenes: number,
      hasFlatTrackList: boolean
    ): TrackBank;

    /**
     * Returns an object that represents the master track of the connected track group. The returned object
     * will only have content if the connected track is a group track.
     *
     * @param numScenes
     *           the number of scenes for bank-wise navigation of the master tracks clip launcher slots.
     * @return an object representing the master track of the connected track group.
     * @since API version 1
     */
    createMasterTrack(numScenes: number): MasterTrack;

    /**
     * Returns a bank of sibling tracks with the given number of tracks, sends and scenes. For more information
     * about track banks and the `bank pattern` in general, see the documentation for {@link #createTrackBank}.
     *
     * @param numTracks
     *           the number of child tracks spanned by the track bank
     * @param numSends
     *           the number of sends spanned by the track bank
     * @param numScenes
     *           the number of scenes spanned by the track bank
     * @param shouldIncludeEffectTracks
     *           specifies whether effect tracks should be included
     * @param shouldIncludeMasterTrack
     *           specifies whether the master should be included
     * @return an object for bank-wise navigation of sibling tracks
     * @see #createTrackBank
     * @since API version 1
     */
    createSiblingsTrackBank(
      numTracks: number,
      numSends: number,
      numScenes: number,
      shouldIncludeEffectTracks: boolean,
      shouldIncludeMasterTrack: boolean
    ): TrackBank;
  }

  // source: com/bitwig/extension/controller/api/TrackBank.java

  /**
   * A track bank provides access to a range of tracks and their scenes (clip launcher slots) in Bitwig Studio.
   * Instances of track bank are configured with a fixed number of tracks and scenes and represent an excerpt of
   * a larger list of tracks and scenes. Various methods are provided for scrolling to different sections of the
   * track/scene list. It basically acts like a 2-dimensional window moving over the grid of tracks and scenes.
   *
   * To receive an instance of track bank that supports all kinds of tracks call {@link ControllerHost#createTrackBank}.
   * Additional methods are provided in the {@link ControllerHost} interface to create track banks that include only main
   * tracks ({@link ControllerHost#createMainTrackBank}) or only effect tracks ({@link ControllerHost#createEffectTrackBank}).
   *
   * @see {@link ControllerHost#createTrackBank}
   * @see {@link ControllerHost#createMainTrackBank}
   * @see {@link ControllerHost#createEffectTrackBank}
   * @since API version 1
   */
  interface TrackBank extends ChannelBank<Track> {
    /**
     * @deprecated use {@link #getChannel(int)} instead.
     * @since API version 1
     */
    getTrack(indexInBank: number): Track;

    /**
     * Returns the track at the given index within the bank.
     *
     * @param indexInBank
     *           the track index within this bank, not the index within the list of all Bitwig Studio tracks.
     *           Must be in the range [0..sizeOfBank-1].
     * @return the requested track object
     * @since API version 1
     * @deprecated Use {@link #getItemAt(int)} instead.
     */
    getChannel(indexInBank: number): Track;

    /**
     * @deprecated use {@link #setChannelScrollStepSize(int)} instead.
     */
    setTrackScrollStepSize(stepSize: number): void;

    /**
     * @deprecated use {@link #scrollChannelsPageUp()} instead.
     */
    scrollTracksPageUp(): void;

    /**
     * @deprecated use {@link #scrollChannelsPageDown()} instead.
     */
    scrollTracksPageDown(): void;

    /**
     * @deprecated use {@link #scrollChannelsUp()} instead.
     */
    scrollTracksUp(): void;

    /**
     * @deprecated use {@link #scrollChannelsDown()} instead.
     */
    scrollTracksDown(): void;

    /**
     * @deprecated use {@link #scrollToChannel(int)} instead.
     */
    scrollToTrack(position: number): void;

    /**
     * @deprecated use {@link #addChannelScrollPositionObserver(Callable, int)} instead.
     */
    addTrackScrollPositionObserver(
      callback: IntegerValueChangedCallback,
      valueWhenUnassigned: number
    ): void;

    /**
     * {@link SceneBank} that represents a view on the scenes in this {@link TrackBank}.
     *
     * @since API version 2
     */
    sceneBank(): SceneBank;

    /**
     * Scrolls the scenes one page up.
     *
     * @since API version 1
     * @deprecated Use {@link #sceneBank()} instead.
     */
    scrollScenesPageUp(): void;

    /**
     * Scrolls the scenes one page down.
     *
     * @since API version 1
     */
    scrollScenesPageDown(): void;

    /**
     * Scrolls the scenes one step up.
     *
     * @since API version 1
     * @deprecated Use {@link #sceneBank()} instead.
     */
    scrollScenesUp(): void;

    /**
     * Scrolls the scenes one step down.
     *
     * @since API version 1
     * @deprecated Use {@link #sceneBank()} instead.
     */
    scrollScenesDown(): void;

    /**
     * Makes the scene with the given position visible in the track bank.
     *
     * @param position
     *           the position of the scene within the underlying full list of scenes
     * @since API version 1
     * @deprecated Use {@link #sceneBank()} instead.
     */
    scrollToScene(position: number): void;

    /**
     * Registers an observer that reports the current scene scroll position.
     *
     * @param callback
     *           a callback function that takes a single integer parameter
     * @param valueWhenUnassigned
     *           the default value that gets reports when the track bank is not yet connected to a Bitwig
     *           Studio document
     * @since API version 1
     * @deprecated Use {@link #sceneBank()} instead.
     */
    addSceneScrollPositionObserver(
      callback: IntegerValueChangedCallback,
      valueWhenUnassigned: number
    ): void;

    /**
     * @deprecated use {@link #canScrollChannelsUp()} instead.
     */
    addCanScrollTracksUpObserver(callback: BooleanValueChangedCallback): void;

    /**
     * @deprecated use {@link #canScrollChannelsDown()} instead.
     */
    addCanScrollTracksDownObserver(callback: BooleanValueChangedCallback): void;

    /**
     * Registers an observer that reports if the scene window can be scrolled further up.
     *
     * @param callback
     *           a callback function that takes a single boolean parameter
     * @since API version 1
     *
     * @deprecated use {@link #canScrollScenesUp()} instead.
     */
    addCanScrollScenesUpObserver(callback: BooleanValueChangedCallback): void;

    /**
     * Registers an observer that reports if the scene window can be scrolled further down.
     *
     * @param callback
     *           a callback function that takes a single boolean parameter
     * @since API version 1
     *
     * @deprecated use {@link #canScrollScenesDown()} instead.
     */
    addCanScrollScenesDownObserver(callback: BooleanValueChangedCallback): void;

    /**
     * Registers an observer that reports the underlying total scene count (not the number of scenes available
     * in the bank window).
     *
     * @param callback
     *           a callback function that receives a single integer parameter
     * @since API version 1
     * @deprecated Use {@link #sceneBank().itemCount().addValueObserver()}
     */
    addSceneCountObserver(callback: IntegerValueChangedCallback): void;

    /**
     * Returns an object that provides access to the clip launcher scenes of the track bank.
     *
     * @return an object that provides access to the clip launcher scenes of the track bank.
     * @since API version 1
     * @deprecated Use {@link #sceneBank()} instead.
     */
    getClipLauncherScenes(): ClipLauncherSlotOrSceneBank;

    /**
     * Launches the scene with the given bank index.
     *
     * @param indexInWindow
     *           the scene index within the bank, not the position of the scene withing the underlying full
     *           list of scenes.
     * @since API version 1
     * @deprecated Use {@link #sceneBank()} instead.
     */
    launchScene(indexInWindow: number): void;

    /**
     * Causes this bank to follow the supplied cursor. When the cursor moves to a new item the bank will be
     * scrolled so that the cursor is within the bank, if possible.
     *
     * @param cursorTrack
     *           The {@link CursorTrack} that this bank should follow.
     *
     * @since API version 2
     */
    followCursorTrack(cursorTrack: CursorTrack): void;
  }

  // source: com/bitwig/extension/controller/api/Transport.java

  /**
   * An interface representing the transport section in Bitwig Studio.
   *
   * @since API version 1
   */
  interface Transport extends ObjectProxy {
    /**
     * Starts playback in the Bitwig Studio transport.
     *
     * @since API version 1
     */
    play(): void;

    /**
     * Stops playback in the Bitwig Studio transport.
     *
     * @since API version 1
     */
    stop(): void;

    /**
     * Toggles the transport playback state between playing and stopped.
     *
     * @since API version 1
     */
    togglePlay(): void;

    /**
     * When the transport is stopped, calling this function starts transport playback, otherwise the transport
     * is first stopped and the playback is restarted from the last play-start position.
     *
     * @since API version 1
     */
    restart(): void;

    /**
     * Starts recording in the Bitwig Studio transport.
     *
     * @since API version 1
     */
    record(): void;

    /**
     * Rewinds the Bitwig Studio transport to the beginning of the arrangement.
     *
     * @since API version 1
     */
    rewind(): void;

    /**
     * Calling this function is equivalent to pressing the fast forward button in the Bitwig Studio transport.
     *
     * @since API version 1
     */
    fastForward(): void;

    /**
     * When calling this function multiple times, the timing of those calls gets evaluated and causes
     * adjustments to the project tempo.
     *
     * @since API version 1
     */
    tapTempo(): void;

    /**
     * Value that reports if the Bitwig Studio transport is playing.
     *
     * @since API version 2
     */
    isPlaying(): SettableBooleanValue;

    /**
     * Registers an observer that reports if the Bitwig Studio transport is playing.
     *
     * @param callback
     *           a callback function that receives a single boolean argument (`true` if playing, `false`
     *           otherwise).
     * @since API version 1
     * @deprecated Use {@link #isPlaying()} instead.
     */
    addIsPlayingObserver(callback: BooleanValueChangedCallback): void;

    /**
     * Value that reports if the Bitwig Studio transport is recording.
     *
     * @since API version 2
     */
    isArrangerRecordEnabled(): SettableBooleanValue;

    /**
     * Registers an observer that reports if the Bitwig Studio transport is recording.
     *
     * @param callback
     *           a callback function that receives a single boolean argument (`true` if recording, `false`
     *           otherwise).
     * @since API version 1
     * @deprecated Use {@link #isArrangerRecordEnabled()} instead.
     */
    addIsRecordingObserver(callback: BooleanValueChangedCallback): void;

    /**
     * Value that reports if overdubbing is enabled in Bitwig Studio.
     *
     * @since API version 2
     */
    isArrangerOverdubEnabled(): SettableBooleanValue;

    /**
     * Registers an observer that reports if over-dubbing is enabled in Bitwig Studio.
     *
     * @param callback
     *           a callback function that receives a single boolean argument (`true` if over-dubbing is
     *           enabled, `false` otherwise).
     * @since API version 1
     * @deprecated Use {@link #isArrangerOverdubEnabled()} instead.
     */
    addOverdubObserver(callback: BooleanValueChangedCallback): void;

    /**
     * Value reports if clip launcher overdubbing is enabled in Bitwig Studio.
     *
     * @since API version 2
     */
    isClipLauncherOverdubEnabled(): SettableBooleanValue;

    /**
     * Registers an observer that reports if clip launcher over-dubbing is enabled in Bitwig Studio.
     *
     * @param callback
     *           a callback function that receives a single boolean argument (`true` if clip launcher
     *           over-dubbing is enabled, `false` otherwise).
     * @since API version 1
     * @deprecated Use {@link #isClipLauncherOverdubEnabled()} instead.
     */
    addLauncherOverdubObserver(callback: BooleanValueChangedCallback): void;

    /**
     * Value that reports the current automation write mode. Possible values are `"latch"`, `"touch"` or
     * `"write"`.
     *
     * @since API version 2
     */
    automationWriteMode(): SettableEnumValue;

    /**
     * Registers an observer that reports the current automation write mode.
     *
     * @param callback
     *           a callback function that receives a single string argument. Possible values are `"latch"`,
     *           `"touch"` or `"write"`.
     * @since API version 1
     * @deprecated Use {@link #automationWriteMode()} instead.
     */
    addAutomationWriteModeObserver(callback: EnumValueChangedCallback): void;

    /**
     * Value that reports if automation write is currently enabled for the arranger.
     *
     * @since API version 2
     */
    isArrangerAutomationWriteEnabled(): SettableBooleanValue;

    /**
     * Registers an observer that reports if Bitwig Studio is currently writing arranger automation.
     *
     * @param callback
     *           a callback function that receives a single boolean argument (`true` when arranger automation
     *           write is enabled, `false` otherwise).
     * @since API version 1
     * @deprecated Use {@link #isArrangerAutomationWriteEnabled()} instead.
     */
    addIsWritingArrangerAutomationObserver(
      callback: BooleanValueChangedCallback
    ): void;

    /**
     * Value that reports if automation write is currently enabled on the clip launcher.
     *
     * @since API version 2
     */
    isClipLauncherAutomationWriteEnabled(): SettableBooleanValue;

    /**
     * Registers an observer that reports if Bitwig Studio is currently writing clip launcher automation.
     *
     * @param callback
     *           a callback function that receives a single boolean argument (`true` when clip launcher
     *           automation write is enabled, `false` otherwise).
     * @since API version 1
     * @deprecated Use {@link #isClipLauncherAutomationWriteEnabled()} instead.
     */
    addIsWritingClipLauncherAutomationObserver(
      callback: BooleanValueChangedCallback
    ): void;

    /**
     * Value that indicates if automation override is currently on.
     *
     * @since API version 2
     */
    isAutomationOverrideActive(): BooleanValue;

    /**
     * Registers an observer that reports if automation is overridden in Bitwig Studio.
     *
     * @param callback
     *           a callback function that receives a single boolean argument (`true` if overridden, `false`
     *           otherwise).
     * @since API version 1
     * @deprecated Use {@link #isAutomationOverrideActive()}.addValueObserver(callback)
     */
    addAutomationOverrideObserver(callback: BooleanValueChangedCallback): void;

    /**
     * Value that indicates if the loop is currently active or not.
     *
     * @since API version 2
     */
    isArrangerLoopEnabled(): SettableBooleanValue;

    /**
     * Registers an observer that reports if arranger looping is enabled in Bitwig Studio.
     *
     * @param callback
     *           a callback function that receives a single boolean argument (`true` when enabled, `false`
     *           otherwise).
     * @since API version 1
     * @deprecated Use {@link #isArrangerLoopEnabled()}.addValueObserver(callback)
     */
    addIsLoopActiveObserver(callback: BooleanValueChangedCallback): void;

    /**
     * Value that reports if punch-in is enabled in the Bitwig Studio transport.
     *
     * @since API version 2
     */
    isPunchInEnabled(): SettableBooleanValue;

    /**
     * Registers an observer that reports if punch-in is enabled in the Bitwig Studio transport.
     *
     * @param callback
     *           a callback function that receives a single boolean argument (`true` when punch-in is enabled,
     *           `false` otherwise).
     * @since API version 1
     * @deprecated Use {@link #isPunchInEnabled()} instead.
     */
    addPunchInObserver(callback: BooleanValueChangedCallback): void;

    /**
     * Value that reports if punch-in is enabled in the Bitwig Studio transport.
     *
     * @since API version 2
     */
    isPunchOutEnabled(): SettableBooleanValue;

    /**
     * Registers an observer that reports if punch-out is enabled in the Bitwig Studio transport.
     *
     * @param callback
     *           a callback function that receives a single boolean argument (`true` when punch-out is enabled,
     *           `false` otherwise).
     * @since API version 1
     * @deprecated Use {@link #isPunchOutEnabled()} instead.
     */
    addPunchOutObserver(callback: BooleanValueChangedCallback): void;

    /**
     * Value that reports if the metronome is enabled in Bitwig Studio.
     *
     * @since API version 2
     */
    isMetronomeEnabled(): SettableBooleanValue;

    /**
     * Registers an observer that reports if the metronome is enabled in Bitwig Studio.
     *
     * @param callback
     *           a callback function that receives a single boolean argument (`true` when the metronome is
     *           enabled, `false` otherwise).
     * @since API version 1
     * @deprecated Use {@link #isMetronomeEnabled()} instead.
     */
    addClickObserver(callback: BooleanValueChangedCallback): void;

    /**
     * Value that reports if the metronome has tick playback enabled.
     *
     * @since API version 2
     */
    isMetronomeTickPlaybackEnabled(): SettableBooleanValue;

    /**
     * Registers an observer that reports if the metronome has tick playback enabled.
     *
     * @param callback
     *           a callback function that receives a single boolean argument (`true` if metronome ticks, are
     *           enabled, `false` otherwise).
     * @since API version 1
     * @deprecated Use {@link #isMetronomeTickPlaybackEnabled()} instead.
     */
    addMetronomeTicksObserver(callback: BooleanValueChangedCallback): void;

    /**
     * Value that reports the metronome volume.
     *
     * @since API version 2
     */
    metronomeVolume(): SettableRangedValue;

    /**
     * Registers an observer that reports the metronome volume.
     *
     * @param callback
     *           a callback function that receives a single numeric argument.
     * @since API version 1
     * @deprecated Use {@link #metronomeVolume()} instead.
     */
    addMetronomeVolumeObserver(callback: DoubleValueChangedCallback): void;

    /**
     * Value that reports if the metronome is audible during pre-roll.
     *
     * @since API version 2
     */
    isMetronomeAudibleDuringPreRoll(): SettableBooleanValue;

    /**
     * Registers an observer that reports if the metronome is audible during pre-roll.
     *
     * @param callback
     *           a callback function that receives a single boolean argument.
     * @since API version 1
     * @deprecated Use {@link #isMetronomeAudibleDuringPreRoll()} instead.
     */
    addPreRollClickObserver(callback: BooleanValueChangedCallback): void;

    /**
     * Value that reports the current pre-roll setting. Possible values are `"none"`, `"one_bar"`,
     * `"two_bars"`, or `"four_bars"`.
     *
     * @since API version 2
     */
    preRoll(): SettableEnumValue;

    /**
     * Registers an observer that reports the current pre-roll setting.
     *
     * @param callback
     *           a callback function that receives a single string argument. Possible values are `"none"`,
     *           `"one_bar"`, `"two_bars"`, or `"four_bars"`.
     * @since API version 1
     * @deprecated Use {@link #preRoll()} instead.
     */
    addPreRollObserver(callback: EnumValueChangedCallback): void;

    /**
     * Toggles the enabled state of the arranger loop in Bitwig Studio.
     *
     * @since API version 1
     * @deprecated Use {@link #isArrangerLoopEnabled()} instead.
     */
    toggleLoop(): void;

    /**
     * Enables of disables the arranger loop according to the given parameter.
     *
     * @param isEnabled
     *           `true` to enable the arranger loop, `false` otherwise
     * @since API version 1
     * @deprecated Use {@link #isArrangerLoopEnabled()} instead.
     */
    setLoop(isEnabled: boolean): void;

    /**
     * Toggles the punch-in enabled state of the Bitwig Studio transport.
     *
     * @since API version 1
     * @deprecated Use {@link #isPunchInEnabled()} instead.
     */
    togglePunchIn(): void;

    /**
     * Toggles the punch-out enabled state of the Bitwig Studio transport.
     *
     * @since API version 1
     * @deprecated Use {@link #isPunchOutEnabled()} instead.
     */
    togglePunchOut(): void;

    /**
     * Toggles the metronome enabled state of the Bitwig Studio transport.
     *
     * @since API version 1
     * @deprecated Use {@link #isMetronomeEnabled()} instead.
     */
    toggleClick(): void;

    /**
     * Enables of disables the metronome according to the given parameter.
     *
     * @param isEnabled
     *           `true` to enable the metronome, `false` otherwise
     * @since API version 1
     * @deprecated Use {@link #isMetronomeEnabled()} instead.
     */
    setClick(isEnabled: boolean): void;

    /**
     * Toggles the enabled state of the metronome ticks.
     *
     * @since API version 1
     * @deprecated Use {@link #isMetronomeTickPlaybackEnabled()} instead.
     */
    toggleMetronomeTicks(): void;

    /**
     * Toggles the enabled state of the metronome during pre-roll.
     *
     * @since API version 1
     * @deprecated Use {@link #isMetronomeAudibleDuringPreRoll()} instead.
     */
    toggleMetronomeDuringPreRoll(): void;

    /**
     * Updates the transport pre-roll setting according to the given parameter.
     *
     * @param value
     *           the new pre-roll setting, either `"none"`, `"one_bar"`, `"two_bars"`, or `"four_bars"`.
     * @since API version 1
     * @deprecated Use {@link #preRoll()} instead.
     */
    setPreRoll(value: string): void;

    /**
     * Sets the metronome volume.
     *
     * @param amount
     *           the new metronome volume relative to the specified range. Values should be in the range
     *           [0..range-1].
     * @param range
     *           the range of the provided amount value
     * @since API version 1
     * @deprecated Use {@link #metronomeVolume()} instead.
     */
    setMetronomeValue(amount: number, range: number): void;

    /**
     * Toggles the over-dubbing enabled state of the Bitwig Studio transport.
     *
     * @since API version 1
     * @deprecated Use {@link #isArrangerOverdubEnabled()} instead.
     */
    toggleOverdub(): void;

    /**
     * Enables of disables arranger over-dubbing according to the given parameter.
     *
     * @param isEnabled
     *           `true` to enable over-dubbing, `false` otherwise
     * @since API version 1
     * @deprecated Use {@link #isArrangerOverdubEnabled()} instead.
     */
    setOverdub(isEnabled: boolean): void;

    /**
     * Toggles clip launcher overdubbing in Bitwig Studio.
     *
     * @since API version 1
     * @deprecated Use {@link #isClipLauncherOverdubEnabled()} instead.
     */
    toggleLauncherOverdub(): void;

    /**
     * Enables of disables clip launcher over-dubbing according to the given parameter.
     *
     * @param isEnabled
     *           `true` to enable the over-dubbing, `false` otherwise
     * @since API version 1
     * @deprecated Use {@link #isClipLauncherOverdubEnabled()} instead.
     */
    setLauncherOverdub(isEnabled: boolean): void;

    /**
     * Sets the automation write mode.
     *
     * @param mode
     *           the string that identifies the new automation write mode. Possible values are `"latch"`,
     *           `"touch"` or `"write"`.
     * @since API version 1
     * @deprecated Use {@link #automationWriteMode()} instead.
     */
    setAutomationWriteMode(mode: string): void;

    /**
     * Toggles the latch automation write mode in the Bitwig Studio transport.
     *
     * @since API version 1
     */
    toggleLatchAutomationWriteMode(): void;

    /**
     * Toggles the arranger automation write enabled state of the Bitwig Studio transport.
     *
     * @since API version 1
     */
    toggleWriteArrangerAutomation(): void;

    /**
     * Toggles the clip launcher automation write enabled state of the Bitwig Studio transport.
     *
     * @since API version 1
     */
    toggleWriteClipLauncherAutomation(): void;

    /**
     * Resets any automation overrides in Bitwig Studio.
     *
     * @since API version 1
     */
    resetAutomationOverrides(): void;

    /**
     * Switches playback to the arrangement sequencer on all tracks.
     *
     * @since API version 1
     */
    returnToArrangement(): void;

    /**
     * Returns an object that provides access to the project tempo.
     *
     * @return the requested tempo value object
     * @since API version 1
     * @deprecated Use {@link #tempo()} instead.
     */
    getTempo(): Parameter;

    /**
     * Returns an object that provides access to the project tempo.
     *
     * @return the requested tempo value object
     * @since API version 1
     */
    tempo(): Parameter;

    /**
     * Increases the project tempo value by the given amount, which is specified relative to the given range.
     *
     * @param amount
     *           the new tempo value relative to the specified range. Values should be in the range
     *           [0..range-1].
     * @param range
     *           the range of the provided amount value
     * @since API version 1
     */
    increaseTempo(amount: number, range: number): void;

    /**
     * Returns an object that provides access to the transport position in Bitwig Studio.
     *
     * @return a beat time object that represents the transport position
     * @since API version 1
     */
    getPosition(): SettableBeatTimeValue;

    /**
     * Sets the transport playback position to the given beat time value.
     *
     * @param beats
     *           the new playback position in beats
     * @since API version 1
     */
    setPosition(beats: number): void;

    /**
     * Increases the transport position value by the given number of beats, which is specified relative to the
     * given range.
     *
     * @param beats
     *           the beat time value that gets added to the current transport position. Values have double
     *           precision and can be positive or negative.
     * @param snap
     *           when `true` the actual new transport position will be quantized to the beat grid, when `false`
     *           the position will be increased exactly by the specified beat time
     * @since API version 1
     */
    incPosition(beats: number, snap: boolean): void;

    /**
     * Returns an object that provides access to the punch-in position in the Bitwig Studio transport.
     *
     * @return a beat time object that represents the punch-in position
     * @since API version 1
     */
    getInPosition(): SettableBeatTimeValue;

    /**
     * Returns an object that provides access to the punch-out position in the Bitwig Studio transport.
     *
     * @return a beat time object that represents the punch-out position
     * @since API version 1
     */
    getOutPosition(): SettableBeatTimeValue;

    /**
     * Returns an object that provides access to the cross-fader, used for mixing between A/B-channels as
     * specified on the Bitwig Studio tracks.
     *
     * @see Track#getCrossFadeMode()
     * @since API version 1
     * @deprecated Use {@link #crossfade()} instead.
     */
    getCrossfade(): Parameter;

    /**
     * Returns an object that provides access to the cross-fader, used for mixing between A/B-channels as
     * specified on the Bitwig Studio tracks.
     *
     * @see Track#getCrossFadeMode()
     * @since API version 5
     */
    crossfade(): Parameter;

    /**
     * Returns an object that provides access to the transport time signature.
     *
     * @return the time signature value object that represents the transport time signature.
     * @since API version 1
     * @deprecated Use {@link #timeSignature()} instead.
     */
    getTimeSignature(): TimeSignatureValue;

    /**
     * Returns an object that provides access to the transport time signature.
     *
     * @return the time signature value object that represents the transport time signature.
     * @since API version 5
     */
    timeSignature(): TimeSignatureValue;

    /**
     * Value that reports the current clip launcher post recording action. Possible values are `"off"`,
     * `"play_recorded"`, `"record_next_free_slot"`, `"stop"`, `"return_to_arrangement"`,
     * `"return_to_previous_clip"` or `"play_random"`.
     *
     * @since API version 2
     */
    clipLauncherPostRecordingAction(): SettableEnumValue;

    /**
     * Registers an observer that reports the current clip launcher post recording action.
     *
     * @param callback
     *           a callback function that receives a single string argument. Possible values are `"off"`,
     *           `"play_recorded"`, `"record_next_free_slot"`, `"stop"`, `"return_to_arrangement"`,
     *           `"return_to_previous_clip"` or `"play_random"`.
     * @since API version 1
     * @deprecated Use {@link #clipLauncherPostRecordingAction()} instead.
     */
    addClipLauncherPostRecordingActionObserver(
      callback: EnumValueChangedCallback
    ): void;

    /**
     * Sets the automation write mode.
     *
     * @param action
     *           the string that identifies the new automation write mode. Possible values are `"off"`,
     *           `"play_recorded"`, `"record_next_free_slot"`, `"stop"`, `"return_to_arrangement"`,
     *           `"return_to_previous_clip"` or `"play_random"`.
     * @since API version 1
     * @deprecated Use {@link #clipLauncherPostRecordingAction()} instead.
     */
    setClipLauncherPostRecordingAction(action: string): void;

    /**
     * Returns an object that provides access to the clip launcher post recording time offset.
     *
     * @return a beat time object that represents the post recording time offset
     * @since API version 1
     */
    getClipLauncherPostRecordingTimeOffset(): SettableBeatTimeValue;
  }

  // source: com/bitwig/extension/controller/api/UserControlBank.java

  /**
   * Instances of this interface represent a bank of custom controls that can be manually learned to device
   * parameters by the user.
   *
   * @since API version 1
   */
  interface UserControlBank {
    /**
     * Gets the user control at the given bank index.
     *
     * @param index
     *           the index of the control within the bank
     * @return the requested user control object
     * @since API version 1
     */
    getControl(index: number): Parameter;
  }

  // source: com/bitwig/extension/controller/api/Value.java

  import ValueChangedCallback = com.bitwig.extension.callback.ValueChangedCallback;

  /**
   * The common interface that is shared by all value objects in the controller API.
   *
   * @since API version 1
   */
  interface Value<
    ObserverType extends ValueChangedCallback = ValueChangedCallback
  > extends Subscribable {
    /**
     * Marks this value as being of interest to the driver. This can only be called once during the driver's
     * init method. A value that is of interest to the driver can be obtained using the value's get method. If
     * a value has not been marked as interested then an error will be reported if the driver attempts to get
     * the current value. Adding an observer to a value will automatically mark this value as interested.
     *
     * @since API version 2
     */
    markInterested(): void;

    /**
     * Registers an observer that reports the current value.
     *
     * @param callback
     *           a callback function that receives a single parameter
     * @since API version 1
     */
    addValueObserver(callback: ObserverType): void;
  }
}

declare namespace java {
  namespace nio {
    type ByteBuffer = any;
  }
  namespace util {
    type List<T> = Array<T>;
    type UUID = string;

    namespace func {
      type BooleanSupplier = {
        getAsBoolean(): boolean;
      };
      type Consumer<T = any> = {
        accept(t: T): void;
        andThen(after: Consumer<T>): Consumer<T>;
      };
      type Supplier<T = any> = {
        get(): T;
      };
      type DoubleConsumer = {
        accept(double: number): void;
        andThen(after: DoubleConsumer): DoubleConsumer;
      };
      type DoubleSupplier = {
        getAsDouble(): number;
      };
      type IntConsumer = {
        accept(int: number): void;
        andThen(after: IntConsumer): IntConsumer;
      };
      type IntSupplier = {
        getAsInt(): number;
      };
      interface Function<T = any, R = any> {
        andThen(after: Function): Function<T>;
        apply(t: T): R;
        compose(before: Function): Function<any, R>;
      }
      type IntFunction<R = any> = {
        apply(value: number): R;
      };
    }
  }
}

import API = com.bitwig.extension.controller.api;

declare const host: API.ControllerHost;
declare const loadAPI: typeof host.loadAPI;
declare const load: typeof host.load;
declare const println: typeof host.println;
declare const errorln: typeof host.errorln;
declare function dump(obj: any): void;
